import { Group, DisplayObject, AABB, IAnimation } from '@antv/g';
import { EdgeShapeMap } from '../types/edge';
import { DisplayMapper, IItem, ItemDisplayModel, ItemModel, ItemModelData, ItemShapeStyles, ITEM_TYPE, State, LodStrategyObj } from '../types/item';
import { NodeShapeMap } from '../types/node';
import { EdgeStyleSet, NodeStyleSet } from '../types/theme';
export default abstract class Item implements IItem {
    destroyed: boolean;
    /** Inner model. */
    model: ItemModel;
    /** Display model, user will not touch it. */
    displayModel: ItemDisplayModel;
    /** The mapper configured at graph with field name 'node' / 'edge' / 'combo'. */
    mapper: DisplayMapper;
    /** The state sstyle mapper configured at traph with field name 'nodeState' / 'edgeState' / 'comboState'. */
    stateMapper: {
        [stateName: string]: DisplayMapper;
    };
    /** The graphic group for item drawing. */
    group: Group;
    /** The keyShape of the item. */
    keyShape: DisplayObject;
    /** render extension for this item. */
    renderExt: any;
    /** Visibility. */
    visible: boolean;
    /** The states on the item. */
    states: {
        name: string;
        value: string | boolean;
    }[];
    /** The map caches the shapes of the item. The key is the shape id, the value is the g shape. */
    shapeMap: NodeShapeMap | EdgeShapeMap;
    afterDrawShapeMap: {};
    /** Set to different value in implements. */
    type: ITEM_TYPE;
    /** The flag of transient item. */
    transient: boolean;
    renderExtensions: any;
    /** Cache the shape animation instances to stop at next lifecycle. */
    animations: IAnimation[];
    /** Cache the group animation instances to stop at next lifecycle. */
    groupAnimations: IAnimation[];
    themeStyles: {
        default: ItemShapeStyles;
        [stateName: string]: ItemShapeStyles;
    };
    /** The zoom strategy to show and hide shapes according to their lod. */
    lodStrategy: LodStrategyObj;
    /** Last zoom ratio. */
    zoom: number;
    /** Cache the chaging states which are not consomed by draw  */
    changedStates: string[];
    /** The listener for the animations frames. */
    onframe: Function;
    private device;
    /** Cache the dirty tags for states when data changed, to re-map the state styles when state changed */
    private stateDirtyMap;
    private cacheStateStyles;
    private cacheHiddenShape;
    constructor(props: any);
    /** Initiate the item. */
    init(props: any): void;
    /**
     * Draws the shapes.
     * @internal
     * */
    draw(displayModel: ItemDisplayModel, diffData?: {
        previous: ItemModelData;
        current: ItemModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }, animate?: boolean, onfinish?: Function): void;
    /**
     * Updates the shapes.
     * @internal
     * */
    update(model: ItemModel, diffData?: {
        previous: ItemModelData;
        current: ItemModelData;
    }, isReplace?: boolean, itemTheme?: {
        styles: NodeStyleSet | EdgeStyleSet;
        lodStrategy: LodStrategyObj;
    }, onlyMove?: boolean, animate?: boolean, onfinish?: Function): void;
    /**
     * Update the group's position, e.g. node, combo.
     * @param displayModel
     * @param diffData
     * @param onfinish
     * @returns
     */
    updatePosition(displayModel: ItemDisplayModel, diffData?: {
        previous: ItemModelData;
        current: ItemModelData;
    }, animate?: boolean, onfinish?: Function): void;
    /**
     * Maps (mapper will be function, value, or encode format) model to displayModel and find out the shapes to be update for incremental updating.
     * @param model inner model
     * @param diffData changes from graphCore changed event
     * @param isReplace whether replace the whole data or partial update
     * @returns
     */
    getDisplayModelAndChanges(innerModel: ItemModel, diffData?: {
        previous: ItemModelData;
        current: ItemModelData;
    }, isReplace?: boolean): {
        model: ItemDisplayModel;
        typeChange?: boolean;
    };
    getID(): import("@antv/graphlib").ID;
    getType(): ITEM_TYPE;
    /** Show the item. */
    show(animate?: boolean): void;
    /** Hides the item. */
    hide(animate?: boolean): void;
    /** Returns the visibility of the item. */
    isVisible(): boolean;
    /** Puts the item to the front in its graphic group. */
    toFront(): void;
    /** Puts the item to the back in its graphic group. */
    toBack(): void;
    /**
     * The state value for the item, false if the item does not have the state.
     * @param state state name
     * @returns { boolean | string } the state value
     */
    hasState(state: string): string | boolean;
    /**
     * Set the state for the item.
     * @param state state name
     * @param value state value
     */
    setState(state: string, value: string | boolean): void;
    /**
     * Clear the states for the item.
     * @param states the states to be cleared. All the states will be cleared if the states is not assigned
     */
    clearStates(states?: string[]): void;
    /**
     * Get all the states of the item.
     * @retruns states array with has { name: string, value: string | boolean } format item
     */
    getStates(): {
        name: string;
        value: string | boolean;
    }[];
    /**
     * Run some thing with animations, e.g. hide, show, destroy.
     * @param animates
     * @param timing
     * @param targetStyleMap
     * @param callback
     * @returns
     */
    private runWithAnimates;
    /**
     * Re-draw the item with merged state styles.
     * @param previousStates previous states
     * @returns
     */
    private drawWithStates;
    /**
     * Get the rendering bouding box of the keyShape.
     * @returns keyShape's rendering bounding box
     */
    getKeyBBox(): AABB;
    /**
     * Get the local bounding box for the keyShape.
     * */
    getLocalKeyBBox(): AABB;
    /**
     * Get the rendering bouding box of the whole item.
     * @returns item's rendering bounding box
     */
    getBBox(): AABB;
    /**
     * Stop all the animations on the item.
     */
    stopAnimations(): any[];
    /**
     * Animations' frame listemer.
     */
    animateFrameListener: Function;
    /**
     * Call render extension's onZoom to response the graph zooming.
     * @param zoom
     */
    updateZoom(zoom: any): void;
    /** Destroy the item. */
    destroy(): void;
}
