import { Group } from '@antv/g';
import { clone } from '@antv/util';
import { updateShapes } from '../util/shape';
import { animateShapes, getAnimatesExcludePosition } from '../util/animate';
import { getCircleIntersectByPoint, getEllipseIntersectByPoint, getNearestPoint, getRectIntersectByPoint, } from '../util/point';
import Item from './item';
export default class Node extends Item {
    constructor(props) {
        super(props);
        this.init(Object.assign(Object.assign({}, props), { type: props.type || 'node' }));
        this.draw(this.displayModel, undefined, undefined, !this.displayModel.data.disableAnimate, props.onfinish);
    }
    draw(displayModel, diffData, diffState, animate = true, onfinish = () => { }) {
        var _a;
        const { group, renderExt, shapeMap: prevShapeMap, model } = this;
        renderExt.mergeStyles(displayModel);
        const firstRendering = !((_a = this.shapeMap) === null || _a === void 0 ? void 0 : _a.keyShape);
        const shapeMap = renderExt.draw(displayModel, this.shapeMap, diffData, diffState);
        // add shapes to group, and update shapeMap
        this.shapeMap = updateShapes(prevShapeMap, shapeMap, group);
        const { animates, disableAnimate, x = 0, y = 0, z = 0 } = displayModel.data;
        if (firstRendering) {
            // first rendering, move the group
            group.setLocalPosition(x, y, z);
        }
        else {
            // terminate previous animations
            this.stopAnimations();
            this.updatePosition(displayModel, diffData, animate, onfinish);
        }
        const { haloShape } = this.shapeMap;
        haloShape === null || haloShape === void 0 ? void 0 : haloShape.toBack();
        super.draw(displayModel, diffData, diffState, animate, onfinish);
        renderExt.updateCache(this.shapeMap);
        if (firstRendering) {
            // update the transform
            renderExt.onZoom(this.shapeMap, this.zoom);
        }
        // handle shape's and group's animate
        if (animate && !disableAnimate && animates) {
            const animatesExcludePosition = getAnimatesExcludePosition(animates);
            this.animations = animateShapes(animatesExcludePosition, // animates
            renderExt.mergedStyles, // targetStylesMap
            this.shapeMap, // shapeMap
            group, firstRendering ? 'buildIn' : 'update', this.changedStates, this.animateFrameListener, (canceled) => onfinish(model.id, canceled));
        }
    }
    update(model, diffData, isReplace, theme, onlyMove, animate, onfinish) {
        super.update(model, diffData, isReplace, theme, onlyMove, animate, onfinish);
    }
    /**
     * Update the node's position,
     * do not update other styles which leads to better performance than updating position by updateData.
     */
    updatePosition(displayModel, diffData, animate, onfinish = () => { }) {
        const { group } = this;
        const { fx, fy, fz, x, y, z = 0, animates, disableAnimate, } = displayModel.data;
        const position = {
            x: fx === undefined ? x : fx,
            y: fy === undefined ? y : fy,
            z: fz === undefined ? z : fz,
        };
        if (isNaN(position.x) ||
            isNaN(position.y) ||
            isNaN(position.z))
            return;
        if (animate && !disableAnimate && (animates === null || animates === void 0 ? void 0 : animates.update)) {
            const groupAnimates = animates.update.filter(({ shapeId, fields = [] }) => (!shapeId || shapeId === 'group') &&
                (fields.includes('x') || fields.includes('y')));
            if (groupAnimates.length) {
                const animations = animateShapes({ update: groupAnimates }, { group: position }, // targetStylesMap
                this.shapeMap, // shapeMap
                group, 'update', [], this.animateFrameListener, (canceled) => onfinish(displayModel.id, canceled));
                this.groupAnimations = animations;
                return;
            }
        }
        group.setLocalPosition([position.x, position.y, position.z]);
        onfinish(displayModel.id, !animate);
    }
    clone(containerGroup, onlyKeyShape, disableAnimate) {
        if (onlyKeyShape) {
            const clonedKeyShape = this.shapeMap.keyShape.cloneNode();
            const pos = this.group.getPosition();
            const clonedGroup = new Group();
            clonedGroup.setPosition(pos);
            clonedGroup.appendChild(clonedKeyShape);
            containerGroup.appendChild(clonedGroup);
            return clonedGroup;
        }
        const clonedModel = clone(this.model);
        clonedModel.data.disableAnimate = disableAnimate;
        const clonedNode = new Node({
            model: clonedModel,
            renderExtensions: this.renderExtensions,
            containerGroup,
            mapper: this.mapper,
            stateMapper: this.stateMapper,
            zoom: this.zoom,
            theme: {
                styles: this.themeStyles,
                lodStrategy: this.lodStrategy,
            },
        });
        Object.keys(this.shapeMap).forEach((shapeId) => {
            if (!this.shapeMap[shapeId].isVisible())
                clonedNode.shapeMap[shapeId].hide();
        });
        return clonedNode;
    }
    getAnchorPoint(point, anchorIdx) {
        const { anchorPoints = [] } = this.displayModel.data;
        return this.getIntersectPoint(point, this.getPosition(), anchorPoints, anchorIdx);
    }
    getIntersectPoint(point, innerPoint, anchorPoints, anchorIdx) {
        const { keyShape } = this.shapeMap;
        const shapeType = keyShape.nodeName;
        const { x, y, z } = innerPoint;
        const keyShapeRenderBBox = keyShape.getRenderBounds();
        const keyShapeWidth = keyShapeRenderBBox.max[0] - keyShapeRenderBBox.min[0];
        const keyShapeHeight = keyShapeRenderBBox.max[1] - keyShapeRenderBBox.min[1];
        const anchorPositions = anchorPoints.map((pointRatio) => {
            const [xRatio, yRatio] = pointRatio;
            return {
                x: keyShapeWidth * xRatio + keyShapeRenderBBox.min[0],
                y: keyShapeHeight * yRatio + keyShapeRenderBBox.min[1],
            };
        });
        if (anchorIdx !== undefined && anchorPositions[anchorIdx]) {
            return anchorPositions[anchorIdx];
        }
        let intersectPoint;
        switch (shapeType) {
            case 'circle':
                intersectPoint = getCircleIntersectByPoint({
                    x,
                    y,
                    r: keyShape.attributes.r,
                }, point);
                break;
            case 'ellipse':
                intersectPoint = getEllipseIntersectByPoint({
                    x,
                    y,
                    rx: keyShape.attributes.rx,
                    ry: keyShape.attributes.ry,
                }, point);
                break;
            case 'mesh':
                intersectPoint = innerPoint;
                break;
            default: {
                intersectPoint = getRectIntersectByPoint({
                    x: keyShapeRenderBBox.min[0],
                    y: keyShapeRenderBBox.min[1],
                    width: keyShapeRenderBBox.max[0] - keyShapeRenderBBox.min[0],
                    height: keyShapeRenderBBox.max[1] - keyShapeRenderBBox.min[1],
                }, point);
            }
        }
        let linkPoint = intersectPoint;
        // If the node has anchorPoints in the data, find the nearest anchor point.
        if (anchorPoints.length) {
            if (!linkPoint) {
                // If the linkPoint is failed to calculate.
                linkPoint = point;
            }
            linkPoint = getNearestPoint(anchorPositions, linkPoint).nearestPoint;
        }
        if (!linkPoint) {
            // If the calculations above are all failed, return the data's position
            return { x, y };
        }
        return linkPoint;
    }
    getPosition() {
        const initiated = this.shapeMap.keyShape && this.group.attributes.x !== undefined;
        if (initiated) {
            const { center } = this.shapeMap.keyShape.getRenderBounds();
            return { x: center[0], y: center[1], z: center[2] };
        }
        const { x = 0, y = 0, z = 0, fx, fy, fz } = this.model.data;
        return {
            x: (fx === undefined ? x : fx),
            y: (fy === undefined ? y : fy),
            z: (fz === undefined ? z : fz),
        };
    }
}
//# sourceMappingURL=node.js.map