var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { Group } from '@antv/g';
import { debounce, each, isArray, isObject, map, throttle, uniq, uniqueId, } from '@antv/util';
import { DirectionalLight, AmbientLight } from '@antv/g-plugin-3d';
import registry from '../../stdlib';
import Node from '../../item/node';
import Edge from '../../item/edge';
import Combo from '../../item/combo';
import { getCombinedBoundsByData, upsertShape } from '../../util/shape';
import { getExtension } from '../../util/extension';
import { upsertTransientItem } from '../../util/item';
import { formatLodStrategy } from '../../util/zoom';
import { deconstructData, graphComboTreeDfs, graphCoreTreeDfs, traverseAncestors, traverseAncestorsAndSucceeds, traverseGraphAncestors, } from '../../util/data';
import { getGroupedChanges } from '../../util/event';
import { EdgeCollisionChecker, QuadTree } from '../../util/polyline';
/**
 * Manages and stores the node / edge / combo items.
 */
export class ItemController {
    constructor(graph) {
        this.nodeExtensions = [];
        this.edgeExtensions = [];
        this.comboExtensions = [];
        /**
         * Node / edge / combo items map
         */
        this.itemMap = new Map();
        this.nodeDataTypeSet = new Set();
        this.edgeDataTypeSet = new Set();
        this.comboDataTypeSet = new Set();
        // The G shapes or groups on transient map drawn by this controller
        this.transientObjectMap = new Map();
        this.transientItemMap = new Map();
        this.onViewportChange = debounce(({ transform, effectTiming }) => {
            const { zoom } = transform;
            if (zoom) {
                const zoomRatio = this.graph.getZoom();
                this.itemMap.forEach((item) => item.updateZoom(zoomRatio));
                this.zoom = zoomRatio;
            }
        }, 500, false);
        this.onThemeChange = ({ theme }) => {
            if (!theme)
                return;
            const { nodeDataTypeSet, edgeDataTypeSet } = this;
            const { node: nodeTheme, edge: edgeTheme } = theme;
            this.itemMap.forEach((item) => {
                const itemTye = item.getType();
                const usingTheme = itemTye === 'node' ? nodeTheme : edgeTheme;
                const usingTypeSet = itemTye === 'node' ? nodeDataTypeSet : edgeDataTypeSet;
                const { dataTypeField } = usingTheme;
                let dataType;
                if (dataTypeField)
                    dataType = item.model.data[dataTypeField];
                const itemTheme = getItemTheme(usingTypeSet, dataTypeField, dataType, usingTheme);
                item.update(item.model, undefined, false, itemTheme);
            });
        };
        this.onDestroy = () => {
            Object.values(this.itemMap).forEach((item) => item.destroy());
            // Fix OOM problem, since this map will hold all the refs of items.
            this.itemMap.clear();
        };
        this.graph = graph;
        // get mapper for node / edge / combo
        const { node, edge, combo, nodeState = {}, edgeState = {}, comboState = {}, } = graph.getSpecification();
        this.nodeMapper = node;
        this.edgeMapper = edge;
        this.comboMapper = combo;
        this.nodeStateMapper = nodeState;
        this.edgeStateMapper = edgeState;
        this.comboStateMapper = comboState;
        this.tap();
    }
    /**
     * Subscribe the lifecycle of graph.
     */
    tap() {
        // item extensions are node / edge / combo type registrations
        const extensions = this.getExtensions();
        this.nodeExtensions = extensions.node;
        this.edgeExtensions = extensions.edge;
        this.comboExtensions = extensions.combo;
        this.graph.hooks.render.tap(this.onRender.bind(this));
        this.graph.hooks.itemchange.tap(this.onChange.bind(this));
        this.graph.hooks.itemstatechange.tap(this.onItemStateChange.bind(this));
        this.graph.hooks.itemvisibilitychange.tap(this.onItemVisibilityChange.bind(this));
        this.graph.hooks.itemzindexchange.tap(this.onItemZIndexChange.bind(this));
        this.graph.hooks.transientupdate.tap(this.onTransientUpdate.bind(this));
        this.graph.hooks.viewportchange.tap(this.onViewportChange.bind(this));
        this.graph.hooks.themechange.tap(this.onThemeChange.bind(this));
        this.graph.hooks.treecollapseexpand.tap(this.onTreeCollapseExpand.bind(this));
        this.graph.hooks.destroy.tap(this.onDestroy.bind(this));
    }
    /**
     * Get the extensions from useLib, stdLib is a subset of useLib.
     */
    getExtensions() {
        // TODO: user need to config using node/edge/combo types from useLib to spec?
        const { node, edge, combo } = this.graph.getSpecification();
        const nodeTypes = Object.keys(registry.useLib.nodes || {});
        const edgeTypes = Object.keys(registry.useLib.edges || {});
        const comboTypes = Object.keys(registry.useLib.combos || {});
        return {
            node: nodeTypes
                .map((config) => getExtension(config, registry.useLib, 'node'))
                .filter(Boolean),
            edge: edgeTypes
                .map((config) => getExtension(config, registry.useLib, 'edge'))
                .filter(Boolean),
            combo: comboTypes
                .map((config) => getExtension(config, registry.useLib, 'combo'))
                .filter(Boolean),
        };
    }
    /**
     * Listener of runtime's render hook.
     * @param param contains inner data stored in graphCore structure
     */
    onRender(param) {
        const { graphCore, theme = {}, transientCanvas } = param;
        const { graph } = this;
        // 0. clear groups on canvas, and create new groups
        graph.canvas.removeChildren();
        const comboGroup = new Group({ id: 'combo-group' });
        const edgeGroup = new Group({ id: 'edge-group' });
        const nodeGroup = new Group({ id: 'node-group' });
        graph.canvas.appendChild(comboGroup);
        graph.canvas.appendChild(edgeGroup);
        graph.canvas.appendChild(nodeGroup);
        this.nodeGroup = nodeGroup;
        this.edgeGroup = edgeGroup;
        this.comboGroup = comboGroup;
        // Also create transient groups on transient canvas.
        transientCanvas.removeChildren();
        this.transientComboGroup = new Group({ id: 'combo-group' });
        this.transientEdgeGroup = new Group({ id: 'edge-group' });
        this.transientNodeGroup = new Group({ id: 'node-group' });
        transientCanvas.appendChild(this.transientComboGroup);
        transientCanvas.appendChild(this.transientEdgeGroup);
        transientCanvas.appendChild(this.transientNodeGroup);
        // 1. create lights for webgl 3d rendering
        if (graph.rendererType === 'webgl-3d') {
            const ambientLight = new AmbientLight({
                style: {
                    fill: 'white',
                    intensity: Math.PI * 2,
                },
            });
            const light = new DirectionalLight({
                style: {
                    fill: 'white',
                    direction: [-1, 0, 1],
                    intensity: Math.PI * 0.7,
                },
            });
            // @ts-ignore
            graph.canvas.appendChild(ambientLight);
            // @ts-ignore
            graph.canvas.appendChild(light);
            const { width, height } = graph.canvas.getConfig();
            graph.canvas.getCamera().setPerspective(0.1, 50000, 45, width / height);
        }
        // 2. create node / edge / combo items, classes from ../../item, and element drawing and updating fns from node/edge/comboExtensions
        const { nodes, edges, combos } = deconstructData({
            nodes: graphCore.getAllNodes(),
            edges: graphCore.getAllEdges(),
        });
        this.renderNodes(nodes, theme.node);
        this.renderCombos(combos, theme.combo, graphCore);
        this.renderEdges(edges, theme.edge);
        this.sortByComboTree(graphCore);
        // collapse the combos which has 'collapsed' in initial data
        if (graphCore.hasTreeStructure('combo')) {
            graphCoreTreeDfs(graphCore, graphCore.getRoots('combo'), (child) => {
                if (child.data.collapsed)
                    this.collapseCombo(graphCore, child);
            }, 'BT', 'combo');
        }
        // collapse the sub tree which has 'collapsed' in initial data
        if (graphCore.hasTreeStructure('tree')) {
            const collapseNodes = [];
            graphCoreTreeDfs(graphCore, graphCore.getRoots('tree'), (child) => {
                if (child.data.collapsed)
                    collapseNodes.push(child);
            }, 'BT', 'tree');
            this.collapseSubTree(collapseNodes, graphCore, false);
        }
    }
    /**
     * Listener of runtime's itemchange lifecycle hook.
     * @param param
     */
    onChange(param) {
        const { changes, graphCore, action, animate = true, upsertAncestors = true, theme = {}, callback = () => { }, } = param;
        const groupedChanges = getGroupedChanges(graphCore, changes);
        const { itemMap } = this;
        // change items according to the order of the keys in groupedChanges
        // === 1. remove edges; 2. remove nodes ===
        [...groupedChanges.EdgeRemoved, ...groupedChanges.NodeRemoved].forEach(({ value }) => {
            const { id } = value;
            const item = itemMap.get(id);
            if (item) {
                item.destroy();
                itemMap.delete(id);
            }
        });
        const { node: nodeTheme = {}, edge: edgeTheme = {}, combo: comboTheme = {}, } = theme;
        // === 3. add nodes ===
        if (groupedChanges.NodeAdded.length) {
            const newNodes = [];
            const newCombos = [];
            groupedChanges.NodeAdded.map((change) => change.value).forEach((model) => {
                if (model.data._isCombo)
                    newCombos.push(model);
                else
                    newNodes.push(model);
            });
            if (newNodes.length) {
                this.renderNodes(newNodes, nodeTheme);
            }
            if (newCombos.length) {
                this.renderCombos(newCombos, comboTheme, graphCore);
            }
        }
        // === 4. add edges ===
        if (groupedChanges.EdgeAdded.length) {
            this.renderEdges(groupedChanges.EdgeAdded.map((change) => change.value), edgeTheme);
        }
        // === 5. update nodes's data ===
        // merge changes for each node or combo
        if (groupedChanges.NodeDataUpdated.length) {
            const nodeComboUpdate = {};
            groupedChanges.NodeDataUpdated.forEach((change) => {
                const { id, propertyName, newValue, oldValue } = change;
                nodeComboUpdate[id] = nodeComboUpdate[id] || {
                    id,
                    previous: {},
                    current: {},
                };
                if (!propertyName) {
                    nodeComboUpdate[id] = {
                        id,
                        isReplace: true,
                        previous: oldValue,
                        current: newValue,
                    };
                }
                else {
                    nodeComboUpdate[id].previous[propertyName] = oldValue;
                    nodeComboUpdate[id].current[propertyName] = newValue;
                }
            });
            const { dataTypeField: nodeDataTypeField } = nodeTheme;
            const edgeIdsToUpdate = new Set();
            const comboIdsToUpdate = new Set();
            const updateRelates = (edgeIds) => {
                const ids = edgeIds
                    ? [...edgeIds]
                    : [...comboIdsToUpdate, ...edgeIdsToUpdate];
                ids.forEach((nid) => {
                    const item = itemMap.get(nid);
                    if (item && !item.destroyed)
                        item.forceUpdate();
                });
            };
            const debounceUpdateRelates = debounce(updateRelates, 16, false);
            Object.values(nodeComboUpdate).forEach((updateObj) => {
                var _a;
                const { isReplace, previous, current, id } = updateObj;
                if (!graphCore.hasNode(id))
                    return;
                const onlyMove = action === 'updatePosition';
                const item = itemMap.get(id);
                if (!item || item.destroyed)
                    return;
                const type = item.getType();
                const innerModel = graphCore.getNode(id);
                if (onlyMove &&
                    type === 'node' &&
                    isNaN(current.x) &&
                    isNaN(current.y)) {
                    callback(innerModel, true);
                    return;
                }
                const nodeRelatedIdsToUpdate = new Set();
                // collapse and expand
                if (graphCore.hasTreeStructure('combo')) {
                    if (type === 'combo' && current.collapsed !== previous.collapsed) {
                        if (current.collapsed) {
                            this.collapseCombo(graphCore, innerModel);
                        }
                        else if (current.collapsed === false) {
                            this.expandCombo(graphCore, innerModel);
                        }
                    }
                    const previousParentId = item.displayModel.data.parentId;
                    // update the current parent combo tree
                    // if the node has previous parent, related previous parent combo should be updated to
                    if (upsertAncestors) {
                        const begins = [innerModel];
                        if (previousParentId &&
                            previousParentId !== current.parentId &&
                            graphCore.hasNode(previousParentId)) {
                            begins.push(graphCore.getNode(previousParentId));
                        }
                        // ancestors and suceeds combos should be updated
                        traverseAncestorsAndSucceeds(this.graph, graphCore, begins, (treeItem) => {
                            if (treeItem.data._isCombo && treeItem.id !== innerModel.id)
                                comboIdsToUpdate.add(treeItem.id);
                            const relatedEdges = graphCore.getRelatedEdges(treeItem.id);
                            relatedEdges.forEach((edge) => {
                                edgeIdsToUpdate.add(edge.id);
                                nodeRelatedIdsToUpdate.add(edge.id);
                            });
                        });
                    }
                    else if (type === 'combo') {
                        // only the succeed combos should be updated
                        graphComboTreeDfs(this.graph, [innerModel], (child) => {
                            const relatedEdges = graphCore.getRelatedEdges(child.id);
                            relatedEdges.forEach((edge) => {
                                edgeIdsToUpdate.add(edge.id);
                                nodeRelatedIdsToUpdate.add(edge.id);
                            });
                            if (child.data._isCombo && child.id !== innerModel.id)
                                comboIdsToUpdate.add(child.id);
                        });
                    }
                }
                // update the theme if the dataType value is changed
                let itemTheme;
                if (nodeDataTypeField &&
                    previous[nodeDataTypeField] !== current[nodeDataTypeField]) {
                    itemTheme = getItemTheme(this.nodeDataTypeSet, nodeDataTypeField, current[nodeDataTypeField], nodeTheme);
                }
                const preventPolylineEdgeOverlap = ((_a = innerModel === null || innerModel === void 0 ? void 0 : innerModel.data) === null || _a === void 0 ? void 0 : _a.preventPolylineEdgeOverlap) || false;
                const relatedEdgeInnerModels = preventPolylineEdgeOverlap
                    ? this.findNearEdgesByNode(id, graphCore).concat(graphCore.getRelatedEdges(id))
                    : graphCore.getRelatedEdges(id);
                relatedEdgeInnerModels.forEach((edge) => {
                    edgeIdsToUpdate.add(edge.id);
                    nodeRelatedIdsToUpdate.add(edge.id);
                });
                item.onframe = () => updateRelates(nodeRelatedIdsToUpdate);
                let statesCache;
                if (innerModel.data._isCombo &&
                    previous.collapsed !== current.collapsed) {
                    statesCache = this.graph.getItemAllStates(id);
                    this.graph.clearItemState(id);
                }
                item.update(innerModel, { previous, current }, isReplace, itemTheme, onlyMove, animate, 
                // call after updating finished
                throttle((_, canceled) => {
                    var _a;
                    (_a = item.onframe) === null || _a === void 0 ? void 0 : _a.call(item, true);
                    item.onframe = undefined;
                    if (statesCache) {
                        statesCache.forEach((state) => this.graph.setItemState(id, state, true));
                    }
                    callback(innerModel, canceled);
                }, 500, {
                    leading: false,
                    trailing: true,
                }));
                const parentItem = this.itemMap.get(current.parentId);
                if (current.parentId && (parentItem === null || parentItem === void 0 ? void 0 : parentItem.model.data.collapsed)) {
                    this.graph.executeWithoutStacking(() => {
                        this.graph.hideItem(innerModel.id, false);
                    });
                }
            });
            debounceUpdateRelates();
        }
        // === 6. update edges' data ===
        if (groupedChanges.EdgeDataUpdated.length) {
            const edgeUpdate = {};
            groupedChanges.EdgeDataUpdated.forEach((change) => {
                const { id, propertyName, newValue, oldValue } = change;
                edgeUpdate[id] = edgeUpdate[id] || { id, previous: {}, current: {} };
                if (!propertyName) {
                    edgeUpdate[id] = {
                        id,
                        isReplace: true,
                        previous: oldValue,
                        current: newValue,
                    };
                }
                else {
                    edgeUpdate[id].previous[propertyName] = oldValue;
                    edgeUpdate[id].current[propertyName] = newValue;
                }
            });
            const { dataTypeField: edgeDataTypeField = '' } = edgeTheme;
            Object.values(edgeUpdate).forEach((updateObj) => {
                const { isReplace, current, previous, id } = updateObj;
                // update the theme if the dataType value is changed
                let itemTheme;
                if (previous[edgeDataTypeField] !== current[edgeDataTypeField]) {
                    itemTheme = getItemTheme(this.edgeDataTypeSet, edgeDataTypeField, current[edgeDataTypeField], edgeTheme);
                }
                const item = itemMap.get(id);
                const innerModel = graphCore.getEdge(id);
                if (!item || item.destroyed)
                    return;
                item.update(innerModel, { current, previous }, isReplace, itemTheme, undefined, animate, (_, canceled) => callback(innerModel, canceled));
            });
        }
        // === 7. update edges' source target ===
        if (groupedChanges.EdgeUpdated.length) {
            const edgeUpdate = {};
            groupedChanges.EdgeUpdated.forEach((change) => {
                // propertyName is 'source' or 'target'
                const { id, propertyName, newValue } = change;
                edgeUpdate[id] = edgeUpdate[id] || { id };
                edgeUpdate[id][propertyName] = newValue;
            });
            Object.values(edgeUpdate).forEach((updateObj) => {
                const { source, target, id } = updateObj;
                const item = itemMap.get(id);
                if (source !== undefined)
                    item.updateEnd('source', this.itemMap.get(source));
                if (target !== undefined)
                    item.updateEnd('target', this.itemMap.get(target));
            });
        }
        // === 8. combo tree structure change, resort the shapes ===
        if (groupedChanges.ComboStructureChanged.length) {
            this.sortByComboTree(graphCore);
        }
        // === 9. tree data structure change, hide the new node and edge while one of the ancestor is collapsed ===
        if (groupedChanges.TreeStructureChanged.length) {
            groupedChanges.TreeStructureChanged.forEach((change) => {
                const { nodeId } = change;
                // hide it when an ancestor is collapsed
                let parent = graphCore.getParent(nodeId, 'tree');
                while (parent) {
                    if (parent.data.collapsed) {
                        this.graph.hideItem(nodeId, true);
                        break;
                    }
                    parent = graphCore.getParent(parent.id, 'tree');
                }
            });
        }
    }
    /**
     * The listener for item state changing.
     * @param param
     * {
     *   ids: ids of the items to be set state
     *   states: state names to set
     *   value: state value
     * }
     */
    onItemStateChange(param) {
        const { ids, states, value } = param;
        ids.forEach((id) => {
            const item = this.itemMap.get(id);
            if (!item) {
                console.warn(`Fail to set state for item ${id}, which is not exist.`);
                return;
            }
            if (!states || !value) {
                // clear all the states
                item.clearStates(states);
            }
            else {
                states.forEach((state) => item.setState(state, value));
            }
        });
    }
    onItemVisibilityChange(param) {
        const { ids, value, graphCore, animate = true } = param;
        ids.forEach((id) => {
            const item = this.itemMap.get(id);
            if (!item) {
                console.warn(`Fail to set visibility for item ${id}, which is not exist.`);
                return;
            }
            const type = item.getType();
            if (value) {
                if (type === 'edge') {
                    item.show(animate);
                }
                else {
                    if (graphCore.hasTreeStructure('combo')) {
                        let anccestorCollapsed = false;
                        traverseAncestors(graphCore, [item.model], (model) => {
                            if (model.data.collapsed)
                                anccestorCollapsed = true;
                            return anccestorCollapsed;
                        });
                        if (anccestorCollapsed)
                            return;
                    }
                    const relatedEdges = graphCore.getRelatedEdges(id);
                    item.show(animate);
                    relatedEdges.forEach(({ id: edgeId, source, target }) => {
                        var _a;
                        if (this.getItemVisible(source) && this.getItemVisible(target))
                            (_a = this.itemMap.get(edgeId)) === null || _a === void 0 ? void 0 : _a.show(animate);
                    });
                }
            }
            else {
                item.hide(animate);
                if (type !== 'edge') {
                    const relatedEdges = graphCore.getRelatedEdges(id);
                    relatedEdges.forEach(({ id: edgeId }) => {
                        var _a;
                        (_a = this.itemMap.get(edgeId)) === null || _a === void 0 ? void 0 : _a.hide(animate);
                    });
                }
            }
        });
    }
    onItemZIndexChange(params) {
        const { ids = [], action, graphCore } = params;
        ids.forEach((id) => {
            const item = this.itemMap.get(id);
            if (!item)
                return;
            if (action === 'front') {
                item.toFront();
                if (graphCore.hasTreeStructure('combo')) {
                    graphComboTreeDfs(this.graph, [item.model], (model) => {
                        if (model.data._isCombo) {
                            const subCombo = this.itemMap.get(model.id);
                            subCombo && subCombo.toFront();
                        }
                    }, 'TB');
                }
            }
            else {
                item.toBack();
                if (graphCore.hasTreeStructure('combo')) {
                    traverseGraphAncestors(this.graph, [item.model], (model) => {
                        var _a;
                        (_a = this.itemMap.get(model.id)) === null || _a === void 0 ? void 0 : _a.toBack();
                    });
                }
            }
        });
    }
    onTransientUpdate(param) {
        const { transientObjectMap } = this;
        const { type, id, config = {}, canvas, graphCore } = param;
        const { style = {}, data = {}, capture, action, onlyDrawKeyShape, upsertAncestors, } = config;
        const isItemType = type === 'node' || type === 'edge' || type === 'combo';
        // Removing
        if (action === 'remove') {
            if (isItemType) {
                const transientItem = this.transientItemMap.get(id);
                if (type === 'combo') {
                    // remove children from bottom to top
                    graphCoreTreeDfs(graphCore, [graphCore.getNode(id)], (child) => {
                        const transientChild = this.transientItemMap.get(child.id);
                        if (transientChild && !transientChild.destroyed) {
                            transientChild.destroy();
                        }
                        this.transientItemMap.delete(child.id);
                    }, 'BT');
                }
                if (transientItem && !transientItem.destroyed) {
                    transientItem.destroy();
                }
                this.transientItemMap.delete(id);
                return;
            }
            else {
                const preObj = transientObjectMap.get(id);
                if (preObj && !preObj.destroyed)
                    preObj.destroy();
                transientObjectMap.delete(id);
                return;
            }
        }
        // Adding / Updating
        if (isItemType) {
            const item = this.itemMap.get(id);
            if (!item) {
                console.warn(`Fail to draw transient item of ${id}, which is not exist.`);
                return;
            }
            const transientItem = upsertTransientItem(item, this.transientNodeGroup, this.transientEdgeGroup, this.transientComboGroup, this.transientItemMap, this.itemMap, graphCore, onlyDrawKeyShape, upsertAncestors);
            if (onlyDrawKeyShape) {
                // only update node positions to cloned node container(group)
                if ((type === 'node' || type === 'combo') &&
                    data.hasOwnProperty('x') &&
                    data.hasOwnProperty('y')) {
                    const { x, y } = data;
                    transientItem.setPosition([x, y]);
                }
                // TODO: edge onlyDrawKeyShape?
            }
            else {
                const transItem = transientItem;
                const positionChanged = data.hasOwnProperty('x') && data.hasOwnProperty('y');
                if (type === 'combo' && positionChanged) {
                    const { x, y } = data;
                    const { x: ox = x, y: oy = y } = transItem.displayModel.data;
                    const dx = x - ox;
                    const dy = y - oy;
                    // move combo = move children nodes from bottom to top
                    graphCoreTreeDfs(graphCore, [transItem.model], (node) => {
                        const transChild = this.transientItemMap.get(node.id);
                        if (!transChild)
                            return;
                        const { x: childX = 0, y: childY = 0 } = transChild.model
                            .data;
                        transChild === null || transChild === void 0 ? void 0 : transChild.update(Object.assign(Object.assign({}, transChild.model), { data: Object.assign(Object.assign({}, transChild.model.data), { x: childX + dx, y: childY + dy }) }));
                    }, 'BT');
                }
                if (graphCore.hasTreeStructure('combo') &&
                    type !== 'edge' &&
                    positionChanged) {
                    // force update ancestor combos in the sametime
                    let currentAncestor = graphCore.getParent(transItem.model.id, 'combo');
                    while (currentAncestor) {
                        const ancestorItem = this.transientItemMap.get(currentAncestor.id);
                        if (ancestorItem)
                            ancestorItem.forceUpdate();
                        currentAncestor = graphCore.getParent(currentAncestor.id, 'combo');
                    }
                }
                // Update the target combo's posiiton and other props
                transItem.update(Object.assign(Object.assign({}, transItem.model), { data: Object.assign(Object.assign({}, transItem.model.data), data) }));
            }
            return;
        }
        const idStr = String(id);
        const shape = upsertShape(type, idStr, style, Object.fromEntries(transientObjectMap));
        transientObjectMap.set(idStr, shape);
        shape.style.pointerEvents = capture ? 'auto' : 'none';
        canvas.getRoot().appendChild(shape);
    }
    getTransient(id) {
        return this.transientObjectMap.get(id);
    }
    getTransientItem(id) {
        return this.transientItemMap[id];
    }
    findDisplayModel(id) {
        var _a;
        return (_a = this.itemMap.get(id)) === null || _a === void 0 ? void 0 : _a.displayModel;
    }
    /**
     * Create nodes with inner data to canvas.
     * @param models nodes' inner datas
     */
    renderNodes(models, nodeTheme = {}) {
        const { nodeExtensions, nodeGroup, nodeDataTypeSet, graph } = this;
        const { dataTypeField = '' } = nodeTheme;
        const zoom = graph.getZoom();
        models.forEach((node) => {
            // get the base styles from theme
            let dataType;
            if (dataTypeField)
                dataType = node.data[dataTypeField];
            const itemTheme = getItemTheme(nodeDataTypeSet, dataTypeField, dataType, nodeTheme);
            this.itemMap.set(node.id, new Node({
                model: node,
                renderExtensions: nodeExtensions,
                containerGroup: nodeGroup,
                mapper: this.nodeMapper,
                stateMapper: this.nodeStateMapper,
                zoom,
                theme: itemTheme,
                device: graph.rendererType === 'webgl-3d'
                    ? // TODO: G type
                        graph.canvas.context.deviceRendererPlugin.getDevice()
                    : undefined,
            }));
        });
    }
    renderCombos(models, comboTheme = {}, graphCore) {
        const { comboExtensions, comboGroup, comboDataTypeSet, graph, itemMap } = this;
        const { dataTypeField = '' } = comboTheme;
        const zoom = graph.getZoom();
        models.forEach((combo) => {
            // get the base styles from theme
            let dataType;
            if (dataTypeField)
                dataType = combo.data[dataTypeField];
            const itemTheme = getItemTheme(comboDataTypeSet, dataTypeField, dataType, comboTheme);
            const getCombinedBounds = () => {
                //  calculate the position of the combo according to its children
                return getCombinedBoundsByData(graph, graphCore
                    .getChildren(combo.id, 'combo')
                    .map(({ id }) => itemMap.get(id))
                    .filter(Boolean));
            };
            const getChildren = () => {
                const childModels = graphCore.getChildren(combo.id, 'combo');
                return childModels.map(({ id }) => itemMap.get(id));
            };
            const comboItem = new Combo({
                model: combo,
                getCombinedBounds,
                getChildren,
                renderExtensions: comboExtensions,
                containerGroup: comboGroup,
                mapper: this.comboMapper,
                stateMapper: this.comboStateMapper,
                zoom,
                theme: itemTheme,
                device: graph.rendererType === 'webgl-3d'
                    ? graph.canvas.context.deviceRendererPlugin.getDevice()
                    : undefined,
            });
            // ------- for integration tests -------
            comboItem.getCombinedBounds = getCombinedBounds;
            comboItem.getChildren = getChildren;
            comboItem.type = 'combo';
            // ------- for integration tests end-------
            itemMap.set(combo.id, comboItem);
        });
    }
    /**
     * Create edges with inner data to canvas.
     * @param models edges' inner datas
     */
    renderEdges(models, edgeTheme = {}) {
        const { edgeExtensions, edgeGroup, itemMap, edgeDataTypeSet, graph } = this;
        const { dataTypeField = '' } = edgeTheme;
        const zoom = graph.getZoom();
        const nodeMap = filterItemMapByType(itemMap, 'node');
        models.forEach((edge) => {
            const { source, target, id } = edge;
            const sourceItem = itemMap.get(source);
            const targetItem = itemMap.get(target);
            if (!sourceItem) {
                console.warn(`The source node ${source} is not exist in the graph for edge ${id}, please add the node first`);
                return;
            }
            if (!targetItem) {
                console.warn(`The source node ${source} is not exist in the graph for edge ${id}, please add the node first`);
                return;
            }
            // get the base styles from theme
            let dataType;
            if (dataTypeField)
                dataType = edge.data[dataTypeField];
            const itemTheme = getItemTheme(edgeDataTypeSet, dataTypeField, dataType, edgeTheme);
            itemMap.set(id, new Edge({
                model: edge,
                renderExtensions: edgeExtensions,
                containerGroup: edgeGroup,
                mapper: this.edgeMapper,
                stateMapper: this.edgeStateMapper,
                sourceItem,
                targetItem,
                nodeMap,
                zoom,
                theme: itemTheme,
            }));
        });
    }
    /**
     * Get the id of the item which have the state with true value
     * @param itemType item's type
     * @param state state name
     * @param value state value, true by default
     * @returns
     */
    findIdByState(itemType, state, value = true) {
        const ids = [];
        this.itemMap.forEach((item) => {
            if (item.getType() !== itemType)
                return;
            if (item.hasState(state) === value)
                ids.push(item.getID());
        });
        return ids;
    }
    /**
     * Get the state value for the item with id
     * @param id item' id
     * @param state state name
     * @returns {boolean | string} the state value
     */
    getItemState(id, state) {
        const item = this.itemMap.get(id);
        if (!item) {
            console.warn(`Fail to get item state, the item with id ${id} does not exist.`);
            return false;
        }
        return item.hasState(state);
    }
    getItemAllStates(id) {
        const item = this.itemMap.get(id);
        if (!item) {
            console.warn(`Fail to get item state, the item with id ${id} does not exist.`);
            return [];
        }
        return item
            .getStates()
            .filter(({ value }) => value)
            .map(({ name }) => name);
    }
    getItemById(id) {
        return this.itemMap.get(id);
    }
    getItemBBox(id, isKeyShape = false, isTransient = false) {
        const item = isTransient
            ? this.transientItemMap.get(id)
            : this.itemMap.get(id);
        if (!item) {
            console.warn(`Fail to get item bbox, the item with id ${id} does not exist.`);
            return false;
        }
        if (item instanceof Group)
            return item.getRenderBounds();
        return isKeyShape ? item.getKeyBBox() : item.getBBox();
    }
    getItemVisible(id) {
        const item = this.itemMap.get(id);
        if (!item) {
            console.warn(`Fail to get item visible, the item with id ${id} does not exist.`);
            return false;
        }
        return item.isVisible();
    }
    /**
     * Identify edges that are intersected by a particular node
     * @param nodeId node id
     * @param graphCore
     * @returns
     */
    findNearEdgesByNode(nodeId, graphCore) {
        var _a;
        const edges = graphCore.getAllEdges();
        const canvasBBox = this.graph.getRenderBBox(undefined);
        const quadTree = new QuadTree(canvasBBox, 4);
        each(edges, (edge) => {
            const { data: { x: sourceX, y: sourceY }, } = graphCore.getNode(edge.source);
            const { data: { x: targetX, y: targetY }, } = graphCore.getNode(edge.target);
            quadTree.insert({
                id: edge.id,
                p1: { x: sourceX, y: sourceY },
                p2: { x: targetX, y: targetY },
                bbox: this.graph.getRenderBBox(edge.id),
            });
        });
        // update node position
        const node = (this.getTransientItem(nodeId) ||
            this.getItemById(nodeId));
        const nodeBBox = this.graph.getRenderBBox(nodeId);
        const nodeData = (_a = node === null || node === void 0 ? void 0 : node.model) === null || _a === void 0 ? void 0 : _a.data;
        if (nodeData) {
            nodeBBox.update([nodeData.x, nodeData.y, 0], nodeBBox.halfExtents);
        }
        const checker = new EdgeCollisionChecker(quadTree);
        const collisions = checker.getCollidingEdges(nodeBBox);
        const collidingEdges = map(collisions, (collision) => graphCore.getEdge(collision.id));
        return collidingEdges;
    }
    sortByComboTree(graphCore) {
        if (!graphCore.hasTreeStructure('combo'))
            return;
        graphCoreTreeDfs(graphCore, graphCore.getRoots('combo'), (node) => {
            const nodeItem = this.itemMap.get(node.id);
            if (node.data._isCombo && nodeItem) {
                nodeItem.toFront();
            }
        });
    }
    collapseCombo(graphCore, comboModel) {
        let relatedEdges = [];
        const succeedIds = [];
        // find the succeeds in collapsed
        graphComboTreeDfs(this.graph, [comboModel], (child) => {
            if (child.id !== comboModel.id) {
                this.graph.executeWithoutStacking(() => {
                    this.graph.hideItem(child.id, false);
                });
            }
            relatedEdges = relatedEdges.concat(graphCore.getRelatedEdges(child.id));
            succeedIds.push(child.id);
        });
        const pairs = [];
        uniq(relatedEdges).forEach((edge) => {
            const { id, source: s, target: t } = edge;
            if (!this.graph.getItemVisible(id))
                return;
            const sourceIsSucceed = succeedIds.includes(s);
            const targetIsSucceed = succeedIds.includes(t);
            // do not add virtual edge if the source and target are both the succeed
            if (sourceIsSucceed && targetIsSucceed)
                return;
            const source = sourceIsSucceed ? comboModel.id : s;
            const target = targetIsSucceed ? comboModel.id : t;
            pairs.push({ source, target });
        });
        // each item in groupedEdges is a virtual edge
        this.graph.executeWithoutStacking(() => {
            this.graph.addData('edge', groupVirtualEdges(pairs));
        });
    }
    expandCombo(graphCore, comboModel) {
        let isAncestorCollapsed = false;
        traverseAncestors(graphCore, [comboModel], (anccestor) => {
            if (anccestor.data.collapsed) {
                isAncestorCollapsed = true;
                return true;
            }
            return false;
        });
        if (isAncestorCollapsed)
            return;
        const relatedVirtualEdgeIds = [];
        let edgesToShow = [];
        const nodesToShow = [];
        // show the succeeds and remove the related virtual edges, including the succeeds' related edges
        graphComboTreeDfs(this.graph, [comboModel], (child) => {
            graphCore.getRelatedEdges(child.id).forEach((edge) => {
                if (edge.data._virtual)
                    relatedVirtualEdgeIds.push(edge.id);
                else
                    edgesToShow.push(edge.id);
            });
            if (child.id !== comboModel.id) {
                if (!graphCore.getNode(child.data.parentId).data.collapsed) {
                    nodesToShow.push(child.id);
                }
                // re-add collapsed succeeds' virtual edges by calling collapseCombo
                if (child.data._isCombo && child.data.collapsed) {
                    this.collapseCombo(graphCore, child);
                }
            }
        });
        const virtualPairs = [];
        const visibleAncestorMap = new Map();
        edgesToShow = uniq(edgesToShow.filter((eid) => {
            const { source, target } = graphCore.getEdge(eid);
            const ends = { source, target };
            const endsVisible = {
                source: this.graph.getItemVisible(source) || nodesToShow.includes(source),
                target: this.graph.getItemVisible(target) || nodesToShow.includes(target),
            };
            // actual edges to show
            if (endsVisible.source && endsVisible.target)
                return true;
            // add virtual edges by finding the visible ancestor
            const virtualEnds = { source: undefined, target: undefined };
            Object.keys(virtualEnds).forEach((end) => {
                if (!endsVisible[end]) {
                    if (!visibleAncestorMap.get(ends[end])) {
                        traverseAncestors(graphCore, [graphCore.getNode(ends[end])], (ancestor) => {
                            if (visibleAncestorMap.has(ends[end]))
                                return;
                            if (this.graph.getItemVisible(ancestor.id) ||
                                nodesToShow.includes(ancestor.id))
                                visibleAncestorMap.set(ends[end], ancestor.id);
                        });
                    }
                    virtualEnds[end] = visibleAncestorMap.get(ends[end]);
                }
                else {
                    virtualEnds[end] = ends[end];
                }
            });
            if (virtualEnds.source !== undefined &&
                virtualEnds.target !== undefined) {
                virtualPairs.push(virtualEnds);
            }
            return false;
        }));
        // remove related virtual edges
        this.graph.executeWithoutStacking(() => {
            this.graph.removeData('edge', uniq(relatedVirtualEdgeIds));
            this.graph.showItem(edgesToShow.concat(nodesToShow));
            // add virtual edges by grouping visible ancestor edges
            this.graph.addData('edge', groupVirtualEdges(virtualPairs));
        });
    }
    /**
     * Collapse or expand a sub tree according to action
     * @param params
     */
    onTreeCollapseExpand(params) {
        const { ids, animate, action, graphCore } = params;
        const rootModels = ids.map((id) => graphCore.getNode(id));
        switch (action) {
            case 'collapse':
                this.collapseSubTree(rootModels, graphCore, animate);
                break;
            case 'expand':
            default:
                this.expandSubTree(rootModels, graphCore, animate);
                break;
        }
    }
    /**
     * Collapse sub tree(s).
     * @param rootModels The root node models of sub trees
     * @param graphCore
     * @param animate Whether enable animations for expanding, true by default
     * @returns
     */
    collapseSubTree(rootModels, graphCore, animate = true) {
        let positions = [];
        rootModels.forEach((root) => {
            let shouldCollapse = true;
            const nodes = [];
            graphCoreTreeDfs(graphCore, [root], (node) => {
                var _a;
                if (node.id === root.id)
                    return;
                const neighbors = graphCore.getNeighbors(node.id);
                if (!((_a = graphCore.getChildren(node.id, 'tree')) === null || _a === void 0 ? void 0 : _a.length) &&
                    neighbors.length > 1) {
                    shouldCollapse = false;
                }
                nodes.push(node);
            }, 'TB', 'tree', {
                stopAllFn: () => !shouldCollapse,
            });
            if (shouldCollapse) {
                positions = positions.concat(nodes.map((node) => ({
                    id: node.id,
                    data: { x: root.data.x, y: root.data.y },
                })));
            }
        });
        if (!positions.length)
            return;
        this.graph.updateNodePosition(positions, undefined, !animate, (model, canceled) => {
            this.graph.hideItem(model.id, canceled);
        }, undefined);
    }
    /**
     * Expand sub tree(s).
     * @param rootModels The root node models of sub trees.
     * @param graphCore
     * @param animate Whether enable animations for expanding, true by default.
     * @returns
     */
    expandSubTree(rootModels, graphCore, animate = true) {
        return __awaiter(this, void 0, void 0, function* () {
            let allNodeIds = [];
            let allEdgeIds = [];
            rootModels.forEach((root) => {
                const nodeIds = [];
                graphCoreTreeDfs(graphCore, [root], (node) => nodeIds.push(node.id), 'TB', 'tree', {
                    stopBranchFn: (node) => {
                        const shouldStop = node.id !== root.id && node.data.collapsed;
                        if (shouldStop)
                            nodeIds.push(node.id);
                        return shouldStop;
                    },
                });
                allEdgeIds = allEdgeIds.concat(graphCore
                    .getAllEdges()
                    .filter((edge) => nodeIds.includes(edge.source) && nodeIds.includes(edge.target))
                    .map((edge) => edge.id));
                allNodeIds = allNodeIds.concat(nodeIds.filter((id) => id !== root.id));
            });
            const ids = uniq(allNodeIds.concat(allEdgeIds));
            this.graph.showItem(ids, !animate);
            yield this.graph.layout(undefined, !animate);
        });
    }
}
const getItemTheme = (dataTypeSet, dataTypeField, dataType, itemTheme) => {
    const { styles: themeStyles = [], lodStrategy } = itemTheme;
    const formattedLodStrategy = formatLodStrategy(lodStrategy);
    if (!dataTypeField) {
        // dataType field is not assigned
        const styles = isArray(themeStyles)
            ? themeStyles[0]
            : Object.values(themeStyles)[0];
        return { styles, lodStrategy: formattedLodStrategy };
    }
    dataTypeSet.add(dataType);
    let themeStyle;
    if (isArray(themeStyles)) {
        const themeStylesLength = themeStyles.length;
        const idx = Array.from(dataTypeSet).indexOf(dataType);
        themeStyle = themeStyles[idx % themeStylesLength];
    }
    else if (isObject(themeStyles)) {
        themeStyle =
            themeStyles[dataType] ||
                themeStyles.others;
    }
    return {
        styles: themeStyle,
        lodStrategy: formattedLodStrategy,
    };
};
const filterItemMapByType = (itemMap, type) => {
    const filteredMap = new Map();
    itemMap.forEach((value, key) => {
        if (value.type === type) {
            filteredMap.set(key, value);
        }
    });
    return filteredMap;
};
const groupVirtualEdges = (pairs) => {
    const groupedVirtualEdges = new Map();
    pairs.forEach((edge) => {
        const { source, target } = edge;
        const key = `${source}_${target}`;
        const group = groupedVirtualEdges.get(key) || {
            edges: [],
            source,
            target,
        };
        group.edges.push(edge);
        groupedVirtualEdges.set(key, group);
    });
    // each item in groupedEdges is a virtual edge
    const virtualEdges = [];
    groupedVirtualEdges.forEach((group) => {
        const { source, target, edges } = group;
        virtualEdges.push({
            id: `virtual-${uniqueId()}`,
            source,
            target,
            data: {
                _virtual: true,
                keyShape: {
                    lineWidth: edges.length,
                },
            },
        });
    });
    return virtualEdges;
};
//# sourceMappingURL=item.js.map