var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { DisplayObject } from '@antv/g';
import { Graph as GraphLib } from '@antv/graphlib';
import { isLayoutWithIterations, Supervisor, } from '@antv/layout';
import registery, { stdLib } from '../../stdlib';
import { isImmediatelyInvokedLayoutOptions, isLayoutWorkerized, } from '../../types';
import { getNodeSizeFn, isComboLayout, layoutOneTree, radialLayout, isTreeLayout, } from '../../util/layout';
/**
 * Manages layout extensions and graph layout.
 * It will also emit `afterlayout` & `tick` events on Graph.
 */
export class LayoutController {
    constructor(graph) {
        this.extensions = {};
        this.graph = graph;
        this.animatedDisplayObject = new DisplayObject({});
        this.tap();
    }
    /**
     * Subscribe the lifecycle of graph.
     */
    tap() {
        this.graph.hooks.layout.tap(this.onLayout.bind(this));
        this.graph.hooks.destroy.tap(this.onDestroy.bind(this));
    }
    onLayout(params) {
        return __awaiter(this, void 0, void 0, function* () {
            /**
             * The final calculated result.
             */
            let positions;
            // Stop currentLayout if any.
            this.stopLayout();
            const { graphCore, options, animate = true } = params;
            let layoutNodes = graphCore.getAllNodes();
            if (!isComboLayout(options)) {
                layoutNodes = layoutNodes.filter((node) => this.graph.getItemVisible(node.id) && !node.data._isCombo);
            }
            const layoutNodesIdMap = {};
            layoutNodes.forEach((node) => (layoutNodesIdMap[node.id] = true));
            const layoutData = {
                nodes: layoutNodes,
                edges: graphCore
                    .getAllEdges()
                    .filter((edge) => layoutNodesIdMap[edge.source] && layoutNodesIdMap[edge.target]),
            };
            const layoutGraphCore = new GraphLib(layoutData);
            if (graphCore.hasTreeStructure('combo')) {
                layoutGraphCore.attachTreeStructure('combo');
                layoutNodes.forEach((node) => {
                    const parent = graphCore.getParent(node.id, 'combo');
                    if (parent && layoutGraphCore.hasNode(parent.id)) {
                        layoutGraphCore.setParent(node.id, parent.id, 'combo');
                    }
                });
            }
            this.graph.emit('startlayout');
            const [width, height] = this.graph.getSize();
            const center = [width / 2, height / 2];
            const nodeSize = getNodeSizeFn(options, 32);
            if (isImmediatelyInvokedLayoutOptions(options)) {
                const { animated = false, animationEffectTiming = {
                    duration: 1000,
                }, execute } = options, rest = __rest(options, ["animated", "animationEffectTiming", "execute"]);
                // It will ignore some layout options such as `type` and `workerEnabled`.
                positions = yield execute(layoutGraphCore, Object.assign({ nodeSize,
                    width,
                    height,
                    center }, rest));
                if (animated) {
                    yield this.animateLayoutWithoutIterations(positions, animationEffectTiming);
                }
            }
            else {
                const { type = 'grid', animated = false, animationEffectTiming = {
                    duration: 1000,
                }, iterations = 300 } = options, rest = __rest(options, ["type", "animated", "animationEffectTiming", "iterations"]);
                let { workerEnabled = false } = options;
                // Find built-in layout algorithms.
                const layoutCtor = stdLib.layouts[type] || registery.useLib.layouts[type];
                if (!layoutCtor) {
                    throw new Error(`Unknown layout algorithm: ${type}`);
                }
                if (isTreeLayout(options)) {
                    // tree layout type
                    yield this.handleTreeLayout(type, Object.assign({ nodeSize,
                        width,
                        height,
                        center }, rest), animationEffectTiming, graphCore, layoutData, animate);
                    return;
                }
                // Initialize layout.
                const layout = new layoutCtor(Object.assign({ nodeSize,
                    width,
                    height,
                    center }, rest));
                this.currentLayout = layout;
                // CustomLayout is not workerized.
                if (!isLayoutWorkerized(options)) {
                    workerEnabled = false;
                    // TODO: console.warn();
                }
                if (workerEnabled) {
                    /**
                     * Run algorithm in WebWorker, `animated` option will be ignored.
                     */
                    const supervisor = new Supervisor(layoutGraphCore, layout, {
                        iterations,
                    });
                    this.currentSupervisor = supervisor;
                    positions = yield supervisor.execute();
                }
                else {
                    // e.g. Force layout
                    if (isLayoutWithIterations(layout)) {
                        if (animated) {
                            /**
                             * `onTick` will get triggered in this case.
                             */
                            positions = yield layout.execute(layoutGraphCore, {
                                onTick: (positionsOnTick) => {
                                    // Display the animated process of layout.
                                    this.updateNodesPosition(positionsOnTick);
                                    this.graph.emit('tick', positionsOnTick);
                                },
                            });
                        }
                        else {
                            /**
                             * Manually step simulation in a sync way. `onTick` won't get triggered in this case,
                             * there will be no animation either.
                             */
                            layout.execute(layoutGraphCore);
                            layout.stop();
                            positions = layout.tick(iterations);
                        }
                    }
                    else {
                        positions = yield layout.execute(layoutGraphCore);
                        if (animated) {
                            yield this.animateLayoutWithoutIterations(positions, animationEffectTiming);
                        }
                    }
                }
            }
            this.graph.emit('endlayout');
            // Update nodes' positions.
            this.updateNodesPosition(positions, animate);
        });
    }
    handleTreeLayout(type, options, animationEffectTiming, graphCore, layoutData, animate) {
        return __awaiter(this, void 0, void 0, function* () {
            const { animated = false, rootIds = [], begin = [0, 0], radial } = options;
            const nodePositions = [];
            const nodeMap = new Map();
            // tree layout with tree data
            const trees = graphCore
                .getRoots('tree')
                .filter((node) => !node.data._isCombo)
                .map((node) => ({ id: node.id, children: [] }));
            trees.forEach((tree) => {
                nodeMap.set(tree.id, tree);
                graphCore.dfsTree(tree.id, (node) => {
                    nodeMap.get(node.id).children = graphCore
                        .getChildren(node.id, 'tree')
                        .filter((node) => !node.data._isCombo)
                        .map((child) => {
                        nodeMap.set(child.id, { id: child.id, children: [] });
                        return nodeMap.get(child.id);
                    });
                }, 'tree');
                layoutOneTree(tree, type, options, nodeMap, nodePositions, begin);
                if (radial) {
                    nodePositions.forEach((pos) => nodeMap.set(pos.id, pos));
                    radialLayout(tree, nodeMap);
                }
            });
            if (animated) {
                yield this.animateLayoutWithoutIterations({ nodes: nodePositions, edges: [] }, animationEffectTiming);
            }
            this.graph.emit('endlayout');
            this.updateNodesPosition({ nodes: nodePositions, edges: [] }, animated || animate);
            return;
        });
    }
    stopLayout() {
        if (this.currentLayout && isLayoutWithIterations(this.currentLayout)) {
            this.currentLayout.stop();
            this.currentLayout = null;
        }
        if (this.currentSupervisor) {
            this.currentSupervisor.stop();
            this.currentSupervisor = null;
        }
        if (this.currentAnimation) {
            this.currentAnimation.finish();
            this.currentAnimation = null;
        }
    }
    getCurrentAnimation() {
        return this.currentAnimation;
    }
    onDestroy() {
        this.stopLayout();
        if (this.currentSupervisor) {
            this.currentSupervisor.kill();
        }
    }
    updateNodesPosition(positions, animate = true) {
        const { nodes, edges } = positions;
        this.graph.updateNodePosition(nodes, undefined, !animate);
        const edgeToUpdate = edges
            .filter((edge) => edge.data.controlPoints)
            .map((edge) => ({
            id: edge.id,
            data: {
                keyShape: {
                    controlPoints: edge.data.controlPoints,
                },
            },
        }));
        if (edgeToUpdate.length) {
            this.graph.updateData('edge', edgeToUpdate);
        }
    }
    /**
     * For those layout without iterations, e.g. circular, random, since they don't have `onTick` callback,
     * we have to translate each node from its initial position to final one after layout,
     * with the help of `onframe` from G's animation API.
     * @param positions
     * @param animationEffectTiming
     */
    animateLayoutWithoutIterations(positions, animationEffectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            // Animation should be executed only once.
            animationEffectTiming.iterations = 1;
            const initialPositions = positions.nodes.map((node) => {
                // Should clone each node's initial data since it will be updated during the layout process.
                return Object.assign({}, this.graph.getNodeData(`${node.id}`));
            });
            // Add a connected displayobject so that we can animate it.
            if (!this.animatedDisplayObject.isConnected) {
                yield this.graph.canvas.ready;
                this.graph.canvas.appendChild(this.animatedDisplayObject);
            }
            // Use `opacity` since it is an interpolated property.
            this.currentAnimation = this.animatedDisplayObject.animate([
                {
                    opacity: 0,
                },
                {
                    opacity: 1,
                },
            ], animationEffectTiming);
            // Update each node's position on each frame.
            // @see https://g.antv.antgroup.com/api/animation/waapi#%E5%B1%9E%E6%80%A7
            this.currentAnimation.onframe = (e) => {
                // @see https://g.antv.antgroup.com/api/animation/waapi#progress
                const progress = e.target.effect.getComputedTiming()
                    .progress;
                const interpolatedNodesPosition = initialPositions.map(({ id, data }, i) => {
                    const { x: fromX = 0, y: fromY = 0 } = data;
                    const { x: toX, y: toY } = positions.nodes[i].data;
                    return {
                        id,
                        data: {
                            x: fromX + (toX - fromX) * progress,
                            y: fromY + (toY - fromY) * progress,
                        },
                    };
                });
                this.updateNodesPosition({
                    nodes: interpolatedNodesPosition,
                    edges: [],
                });
            };
            yield this.currentAnimation.finished;
        });
    }
}
//# sourceMappingURL=layout.js.map