import { uniqueId } from '@antv/util';
import registry from '../../stdlib';
import { getExtension } from '../../util/extension';
import { Plugin as PluginBase } from '../../types/plugin';
/**
 * Wraps the listener with error logging.
 * @returns a new listener with error logging.
 */
const wrapListener = (type, eventName, listener) => {
    return (event) => {
        try {
            listener(event);
        }
        catch (error) {
            console.error(`G6: Error occurred in "${eventName}" phase of the plugin "${type}"!`);
            throw error;
        }
    };
};
/**
 * Manages free plugin extensions and graph layout.
 */
export class PluginController {
    constructor(graph) {
        this.extensions = [];
        /**
         * Plugins on graph.
         * @example
         * { 'minimap': Minimap, 'tooltip': Tooltip }
         */
        this.pluginMap = new Map();
        /**
         * Listeners added by all current plugins.
         * @example
         * {
         *   'minimap': { 'afterlayout': function },
         * }
         */
        this.listenersMap = {};
        this.addListeners = (key, plugin) => {
            const events = plugin.getEvents();
            this.listenersMap[key] = {};
            Object.keys(events).forEach((eventName) => {
                // Wrap the listener with error logging.
                const listener = wrapListener(key, eventName, events[eventName].bind(plugin));
                this.graph.on(eventName, listener);
                this.listenersMap[key][eventName] = listener;
            });
        };
        this.removeListeners = (key) => {
            const listeners = this.listenersMap[key];
            Object.keys(listeners).forEach((eventName) => {
                const listener = listeners[eventName];
                if (listener) {
                    this.graph.off(eventName, listener);
                }
            });
        };
        this.graph = graph;
        this.tap();
    }
    /**
     * Subscribe the lifecycle of graph.
     */
    tap() {
        this.graph.hooks.init.tap(this.onPluginInit.bind(this));
        this.graph.hooks.pluginchange.tap(this.onPluginChange.bind(this));
        this.graph.hooks.destroy.tap(this.onDestroy.bind(this));
    }
    onPluginInit() {
        // 1. Initialize new behaviors.
        this.pluginMap.clear();
        const { graph } = this;
        const pluginConfigs = graph.getSpecification().plugins || [];
        pluginConfigs.forEach(this.initPlugin.bind(this));
        // 2. Add listeners for each behavior.
        this.listenersMap = {};
        this.pluginMap.forEach((item, key) => {
            const { plugin } = item;
            this.addListeners(key, plugin);
        });
    }
    initPlugin(config) {
        var _a, _b;
        const { graph } = this;
        if (config instanceof PluginBase) {
            config.init(graph);
            const key = config.key || ((_a = config.options) === null || _a === void 0 ? void 0 : _a.key) || `plugin-${uniqueId()}`;
            this.pluginMap.set(key, { type: key, plugin: config });
            return { key, plugin: config };
        }
        const Plugin = getExtension(config, registry.useLib, 'plugin');
        const options = typeof config === 'string' ? {} : config;
        const type = typeof config === 'string' ? config : config.type;
        const key = typeof config === 'string'
            ? config
            : config.key || ((_b = config.options) === null || _b === void 0 ? void 0 : _b.key) || type;
        if (!Plugin) {
            throw new Error(`Plugin ${type} not found, please make sure you have registered it first`);
        }
        const plugin = new Plugin(Object.assign(Object.assign({}, options), { key }));
        plugin.init(graph);
        this.pluginMap.set(key, { type, plugin });
        return { key, type, plugin };
    }
    onPluginChange(params) {
        const { action, plugins: pluginCfgs } = params;
        if (action === 'add') {
            pluginCfgs.forEach((config) => {
                const { key, plugin } = this.initPlugin(config);
                this.addListeners(key, plugin);
            });
            return;
        }
        if (action === 'remove') {
            pluginCfgs.forEach((config) => {
                var _a;
                const key = (typeof config === 'string' ? config : config.key) ||
                    ((_a = config.options) === null || _a === void 0 ? void 0 : _a.key) ||
                    config.type;
                const item = this.pluginMap.get(key);
                if (!item)
                    return;
                const { plugin } = item;
                this.removeListeners(key);
                plugin.destroy();
                this.pluginMap.delete(key);
            });
            return;
        }
        if (action === 'update') {
            pluginCfgs.forEach((config) => {
                if (typeof config === 'string')
                    return;
                const key = config.key || config.type;
                const item = this.pluginMap.get(key);
                if (!item)
                    return;
                const { plugin } = item;
                plugin.updateCfgs(config);
                this.removeListeners(key);
                this.addListeners(key, plugin);
            });
            return;
        }
    }
    /**
     * Check if a plugin with the specified plugin key exists.
     * @param {string} pluginKey The key of the plugin to check.
     */
    hasPlugin(pluginKey) {
        return this.pluginMap.has(pluginKey);
    }
    /**
     * Retrieve the plugin with the specified plugin key.
     * @param {string} pluginKey The key of the plugin to check.
     */
    getPlugin(pluginKey) {
        const { plugin } = this.pluginMap.get(pluginKey);
        if (!plugin) {
            throw new Error('Plugin not found for key: ' + pluginKey);
        }
        return plugin;
    }
    onDestroy() {
        this.pluginMap.forEach((item) => {
            const { plugin } = item;
            plugin.destroy();
        });
    }
    destroy() { }
}
//# sourceMappingURL=plugin.js.map