import { Behavior } from '../../types/behavior';
const KEYBOARD_TRIGGERS = ['shift', 'ctrl', 'alt', 'meta'];
const MOUSE_TRIGGERS = ['mouseenter', 'click'];
const compare = (prev, curr) => {
    const inactive = prev.filter((v) => !curr.includes(v));
    const active = curr.filter((v) => !prev.includes(v));
    return {
        active,
        inactive,
    };
};
const DEFAULT_OPTIONS = {
    multiple: true,
    shouldBegin: () => true,
    shouldUpdate: () => true,
    trigger: 'click',
    activeState: 'selected',
};
export class ActivateRelations extends Behavior {
    constructor(options) {
        super(Object.assign({}, DEFAULT_OPTIONS, options));
        this.getEvents = () => {
            const { trigger } = this.options;
            if (trigger === 'mouseenter') {
                return {
                    'node:mouseenter': this.setAllItemStates,
                    'combo:mouseenter': this.setAllItemStates,
                    'node:mouseleave': this.clearActiveState,
                    'combo:mouseleave': this.clearActiveState,
                };
            }
            return {
                'node:click': this.setAllItemStates,
                'combo:click': this.setAllItemStates,
                'canvas:click': this.clearActiveState,
            };
        };
        // Validate options
        if (options.trigger && !MOUSE_TRIGGERS.includes(options.trigger)) {
            console.warn(`G6: Invalid trigger option "${options.trigger}" for activate-relations behavior!`);
            this.options.trigger = DEFAULT_OPTIONS.trigger;
        }
        this.prevEdgeIds = [];
        this.prevNodeIds = [];
    }
    setAllItemStates(e) {
        const { itemId } = e;
        const { graph } = this;
        const { activeState: ACTIVE_STATE } = this.options;
        if (!graph || graph.destroyed)
            return;
        if (!this.options.shouldBegin(e))
            return;
        const ids = graph
            .getNeighborNodesData(itemId, 'both')
            .map((item) => item.id);
        const edgeIds = graph
            .getRelatedEdgesData(itemId, 'both')
            .map((item) => item.id);
        const nodeIds = [itemId, ...ids];
        /** 数据对比，处理得到最小改动的高亮和非高亮的数据 */
        const { active: activeNodeIds, inactive: inactiveNodeIds } = compare(this.prevNodeIds, nodeIds);
        const { active: activeEdgeIds, inactive: inactiveEdgeIds } = compare(this.prevEdgeIds, edgeIds);
        graph.batch(() => {
            /** 节点 */
            graph.setItemState(activeNodeIds, ACTIVE_STATE, true);
            graph.setItemState(inactiveNodeIds, ACTIVE_STATE, false);
            /** 边 */
            graph.setItemState(activeEdgeIds, ACTIVE_STATE, true);
            graph.setItemState(inactiveEdgeIds, ACTIVE_STATE, false);
        });
        this.prevNodeIds = nodeIds;
        this.prevEdgeIds = edgeIds;
    }
    clearActiveState(e) {
        const { activeState: ACTIVE_STATE } = this.options;
        this.graph.batch(() => {
            this.graph.setItemState(this.prevNodeIds, ACTIVE_STATE, false);
            this.graph.setItemState(this.prevEdgeIds, ACTIVE_STATE, false);
        });
        this.prevNodeIds = [];
        this.prevEdgeIds = [];
    }
}
//# sourceMappingURL=activate-relations.js.map