import { Behavior } from '../../types/behavior';
const ALLOWED_TRIGGERS = ['shift', 'ctrl', 'alt', 'meta'];
const DEFAULT_OPTIONS = {
    multiple: true,
    trigger: 'shift',
    selectedState: 'selected',
    itemTypes: ['node', 'edge'],
    eventName: '',
    shouldBegin: () => true,
    shouldUpdate: () => true,
};
export class ClickSelect extends Behavior {
    constructor(options) {
        super(Object.assign({}, DEFAULT_OPTIONS, options));
        /**
         * Cache the ids of items selected by this behavior
         */
        this.selectedIds = [];
        /**
         * Two flag to avoid onCanvasClick triggered while dragging canvas
         */
        this.canvasPointerDown = undefined;
        this.canvasPointerMove = false;
        this.timeout = undefined;
        this.getEvents = () => {
            return {
                'node:click': this.onClick,
                'edge:click': this.onClick,
                'combo:click': this.onClickBesideDblClick,
                'canvas:pointerdown': this.onCanvasPointerDown,
                'canvas:pointermove': this.onCanvasPointerMove,
                'canvas:pointerup': this.onCanvasPointerUp,
            };
        };
        // Validate options
        if (options.trigger && !ALLOWED_TRIGGERS.includes(options.trigger)) {
            console.warn(`G6: Invalid trigger option "${options.trigger}" for click-select behavior!`);
            this.options.trigger = DEFAULT_OPTIONS.trigger;
        }
    }
    isMultipleSelect(event) {
        if (!this.options.multiple)
            return false;
        const keyMap = {
            shift: event.shiftKey,
            ctrl: event.ctrlKey,
            alt: event.altKey,
            meta: event.metaKey,
        };
        return keyMap[this.options.trigger];
    }
    onClickBesideDblClick(event) {
        if (this.timeout) {
            clearTimeout(this.timeout);
            this.timeout = undefined;
            return;
        }
        this.timeout = setTimeout(() => {
            this.timeout = undefined;
            this.onClick(event);
        }, 200);
    }
    onClick(event) {
        if (!this.options.shouldBegin(event))
            return;
        // Will not be 'canvas' because this method is listened on node and edge click.
        const itemType = event.itemType;
        const itemId = event.itemId;
        const selectable = this.options.itemTypes.includes(itemType);
        if (!selectable)
            return;
        const state = this.options.selectedState;
        const multiple = this.isMultipleSelect(event);
        // FIXME: should use graph.getItemState() instead
        const isSelectAction = !this.graph
            .findIdByState(itemType, state)
            .includes(itemId);
        const action = isSelectAction
            ? 'select'
            : 'unselect';
        // Select/Unselect item.
        if (this.options.shouldUpdate(event)) {
            this.graph.batch(() => {
                if (!multiple) {
                    // Not multiple, clear all currently selected items
                    this.graph.setItemState(this.selectedIds, state, false);
                }
                this.graph.setItemState(itemId, state, isSelectAction);
            });
            if (isSelectAction) {
                this.selectedIds.push(itemId);
            }
            else {
                this.selectedIds.filter((id) => id !== itemId);
            }
        }
        // Emit an event.
        if (this.options.eventName) {
            const selectedNodeIds = this.graph.findIdByState('node', state);
            const selectedEdgeIds = this.graph.findIdByState('edge', state);
            const selectedComboIds = this.graph.findIdByState('combo', state);
            this.graph.emit(this.options.eventName, {
                action,
                itemId,
                itemType,
                selectedNodeIds,
                selectedEdgeIds,
                selectedComboIds,
            });
        }
    }
    onCanvasPointerDown(event) {
        this.canvasPointerDown = {
            x: event.canvas.x,
            y: event.canvas.y,
        };
        this.canvasPointerMove = false;
    }
    onCanvasPointerMove(event) {
        if (this.canvasPointerDown) {
            const deltaX = Math.abs(this.canvasPointerDown.x - event.canvas.x);
            const deltaY = Math.abs(this.canvasPointerDown.y - event.canvas.y);
            if (deltaX > 1 || deltaY > 1)
                this.canvasPointerMove = true;
        }
    }
    onCanvasPointerUp(event) {
        if (this.canvasPointerDown && !this.canvasPointerMove)
            this.onCanvasClick(event);
        this.canvasPointerDown = undefined;
        this.canvasPointerMove = false;
    }
    onCanvasClick(event) {
        if (!this.options.shouldBegin(event))
            return;
        // Find current selected items.
        const state = this.options.selectedState;
        if (!this.selectedIds.length)
            return;
        // Unselect all items.
        if (this.options.shouldUpdate(event)) {
            this.graph.setItemState(this.selectedIds, state, false);
        }
        // Emit an event.
        if (this.options.eventName) {
            this.graph.emit(this.options.eventName, {
                action: 'unselect',
                itemId: '',
                itemType: '',
                selectedNodeIds: [],
                selectedEdgeIds: [],
                selectedComboIds: [],
            });
        }
    }
}
//# sourceMappingURL=click-select.js.map