import { Behavior } from '../../types/behavior';
const VALID_TRIGGERS = ['drag', 'directionKeys'];
const DEFAULT_OPTIONS = {
    enableOptimize: false,
    dragOnItems: false,
    trigger: 'drag',
    direction: 'both',
    secondaryKey: '',
    secondaryKeyToDisable: 'shift',
    speedUpKey: '',
    scalableRange: 0,
    eventName: '',
    shouldBegin: () => true,
};
export class DragCanvas extends Behavior {
    constructor(options) {
        const finalOptions = Object.assign({}, DEFAULT_OPTIONS, options);
        if (!VALID_TRIGGERS.includes(finalOptions.trigger)) {
            console.warn(`The trigger ${finalOptions.trigger} is not valid, 'drag' will take effect.`);
            finalOptions.trigger = 'drag';
        }
        super(finalOptions);
    }
    getEvents() {
        if (typeof window !== 'undefined') {
            window.addEventListener('keydown', this.onKeydown.bind(this));
            window.addEventListener('keyup', this.onKeyup.bind(this));
        }
        if (this.options.trigger === 'directionKeys') {
            return {
                keydown: this.onKeydown,
                keyup: this.onKeyup,
            };
        }
        return {
            pointerdown: this.onPointerDown,
            pointermove: this.onPointerMove,
            pointerup: this.onPointerUp,
            keydown: this.onKeydown,
            keyup: this.onKeyup,
        };
    }
    onPointerDown(event) {
        const { secondaryKey, secondaryKeyToDisable, dragOnItems, shouldBegin } = this.options;
        // disabled key is pressing
        if (secondaryKeyToDisable && this.disableKeydown)
            return;
        // assistant key is not pressing
        if (secondaryKey && !this.keydown)
            return;
        // should not begin
        if (!shouldBegin(event))
            return;
        // begin on an item
        const isDragOnItem = event.itemId && event.itemId !== 'CANVAS';
        if (isDragOnItem && !dragOnItems)
            return;
        const { client } = event;
        this.pointerDownAt = { x: client.x, y: client.y };
        document.addEventListener('mouseup', (evt) => {
            this.onPointerUp();
        }, { once: true });
    }
    hideShapes() {
        const { graph } = this;
        if (this.options.enableOptimize) {
            this.hiddenEdgeIds = graph
                .getAllEdgesData()
                .map((edge) => edge.id)
                .filter((id) => graph.getItemVisible(id) === true);
            graph.executeWithoutStacking(() => {
                graph.hideItem(this.hiddenEdgeIds, true);
                this.hiddenNodeIds = graph
                    .getAllNodesData()
                    .map((node) => node.id)
                    .filter((id) => graph.getItemVisible(id) === true);
                // draw node's keyShapes on transient, and then hidden the real nodes;
                // this.hiddenNodeIds.forEach((id) => {
                //   graph.drawTransient('node', id, {
                //     onlyDrawKeyShape: true,
                //   });
                // });
                // graph.hideItem(this.hiddenNodeIds, true);
            });
        }
    }
    formatDisplacement(diffX, diffY) {
        const { graph } = this;
        const { scalableRange, direction } = this.options;
        const [width, height] = graph.getSize();
        const graphBBox = graph.canvas.getRoot().getRenderBounds();
        let rangeNum = Number(scalableRange);
        let isPixel;
        if (typeof scalableRange === 'string') {
            if (scalableRange.includes('px')) {
                isPixel = scalableRange.includes('px');
                rangeNum = Number(scalableRange.replace('px', ''));
            }
            if (scalableRange.includes('%')) {
                rangeNum = Number(scalableRange.replace('%', '')) / 100;
            }
        }
        if (rangeNum === 0)
            return { dx: diffX, dy: diffY };
        let expandWidth = rangeNum;
        let expandHeight = rangeNum;
        // If it is not a string with 'px', regard as ratio
        if (!isPixel) {
            expandWidth = width * rangeNum;
            expandHeight = height * rangeNum;
        }
        const leftTopClient = graph.getViewportByCanvas({
            x: graphBBox.min[0],
            y: graphBBox.min[1],
        });
        const rightBottomClient = graph.getViewportByCanvas({
            x: graphBBox.max[0],
            y: graphBBox.max[1],
        });
        let dx = diffX;
        let dy = diffY;
        if (direction === 'y' ||
            (diffX > 0 && rightBottomClient.x + diffX > width + expandWidth) ||
            (diffX < 0 && leftTopClient.x + expandWidth + diffX < 0)) {
            dx = 0;
        }
        if (direction === 'x' ||
            (diffY > 0 && rightBottomClient.y + diffY > height + expandHeight) ||
            (diffY < 0 && leftTopClient.y + expandHeight + diffY < 0)) {
            dy = 0;
        }
        return { dx, dy };
    }
    onPointerMove(event) {
        if (!this.pointerDownAt)
            return;
        const { eventName, direction, secondaryKeyToDisable } = this.options;
        // disabled key is pressing
        if (secondaryKeyToDisable && this.disableKeydown)
            return;
        const { graph } = this;
        const { client } = event;
        const diffX = client.x - this.pointerDownAt.x;
        const diffY = client.y - this.pointerDownAt.y;
        if (direction === 'x' && !diffX)
            return;
        if (direction === 'y' && !diffY)
            return;
        if (direction === 'both' && !diffX && !diffY)
            return;
        // begin dragging
        if (!this.dragging) {
            this.hideShapes();
            this.dragging = true;
        }
        const { dx, dy } = this.formatDisplacement(diffX, diffY);
        graph.translate({ dx, dy });
        this.pointerDownAt = { x: client.x, y: client.y };
        // Emit event.
        if (eventName) {
            this.graph.emit(eventName, { translate: { dx, dy } });
        }
    }
    onPointerUp() {
        this.pointerDownAt = undefined;
        this.dragging = false;
        const { graph } = this;
        if (this.options.enableOptimize) {
            graph.startBatch();
            if (this.hiddenEdgeIds) {
                graph.showItem(this.hiddenEdgeIds, true);
            }
            // if (this.hiddenNodeIds) {
            //   this.hiddenNodeIds.forEach((id) => {
            //     this.graph.drawTransient('node', id, { action: 'remove' });
            //   });
            //   graph.showItem(this.hiddenNodeIds, true);
            // }
            graph.stopBatch();
        }
    }
    onKeydown(event) {
        const { key } = event;
        const { secondaryKey, secondaryKeyToDisable, trigger, speedUpKey, eventName, shouldBegin, } = this.options;
        if (secondaryKey && secondaryKey === key.toLowerCase()) {
            this.keydown = true;
        }
        if (speedUpKey && speedUpKey === key.toLowerCase()) {
            this.speedupKeydown = true;
        }
        if (secondaryKeyToDisable && secondaryKeyToDisable === key.toLowerCase()) {
            this.disableKeydown = true;
        }
        if (trigger === 'directionKeys') {
            if (secondaryKey && !this.keydown)
                return;
            if (secondaryKeyToDisable && this.disableKeydown)
                return;
            if (!shouldBegin(event))
                return;
            const { graph, speedupKeydown } = this;
            const speed = speedupKeydown ? 20 : 1;
            let dx = 0;
            let dy = 0;
            switch (key) {
                case 'ArrowLeft':
                    dx = -speed;
                    break;
                case 'ArrowRight':
                    dx = speed;
                    break;
                case 'ArrowUp':
                    dy = -speed;
                    break;
                case 'ArrowDown':
                    dy = speed;
                    break;
            }
            if (dx || dy) {
                const { dx: formattedDx, dy: formattedDy } = this.formatDisplacement(dx, dy);
                graph.translate({ dx: formattedDx, dy: formattedDy });
                if (eventName) {
                    this.graph.emit(eventName, {
                        translate: { dx: formattedDx, dy: formattedDy },
                    });
                }
            }
        }
    }
    onKeyup(event) {
        const { key } = event;
        const { secondaryKey, secondaryKeyToDisable, speedUpKey } = this.options;
        if (secondaryKey && secondaryKey === key.toLowerCase()) {
            this.keydown = false;
        }
        if (speedUpKey && speedUpKey === key.toLowerCase()) {
            this.speedupKeydown = false;
        }
        if (secondaryKeyToDisable && secondaryKeyToDisable === key.toLowerCase()) {
            this.disableKeydown = false;
        }
    }
    destroy() {
        if (typeof window !== 'undefined') {
            window.removeEventListener('keydown', this.onKeydown.bind(this));
            window.removeEventListener('keyup', this.onKeyup.bind(this));
        }
    }
}
//# sourceMappingURL=drag-canvas.js.map