import { AABB, DisplayObject } from '@antv/g';
import { ID } from 'types';
import { Point } from '../../../types/common';
import { ArrowStyle, EdgeDisplayModel, EdgeModelData, EdgeShapeMap, EdgeShapeStyles } from '../../../types/edge';
import { SHAPE_TYPE, ShapeStyle, State, LodStrategyObj } from '../../../types/item';
import Node from '../../../item/node';
export declare abstract class BaseEdge {
    type: string;
    defaultStyles: EdgeShapeStyles;
    themeStyles: EdgeShapeStyles;
    mergedStyles: EdgeShapeStyles;
    sourcePoint: Point;
    targetPoint: Point;
    nodeMap: Map<ID, Node>;
    lodStrategy?: LodStrategyObj;
    labelPosition: {
        x: number;
        y: number;
        transform: string;
        isRevert: boolean;
    };
    boundsCache: {
        labelShapeGeometry?: AABB;
        labelBackgroundShapeGeometry?: AABB;
        labelShapeTransform?: string;
        labelBackgroundShapeTransform?: string;
    };
    private zoomCache;
    constructor(props: any);
    /**
     * Get merged styles from `getMergedStyles` and assigns the merged styles to the 'mergedStyles' property.
     * @param model - The EdgeDisplayModel to merge the styles from.
     */
    mergeStyles(model: EdgeDisplayModel): void;
    /**
     * Merge style
     * @param model - The EdgeDisplayModel to retrieve the merged styles from.
     * @returns The merged styles as a EdgeShapeStyles object.
     */
    getMergedStyles(model: EdgeDisplayModel): EdgeShapeStyles;
    /**
     * Call it after calling draw function to update cache about bounds and zoom levels.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     */
    updateCache(shapeMap: any): void;
    /**
     * Draw all elements related to the edge.
     * You should call `drawKeyShape` and `drawAnchorShape`,`drawLabelShape`,`drawIconShape`...as you like.
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns An object containing the keyShape and optional labelShape, iconShape, and some otherShapes properties
     */
    abstract draw(model: EdgeDisplayModel, sourcePoint: Point, targetPoint: Point, shapeMap: {
        [shapeId: string]: DisplayObject;
    }, diffData?: {
        previous: EdgeModelData;
        current: EdgeModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): {
        keyShape: DisplayObject;
        labelShape?: DisplayObject;
        iconShape?: DisplayObject;
        [otherShapeId: string]: DisplayObject;
    };
    /**
     * Perform additional drawing operations or add custom shapes after drawing edge.
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param shapesChanged An array of shape IDs that have changed and need to be updated.
     * @returns An object that contains some new shapes to be added to the edge.
     */
    afterDraw(model: EdgeDisplayModel, shapeMap: {
        [shapeId: string]: DisplayObject;
    }, shapesChanged?: string[]): {
        [otherShapeId: string]: DisplayObject;
    };
    /**
     * Set the state for the edge.
     * @param state state name
     * @param value state value
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     */
    setState: (name: string, value: boolean, shapeMap: {
        [shapeId: string]: DisplayObject;
    }) => void;
    /**
     * Draw the label shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the label shape of the edge.
     */
    drawLabelShape(model: EdgeDisplayModel, shapeMap: EdgeShapeMap, diffData?: {
        previous: EdgeModelData;
        current: EdgeModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the label background shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the label background shape of the edge.
     */
    drawLabelBackgroundShape(model: EdgeDisplayModel, shapeMap: EdgeShapeMap, diffData?: {
        previous: EdgeModelData;
        current: EdgeModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the icon shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the icon shape of the edge.
     */
    drawIconShape(model: EdgeDisplayModel, shapeMap: EdgeShapeMap, diffData?: {
        previous: EdgeModelData;
        current: EdgeModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the halo shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the halo shape of the edge.
     */
    drawHaloShape(model: EdgeDisplayModel, shapeMap: EdgeShapeMap, diffData?: {
        previous: EdgeModelData;
        current: EdgeModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * The listener for graph zooming.
     * 1. show / hide some shapes while zoom level changed;
     * 2. change the shapes' sizes to make them have same visual size while zooming, e.g. labelShape, labelBackgroundShape.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param zoom The zoom level of the graph.
     */
    onZoom: (shapeMap: EdgeShapeMap, zoom: number) => void;
    /**
     * Update the shapes' sizes e.g. labelShape, labelBackgroundShape, to keep the visual size while zooming.
     * @param shapeMap
     * @param zoom
     * @returns
     */
    private balanceShapeSize;
    /**
     * Update the source point { x, y } for the edge. Called in item's draw func.
     * @param point source point
     */
    setSourcePoint(point: Point): void;
    /**
     * Update the target point { x, y } for the edge. Called in item's draw func.
     * @param point target point
     */
    setTargetPoint(point: Point): void;
    /**
     * Update all visible nodes on the canvas to be aware of obstacles. Called in item's draw func.
     * @param nodeMap The Map object representing the node map, where keys are node IDs and values are nodes.
     */
    setNodeMap(nodeMap: Map<ID, Node>): void;
    /**
     * Adds or updates an arrow marker on the specified position of an edge.
     * @param position - The position where the arrow marker should be added or updated. Can be either 'start' or 'end'.
     * @param arrowConfig - The configuration for the arrow marker. Can be a boolean indicating whether to use the default arrow configuration, or an ArrowStyle object with custom arrow properties.
     * @param bodyStyle - The style of the edge body.
     * @param model - The EdgeDisplayModel that contains the data and style information for the edge.
     * @param resultStyle - The style object where the arrow marker properties will be added or updated.
     */
    upsertArrow(position: 'start' | 'end', arrowConfig: boolean | ArrowStyle, bodyStyle: ShapeStyle, model: EdgeDisplayModel, resultStyle: ShapeStyle): void;
    /**
     * Create (if does not exit in shapeMap) or update the shape according to the configurations.
     * @param type shape's type
     * @param id unique string to indicates the shape
     * @param style style to be updated
     * @param shapeMap the shape map of a edge
     * @param model data model of the edge
     * @returns The display object representing the shape.
     */
    upsertShape(type: SHAPE_TYPE, id: string, style: ShapeStyle, shapeMap: {
        [shapeId: string]: DisplayObject;
    }, model: EdgeDisplayModel): DisplayObject;
}
