var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { isNumber, isBoolean } from '@antv/util';
import { DEFAULT_LABEL_BG_PADDING, OTHER_SHAPES_FIELD_NAME, RESERVED_SHAPE_IDS, } from '../../../constant';
import { LOCAL_BOUNDS_DIRTY_FLAG_KEY, formatPadding, mergeStyles, upsertShape, } from '../../../util/shape';
import { DEFAULT_ANIMATE_CFG, fadeIn, fadeOut } from '../../../util/animate';
import { getWordWrapWidthByEnds } from '../../../util/text';
import { getZoomLevel } from '../../../util/zoom';
import { DEFAULT_ARROW_CONFIG, getArrowPath } from '../../../util/arrow';
export class BaseEdge {
    constructor(props) {
        this.defaultStyles = {};
        // cache the zoom level infomations
        this.zoomCache = {
            hiddenShape: {},
            balanceRatio: 1,
            zoom: 1,
            zoomLevel: 0,
            levelShapes: {},
            wordWrapWidth: 50,
            animateConfig: DEFAULT_ANIMATE_CFG.zoom,
            firstRender: true,
        };
        /**
         * The listener for graph zooming.
         * 1. show / hide some shapes while zoom level changed;
         * 2. change the shapes' sizes to make them have same visual size while zooming, e.g. labelShape, labelBackgroundShape.
         * @param shapeMap The shape map that contains all of the elements to show on the edge.
         * @param zoom The zoom level of the graph.
         */
        this.onZoom = (shapeMap, zoom) => {
            var _a, _b, _c;
            // balance the size for label, badges
            this.balanceShapeSize(shapeMap, zoom);
            // zoomLevel changed
            if (!this.lodStrategy)
                return;
            const { levels } = this.lodStrategy;
            const { levelShapes, hiddenShape, animateConfig, firstRender = true, zoomLevel: previousLevel, } = this.zoomCache;
            // last zoom ratio responsed by zoom changing, which might not equal to zoom.previous in props since the function is debounced.
            const currentLevel = getZoomLevel(levels, zoom);
            const levelNums = Object.keys(levelShapes).map(Number);
            const maxLevel = Math.max(...levelNums);
            const minLevel = Math.min(...levelNums);
            if (currentLevel < previousLevel) {
                // zoomLevel changed, from higher to lower, hide something
                if (firstRender) {
                    for (let i = currentLevel + 1; i <= maxLevel; i++) {
                        (_a = levelShapes[String(i)]) === null || _a === void 0 ? void 0 : _a.forEach((id) => { var _a; return (_a = shapeMap[id]) === null || _a === void 0 ? void 0 : _a.hide(); });
                    }
                }
                else {
                    for (let i = currentLevel + 1; i <= maxLevel; i++) {
                        (_b = levelShapes[String(i)]) === null || _b === void 0 ? void 0 : _b.forEach((id) => fadeOut(id, shapeMap[id], hiddenShape, animateConfig));
                    }
                }
            }
            else if (currentLevel > previousLevel) {
                // zoomLevel changed, from lower to higher, show something
                for (let i = currentLevel; i >= minLevel; i--) {
                    (_c = levelShapes[String(i)]) === null || _c === void 0 ? void 0 : _c.forEach((id) => fadeIn(id, shapeMap[id], this.mergedStyles[id] ||
                        this.mergedStyles[id.replace('Background', '')], hiddenShape, animateConfig));
                }
            }
            this.zoomCache.zoom = zoom;
            this.zoomCache.zoomLevel = currentLevel;
            this.zoomCache.firstRender = false;
        };
        const { themeStyles, lodStrategy, zoom } = props;
        if (themeStyles)
            this.themeStyles = themeStyles;
        this.lodStrategy = lodStrategy;
        this.boundsCache = {};
        this.zoomCache.zoom = zoom;
        this.zoomCache.balanceRatio = 1 / zoom;
        this.zoomCache.animateConfig = Object.assign(Object.assign({}, DEFAULT_ANIMATE_CFG.zoom), lodStrategy === null || lodStrategy === void 0 ? void 0 : lodStrategy.animateCfg);
    }
    /**
     * Get merged styles from `getMergedStyles` and assigns the merged styles to the 'mergedStyles' property.
     * @param model - The EdgeDisplayModel to merge the styles from.
     */
    mergeStyles(model) {
        this.mergedStyles = this.getMergedStyles(model);
    }
    /**
     * Merge style
     * @param model - The EdgeDisplayModel to retrieve the merged styles from.
     * @returns The merged styles as a EdgeShapeStyles object.
     */
    getMergedStyles(model) {
        var _a;
        const { data } = model;
        const dataStyles = {};
        Object.keys(data).forEach((fieldName) => {
            if (RESERVED_SHAPE_IDS.includes(fieldName))
                dataStyles[fieldName] = data[fieldName];
            else if (fieldName === OTHER_SHAPES_FIELD_NAME) {
                Object.keys(data[fieldName]).forEach((otherShapeId) => (dataStyles[otherShapeId] = data[fieldName][otherShapeId]));
            }
        });
        const merged = mergeStyles([
            this.themeStyles,
            this.defaultStyles,
            dataStyles,
        ]);
        const padding = (_a = merged.labelBackgroundShape) === null || _a === void 0 ? void 0 : _a.padding;
        if (padding) {
            merged.labelBackgroundShape.padding = formatPadding(padding, DEFAULT_LABEL_BG_PADDING);
        }
        return merged;
    }
    /**
     * Call it after calling draw function to update cache about bounds and zoom levels.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     */
    updateCache(shapeMap) {
        ['labelShape', 'labelBackgroundShape'].forEach((id) => {
            const shape = shapeMap[id];
            if (shape === null || shape === void 0 ? void 0 : shape.getAttribute(LOCAL_BOUNDS_DIRTY_FLAG_KEY)) {
                this.boundsCache[`${id}Geometry`] = shape.getGeometryBounds();
                this.boundsCache[`${id}Transform`] = shape.style.transform;
                shape.setAttribute(LOCAL_BOUNDS_DIRTY_FLAG_KEY, false);
            }
        });
        const { zoom } = this.zoomCache;
        const levelShapes = {};
        Object.keys(shapeMap).forEach((shapeId) => {
            const { lod } = shapeMap[shapeId].attributes;
            if (lod !== undefined) {
                levelShapes[lod] = levelShapes[lod] || [];
                levelShapes[lod].push(shapeId);
            }
        });
        this.zoomCache.levelShapes = levelShapes;
        const { maxWidth = '60%' } = this.mergedStyles.labelShape || {};
        this.zoomCache.wordWrapWidth = getWordWrapWidthByEnds([this.sourcePoint, this.targetPoint], maxWidth, 1);
        this.zoomCache.zoom = 1;
        this.zoomCache.zoomLevel = 0;
        if (zoom !== 1)
            this.onZoom(shapeMap, zoom);
    }
    /**
     * Perform additional drawing operations or add custom shapes after drawing edge.
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param shapesChanged An array of shape IDs that have changed and need to be updated.
     * @returns An object that contains some new shapes to be added to the edge.
     */
    afterDraw(model, shapeMap, shapesChanged) {
        return {};
    }
    /**
     * Draw the label shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the label shape of the edge.
     */
    drawLabelShape(model, shapeMap, diffData, diffState) {
        const { keyShape } = shapeMap;
        const { labelShape: shapeStyle } = this.mergedStyles;
        if (!shapeStyle || !shapeStyle.text || !model.data.labelShape)
            return;
        const { position, offsetX: propsOffsetX, offsetY: propsOffsetY, autoRotate = true, maxWidth } = shapeStyle, otherStyle = __rest(shapeStyle, ["position", "offsetX", "offsetY", "autoRotate", "maxWidth"]);
        const positionPreset = {
            textAlign: 'center',
            offsetX: 0,
            offsetY: 0,
            pointRatio: [0.5, 0.501],
        };
        if (isNumber(position)) {
            positionPreset.pointRatio = [position, position + 0.01];
        }
        switch (position) {
            case 'start':
                positionPreset.pointRatio = [0, 0.01];
                positionPreset.textAlign = 'left';
                positionPreset.offsetX = 4;
                break;
            case 'end':
                positionPreset.pointRatio = [0.99, 1];
                positionPreset.textAlign = 'right';
                positionPreset.offsetX = -4;
                break;
            default: // at middle by default
                break;
        }
        const point = keyShape.getPoint(positionPreset.pointRatio[0]);
        let positionStyle = { x: point.x, y: point.y };
        let isRevert = false;
        if (autoRotate) {
            const pointOffset = keyShape.getPoint(positionPreset.pointRatio[1]);
            const angle = Math.atan((point.y - pointOffset.y) / (point.x - pointOffset.x)); // TODO: NaN
            // revert
            isRevert = pointOffset.x < point.x;
            if (isRevert) {
                if (position === 'start') {
                    positionPreset.textAlign = 'right';
                    positionPreset.offsetX = -4;
                }
                else if (position === 'end') {
                    positionPreset.textAlign = 'left';
                    positionPreset.offsetX = 4;
                }
            }
            const offsetX = (propsOffsetX === undefined ? positionPreset.offsetX : propsOffsetX);
            const offsetY = (propsOffsetY === undefined ? positionPreset.offsetY : propsOffsetY);
            // the projection is |offsetX| away from point, along the tangent line of the keyShape's path at point
            const projection = {
                x: point.x + offsetX * Math.cos(angle),
                y: point.y + offsetX * Math.sin(angle),
            };
            // the position of the text is |offsetY| away from projection, along the vertical line of the tangent line at point
            positionStyle = {
                x: projection.x + offsetY * Math.cos(Math.PI - angle),
                y: projection.y + offsetY * Math.sin(Math.PI - angle),
                transform: `rotate(${(angle / Math.PI) * 180})`,
            };
        }
        this.labelPosition = Object.assign(Object.assign({}, positionStyle), { isRevert });
        const wordWrapWidth = getWordWrapWidthByEnds([this.sourcePoint, this.targetPoint], maxWidth, this.zoomCache.zoom);
        const style = Object.assign(Object.assign(Object.assign(Object.assign({}, this.defaultStyles.labelShape), { textAlign: positionPreset.textAlign, wordWrapWidth, isBillboard: true }), positionStyle), otherStyle);
        return this.upsertShape('text', 'labelShape', style, shapeMap, model);
    }
    /**
     * Draw the label background shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the label background shape of the edge.
     */
    drawLabelBackgroundShape(model, shapeMap, diffData, diffState) {
        const { labelShape } = shapeMap;
        if (!labelShape || !labelShape.style.text || !model.data.labelShape)
            return;
        const { labelBackgroundShape, labelShape: labelShapeStyle } = this.mergedStyles;
        const textBBox = this.boundsCache.labelShapeGeometry || labelShape.getGeometryBounds();
        const { x, y, transform, isRevert } = this.labelPosition;
        const { padding } = labelBackgroundShape, backgroundStyle = __rest(labelBackgroundShape, ["padding"]);
        const { balanceRatio = 1 } = this.zoomCache;
        const textWidth = textBBox.max[0] - textBBox.min[0];
        const textHeight = textBBox.max[1] - textBBox.min[1];
        const bgStyle = Object.assign(Object.assign({ fill: '#fff' }, backgroundStyle), { x: textBBox.min[0] - padding[3] + x, y: textBBox.min[1] - padding[0] / balanceRatio + y, width: textWidth + padding[1] + padding[3], height: textHeight + (padding[0] + padding[2]) / balanceRatio, transform: transform });
        if (labelShapeStyle.position === 'start') {
            if (isRevert) {
                bgStyle.transformOrigin = `${bgStyle.width - padding[1]} ${bgStyle.height / 2}`;
            }
            else {
                bgStyle.transformOrigin = `${padding[3]} ${bgStyle.height / 2}`;
            }
        }
        else if (labelShapeStyle.position === 'end') {
            if (isRevert) {
                bgStyle.transformOrigin = `${padding[3]} ${bgStyle.height / 2}`;
            }
            else {
                bgStyle.transformOrigin = `${bgStyle.width - padding[1]} ${bgStyle.height / 2}`;
            }
        }
        else {
            bgStyle.transformOrigin = `${textWidth / 2 + padding[3]} ${textHeight / 2 + padding[0]}`;
        }
        return this.upsertShape('rect', 'labelBackgroundShape', bgStyle, shapeMap, model);
    }
    /**
     * Draw the icon shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the icon shape of the edge.
     */
    drawIconShape(model, shapeMap, diffData, diffState) {
        const { labelShape, labelBackgroundShape, keyShape } = shapeMap;
        const { iconShape: shapeStyle, labelShape: labelShapeProps } = this.mergedStyles;
        const { width, height, fontSize, text, offsetX = 0, offsetY = 0, } = shapeStyle;
        const w = (width || fontSize);
        const h = (height || fontSize);
        const iconShapeType = text ? 'text' : 'image';
        if (iconShapeType === 'text') {
            shapeStyle.textAlign = 'left';
            shapeStyle.textBaseline = 'top';
            shapeStyle.fontSize = w;
        }
        else {
            shapeStyle.width = w;
            shapeStyle.height = h;
        }
        if (labelShape) {
            const referShape = labelBackgroundShape || labelShape;
        }
        if (labelShapeProps) {
            const referShape = labelBackgroundShape || labelShape;
            const referBounds = this.boundsCache.labelBackgroundShapeGeometry ||
                this.boundsCache.labelShapeGeometry ||
                referShape.getGeometryBounds();
            const { min: referMin, max: referMax, halfExtents: referHalExtents, } = referBounds;
            const referHeight = referMax[1] - referMin[1];
            const referWidth = referMax[0] - referMin[0];
            const { x: referX, y: referY, transform: referTransform, } = referShape.attributes;
            const { textAlign: labelAlign } = labelShape.attributes;
            shapeStyle.x = referMin[0] - w + 4 + referX + offsetX;
            shapeStyle.y = referMin[1] + (referHeight - h) / 2 + referY + offsetY;
            if (referTransform) {
                shapeStyle.transform = referTransform;
                if (labelAlign === 'right') {
                    shapeStyle.transformOrigin = `${referWidth / 2 - w / 2 + 4 + referHalExtents[0] - offsetX} ${h / 2 - offsetY}`;
                }
                else if (labelAlign === 'left') {
                    shapeStyle.transformOrigin = `${w + 4 - offsetX} ${h / 2 - offsetY}`;
                }
                else {
                    // labelShape align 'center'
                    shapeStyle.transformOrigin = `${(w + referWidth) / 2 - offsetX} ${h / 2 - offsetY}`;
                }
            }
        }
        else {
            const midPoint = keyShape.getPoint(0.5);
            shapeStyle.x = midPoint.x + offsetX;
            shapeStyle.y = midPoint.y + offsetY;
            // TODO: rotate
        }
        return this.upsertShape(iconShapeType, 'iconShape', shapeStyle, shapeMap, model);
    }
    /**
     * Draw the halo shape of the edge
     * @param model The displayed model of this edge, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the edge.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current edge's state.
     * @returns The display object representing the halo shape of the edge.
     */
    drawHaloShape(model, shapeMap, diffData, diffState) {
        const { keyShape } = shapeMap;
        const { haloShape: haloShapeStyle } = this.mergedStyles;
        if (haloShapeStyle.visible === false)
            return;
        const { nodeName, attributes } = keyShape;
        return this.upsertShape(nodeName, 'haloShape', Object.assign(Object.assign(Object.assign({}, attributes), { stroke: attributes.stroke }), haloShapeStyle), shapeMap, model);
    }
    /**
     * Update the shapes' sizes e.g. labelShape, labelBackgroundShape, to keep the visual size while zooming.
     * @param shapeMap
     * @param zoom
     * @returns
     */
    balanceShapeSize(shapeMap, zoom) {
        const { labelShape, labelBackgroundShape } = shapeMap;
        const balanceRatio = 1 / zoom || 1;
        this.zoomCache.balanceRatio = balanceRatio;
        const { labelShape: labelStyle = {} } = this.mergedStyles;
        const { position = 'bottom' } = labelStyle;
        if (!labelShape)
            return;
        if (position === 'bottom')
            labelShape.style.transformOrigin = '0';
        else
            labelShape.style.transformOrigin = '';
        const oriTransform = this.boundsCache.labelShapeTransform;
        labelShape.style.transform = `${oriTransform} scale(${balanceRatio}, ${balanceRatio})`;
        const wordWrapWidth = this.zoomCache.wordWrapWidth * zoom;
        labelShape.style.wordWrapWidth = wordWrapWidth;
        if (!labelBackgroundShape)
            return;
        const oriBgTransform = this.boundsCache.labelBackgroundShapeTransform;
        labelBackgroundShape.style.transform = `${oriBgTransform} scale(${balanceRatio}, ${balanceRatio})`;
    }
    /**
     * Update the source point { x, y } for the edge. Called in item's draw func.
     * @param point source point
     */
    setSourcePoint(point) {
        this.sourcePoint = point;
    }
    /**
     * Update the target point { x, y } for the edge. Called in item's draw func.
     * @param point target point
     */
    setTargetPoint(point) {
        this.targetPoint = point;
    }
    /**
     * Update all visible nodes on the canvas to be aware of obstacles. Called in item's draw func.
     * @param nodeMap The Map object representing the node map, where keys are node IDs and values are nodes.
     */
    setNodeMap(nodeMap) {
        this.nodeMap = nodeMap;
    }
    /**
     * Adds or updates an arrow marker on the specified position of an edge.
     * @param position - The position where the arrow marker should be added or updated. Can be either 'start' or 'end'.
     * @param arrowConfig - The configuration for the arrow marker. Can be a boolean indicating whether to use the default arrow configuration, or an ArrowStyle object with custom arrow properties.
     * @param bodyStyle - The style of the edge body.
     * @param model - The EdgeDisplayModel that contains the data and style information for the edge.
     * @param resultStyle - The style object where the arrow marker properties will be added or updated.
     */
    upsertArrow(position, arrowConfig, bodyStyle, model, resultStyle) {
        const markerField = `marker${position === 'start' ? 'Start' : 'End'}`;
        if (!arrowConfig) {
            resultStyle[markerField] = null;
            resultStyle[`${markerField}Offset`] = 0;
            return;
        }
        let arrowStyle = {};
        if (isBoolean(arrowConfig)) {
            arrowStyle = Object.assign({}, DEFAULT_ARROW_CONFIG);
        }
        else {
            arrowStyle = arrowConfig;
        }
        const { type = 'triangle', width = 10, height = 10, path: propPath, offset = 0 } = arrowStyle, others = __rest(arrowStyle, ["type", "width", "height", "path", "offset"]);
        const path = propPath ? propPath : getArrowPath(type, width, height);
        resultStyle[markerField] = this.upsertShape('path', `${markerField}Shape`, Object.assign(Object.assign(Object.assign({}, bodyStyle), { fill: type === 'simple' ? '' : bodyStyle.stroke, path, anchor: '0.5 0.5', transformOrigin: 'center' }), others), {}, model);
        resultStyle[`${markerField}Offset`] = width / 2 + offset;
    }
    /**
     * Create (if does not exit in shapeMap) or update the shape according to the configurations.
     * @param type shape's type
     * @param id unique string to indicates the shape
     * @param style style to be updated
     * @param shapeMap the shape map of a edge
     * @param model data model of the edge
     * @returns The display object representing the shape.
     */
    upsertShape(type, id, style, shapeMap, model) {
        return upsertShape(type, id, style, shapeMap, model);
    }
}
//# sourceMappingURL=base.js.map