import { AABB, DisplayObject } from '@antv/g';
import { NodeDisplayModel } from '../../../types';
import { SHAPE_TYPE, SHAPE_TYPE_3D, ShapeStyle, State, LodStrategyObj } from '../../../types/item';
import { NodeModelData, NodeShapeMap, NodeShapeStyles, IAnchorPositionMap } from '../../../types/node';
import { ComboDisplayModel, ComboModelData, ComboShapeStyles, ComboShapeMap } from '../../../types/combo';
import { AnimateCfg } from '../../../types/animate';
export declare abstract class BaseNode {
    type: string;
    defaultStyles: NodeShapeStyles | ComboShapeStyles;
    themeStyles: NodeShapeStyles | ComboShapeStyles;
    mergedStyles: NodeShapeStyles | ComboShapeStyles;
    lodStrategy?: LodStrategyObj;
    boundsCache: {
        keyShapeLocal?: AABB;
        labelShapeGeometry?: AABB;
    };
    /**
     * Cache the scale transform calculated by balancing size, for restoring.
     */
    protected scaleTransformCache: string;
    protected zoomCache: {
        hiddenShape: {
            [shapeId: string]: boolean;
        };
        balanceRatio: number;
        zoom: number;
        zoomLevel: number;
        levelShapes: {
            [level: string]: string[];
        };
        wordWrapWidth: number;
        animateConfig: AnimateCfg;
        firstRender: boolean;
    };
    constructor(props: any);
    /**
     * Get merged styles from `getMergedStyles` and assigns the merged styles to the 'mergedStyles' property.
     * @param model - The NodeDisplayModel or ComboDisplayModel to merge the styles from.
     */
    mergeStyles(model: NodeDisplayModel | ComboDisplayModel): void;
    /**
     * Merge style
     * @param model - The NodeDisplayModel or ComboDisplayModel to retrieve the merged styles from.
     * @returns The merged styles as a NodeShapeStyles object.
     */
    getMergedStyles(model: NodeDisplayModel | ComboDisplayModel): NodeShapeStyles;
    /**
     * Call it after calling draw function to update cache about bounds and zoom levels.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     */
    updateCache(shapeMap: any): void;
    /**
     * Draw all elements related to the node.
     * You should call `drawKeyShape` and `drawAnchorShape`,`drawLabelShape`,`drawIconShape`...as you like.
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns An object containing the keyShape and optional labelShape, iconShape, and some otherShapes properties
     */
    abstract draw(model: NodeDisplayModel | ComboDisplayModel, shapeMap: {
        [shapeId: string]: DisplayObject;
    }, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): {
        keyShape: DisplayObject;
        labelShape?: DisplayObject;
        iconShape?: DisplayObject;
        [otherShapeId: string]: DisplayObject;
    };
    /**
     * Perform additional drawing operations or add custom shapes after drawing node.
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param shapesChanged An array of shape IDs that have changed and need to be updated.
     * @returns An object that contains some new shapes to be added to the node.
     */
    afterDraw(model: NodeDisplayModel | ComboDisplayModel, shapeMap: {
        [shapeId: string]: DisplayObject;
    }, shapesChanged?: string[]): {
        [otherShapeId: string]: DisplayObject;
    };
    /**
     * Set the state for the node.
     * @param state state name
     * @param value state value
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     */
    setState: (name: string, value: boolean, shapeMap: {
        [shapeId: string]: DisplayObject;
    }) => void;
    /**
     * The key function of drawing shape.
     * Draw the key shape of the node based on the provided model and shape map.
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the key shape of the node.
     */
    abstract drawKeyShape(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the label shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the label shape of the node.
     */
    drawLabelShape(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the label background shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the label background shape of the node.
     */
    drawLabelBackgroundShape(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        oldState: State[];
        newState: State[];
    }): DisplayObject;
    /**
     * Draw the icon shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the icon shape of the node.
     */
    drawIconShape(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the halo shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the halo shape of the node.
     */
    drawHaloShape(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): DisplayObject;
    /**
     * Draw the anchors shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the anchors shape of the node.
     */
    drawAnchorShapes(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): {
        [shapeId: string]: DisplayObject;
    };
    private getAnchorPosition;
    /**
     * Configures the anchor positions based on the provided keyShapeStyle and returns the configuration.
     * e.g for a CircleNode, it returns: `{"right":keyShapeStyle.x+keyShapeStyle.r, keyShapeStyle.y}`
     * @param keyShapeStyle - The keyShapeStyle object that contains the style information of the key shape.
     * @returns The anchor position configuration as an IAnchorPositionMap object.
     */
    calculateAnchorPosition(keyShapeStyle: any): IAnchorPositionMap;
    /**
     * Draw the badges shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the badges shape of the node.
     */
    drawBadgeShapes(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): {
        [shapeId: string]: DisplayObject;
    };
    /**
     * Draw other shapes(such as preRect,stateIcon) of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the other shapes of the node.
     */
    drawOtherShapes(model: NodeDisplayModel | ComboDisplayModel, shapeMap: NodeShapeMap | ComboShapeMap, diffData?: {
        previous: NodeModelData | ComboModelData;
        current: NodeModelData | ComboModelData;
    }, diffState?: {
        previous: State[];
        current: State[];
    }): {
        [id: string]: DisplayObject;
    };
    /**
     * The listener for graph zooming.
     * 1. show / hide some shapes while zoom level changed;
     * 2. change the shapes' sizes to make them have same visual size while zooming, e.g. labelShape, labelBackgroundShape.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param zoom The zoom level of the graph.
     */
    onZoom: (shapeMap: NodeShapeMap | ComboShapeMap, zoom: number) => void;
    /**
     * Update the shapes' sizes e.g. labelShape, labelBackgroundShape, to keep the visual size while zooming.
     * @param shapeMap
     * @param zoom
     * @returns
     */
    private balanceShapeSize;
    /**
     * Create (if does not exit in shapeMap) or update the shape according to the configurations.
     * @param type shape's type
     * @param id unique string to indicates the shape
     * @param style style to be updated
     * @param shapeMap the shape map of a node / combo
     * @param model data model of the node / combo
     * @returns The display object representing the shape.
     */
    upsertShape(type: SHAPE_TYPE | SHAPE_TYPE_3D, id: string, style: ShapeStyle, shapeMap: NodeShapeMap | ComboShapeMap, model: NodeDisplayModel | ComboDisplayModel): DisplayObject;
}
