import { each } from '@antv/util';
import { BaseNode } from './base';
const defaultDonutPalette = [
    '#61DDAA',
    '#65789B',
    '#F6BD16',
    '#7262FD',
    '#78D3F8',
    '#9661BC',
    '#F6903D',
    '#008685',
    '#F08BB4',
];
export class DonutNode extends BaseNode {
    constructor(props) {
        super(props);
        this.defaultStyles = {
            keyShape: {
                r: 16,
                x: 0,
                y: 0,
            },
            donutShapes: {
                innerSize: 0.6,
                attrs: {},
                colorMap: {},
                zIndex: 1,
            },
        };
        /**
         * Draw a single donut segment
         * @param cfg The configurations of donut segments
         * @param shapes The collections of donut segment shapes
         * @param model
         * @param shapeMap
         * @param diffData
         * @param diffState
         * @returns
         */
        this.drawDonutSegment = (cfg, shapes, model, shapeMap, diffData, diffState) => {
            const { arcR, beginAngle, config, index, lineWidth, zIndex, totalValue, drawWhole = false, } = cfg;
            const id = `donutShape${index}`;
            const percent = config.value / totalValue;
            if (percent < 0.001) {
                // too small to add a fan
                return {
                    beginAngle,
                    shouldEnd: false,
                };
            }
            let arcEnd, endAngle, isLargeArc;
            const arcBegin = calculateArcEndpoint(arcR, beginAngle);
            // draw a path represents the whole circle, or the percentage is close to 1
            if (drawWhole || percent > 0.999) {
                arcEnd = [arcR, 0.0001]; // [arcR * cos(2 * PI), -arcR * sin(2 * PI)]
                isLargeArc = 1;
            }
            else {
                const angle = percent * Math.PI * 2;
                endAngle = beginAngle + angle;
                arcEnd = calculateArcEndpoint(arcR, endAngle);
                isLargeArc = angle > Math.PI ? 1 : 0;
            }
            const style = {
                path: [
                    ['M', arcBegin[0], arcBegin[1]],
                    ['A', arcR, arcR, 0, isLargeArc, 0, arcEnd[0], arcEnd[1]],
                ],
                stroke: config.color || defaultDonutPalette[index % defaultDonutPalette.length],
                lineWidth,
                zIndex,
            };
            shapes[id] = this.upsertShape('path', id, style, shapeMap, model);
            return {
                beginAngle: endAngle,
                shouldEnd: drawWhole || percent > 0.999,
            };
        };
        // suggest to merge default styles like this to avoid style value missing
        // this.defaultStyles = mergeStyles([this.baseDefaultStyles, this.defaultStyles]);
    }
    draw(model, shapeMap, diffData, diffState) {
        const { data = {} } = model;
        let shapes = { keyShape: undefined };
        // keyShape
        shapes.keyShape = this.drawKeyShape(model, shapeMap, diffData);
        // haloShape
        if (data.haloShape && this.drawHaloShape) {
            shapes.haloShape = this.drawHaloShape(model, shapeMap, diffData);
        }
        // labelShape
        if (data.labelShape) {
            shapes.labelShape = this.drawLabelShape(model, shapeMap, diffData);
        }
        // labelBackgroundShape
        if (data.labelBackgroundShape) {
            shapes.labelBackgroundShape = this.drawLabelBackgroundShape(model, shapeMap, diffData);
        }
        // anchor shapes
        if (data.anchorShapes) {
            const anchorShapes = this.drawAnchorShapes(model, shapeMap, diffData, diffState);
            shapes = Object.assign(Object.assign({}, shapes), anchorShapes);
        }
        // iconShape
        if (data.iconShape) {
            shapes.iconShape = this.drawIconShape(model, shapeMap, diffData);
        }
        // badgeShape
        if (data.badgeShapes) {
            const badgeShapes = this.drawBadgeShapes(model, shapeMap, diffData, diffState);
            shapes = Object.assign(Object.assign({}, shapes), badgeShapes);
        }
        // donutShapes
        if (data.donutShapes) {
            const donutShapes = this.drawDonutShapes(model, shapeMap, diffData, diffState);
            shapes = Object.assign(Object.assign({}, shapes), donutShapes);
        }
        // otherShapes
        if (data.otherShapes && this.drawOtherShapes) {
            shapes = Object.assign(Object.assign({}, shapes), this.drawOtherShapes(model, shapeMap, diffData));
        }
        return shapes;
    }
    /**
     * Draw a complete donut composed of several segments
     * @param model
     * @param shapeMap
     * @param diffData
     * @param diffState
     * @returns
     */
    drawDonutShapes(model, shapeMap, diffData, diffState) {
        const { donutShapes: { innerSize, attrs, colorMap, zIndex }, } = this.mergedStyles;
        const attrNum = Object.keys(attrs).length;
        if (!attrNum)
            return;
        const { configs, totalValue } = getDonutConfig(attrs, colorMap);
        if (!totalValue)
            return;
        const { lineWidth, arcR } = getDonutSize(shapeMap.keyShape, innerSize);
        let beginAngle = 0;
        const shapes = {};
        each(configs, (config, index) => {
            const result = this.drawDonutSegment({
                arcR,
                beginAngle,
                config,
                index,
                lineWidth,
                zIndex,
                totalValue,
                drawWhole: attrNum === 1,
            }, shapes, model, shapeMap, diffData, diffState);
            if (result.shouldEnd)
                return;
            beginAngle = result.beginAngle;
        });
        return shapes;
    }
    drawKeyShape(model, shapeMap, diffData, diffState) {
        return this.upsertShape('circle', 'keyShape', this.mergedStyles.keyShape, shapeMap, model);
    }
}
/**
 * Calculate the endpoint of an arc segment.
 * @param arcR Radius of the arc.
 * @param angle angle in degrees subtended by arc.
 */
const calculateArcEndpoint = (arcR, angle) => [
    arcR * Math.cos(angle),
    -arcR * Math.sin(angle),
];
/**
 * calculate the total value and format single value for each fan
 * @param donutAttrs
 * @param donutColorMap
 * @returns
 */
const getDonutConfig = (donutAttrs, donutColorMap) => {
    let totalValue = 0;
    const configs = [];
    Object.keys(donutAttrs).forEach((name) => {
        const value = +donutAttrs[name];
        if (isNaN(value))
            return;
        configs.push({
            key: name,
            value,
            color: donutColorMap[name],
        });
        totalValue += value;
    });
    return { totalValue, configs };
};
/**
 * calculate the lineWidth and radius for fan shapes according to the keyShape's radius
 * @param keyShape
 * @returns
 */
const getDonutSize = (keyShape, innerSize) => {
    const keyShapeR = keyShape.attr('r');
    const innerR = innerSize * keyShapeR; // The radius of the inner ring of the donut
    const arcR = (keyShapeR + innerR) / 2; // The average of the radius of the inner ring and the radius of the outer ring
    const lineWidth = keyShapeR - innerR;
    return { lineWidth, arcR };
};
//# sourceMappingURL=donut.js.map