import { BaseNode } from './base';
export class TriangleNode extends BaseNode {
    constructor(props) {
        super(props);
        this.defaultStyles = {
            keyShape: {
                r: 12,
                x: 0,
                y: 0,
                direction: 'up', //'up'|'left'|'right'|'down'
            },
        };
        // suggest to merge default styles like this to avoid style value missing
        // this.defaultStyles = mergeStyles([this.baseDefaultStyles, this.defaultStyles]);
    }
    draw(model, shapeMap, diffData, diffState) {
        const { data = {} } = model;
        let shapes = { keyShape: undefined };
        // keyShape
        shapes.keyShape = this.drawKeyShape(model, shapeMap, diffData);
        // haloShape
        if (data.haloShape && this.drawHaloShape) {
            shapes.haloShape = this.drawHaloShape(model, shapeMap, diffData);
        }
        // labelShape
        if (data.labelShape) {
            shapes.labelShape = this.drawLabelShape(model, shapeMap, diffData);
        }
        // labelBackgroundShape
        if (data.labelBackgroundShape) {
            shapes.labelBackgroundShape = this.drawLabelBackgroundShape(model, shapeMap, diffData);
        }
        // anchor shapes
        if (data.anchorShapes) {
            const anchorShapes = this.drawAnchorShapes(model, shapeMap, diffData, diffState);
            shapes = Object.assign(Object.assign({}, shapes), anchorShapes);
        }
        // iconShape
        if (data.iconShape) {
            shapes.iconShape = this.drawIconShape(model, shapeMap, diffData);
        }
        // badgeShape
        if (data.badgeShapes) {
            const badgeShapes = this.drawBadgeShapes(model, shapeMap, diffData, diffState);
            shapes = Object.assign(Object.assign({}, shapes), badgeShapes);
        }
        // otherShapes
        if (data.otherShapes && this.drawOtherShapes) {
            shapes = Object.assign(Object.assign({}, shapes), this.drawOtherShapes(model, shapeMap, diffData));
        }
        return shapes;
    }
    drawKeyShape(model, shapeMap, diffData, diffState) {
        const { keyShape: keyShapeStyle } = this.mergedStyles;
        return this.upsertShape('path', 'keyShape', Object.assign(Object.assign({}, this.mergedStyles.keyShape), { path: this.getTrianglePath(keyShapeStyle.r, keyShapeStyle.direction) }), shapeMap, model);
    }
    getTrianglePath(r, direction) {
        const halfHeight = (3 * r) / 2;
        const halfLength = (3 * r) / Math.sin((1 / 3) * Math.PI) / 2;
        let path;
        if (direction === 'down') {
            path = [
                ['M', 0, halfHeight],
                ['L', halfLength, -halfHeight],
                ['L', -halfLength, -halfHeight],
                ['Z'],
            ];
        }
        else if (direction === 'left') {
            path = [
                ['M', -halfHeight, 0],
                ['L', halfHeight, halfLength],
                ['L', halfHeight, -halfLength],
                ['Z'],
            ];
        }
        else if (direction === 'right') {
            path = [
                ['M', halfHeight, 0],
                ['L', -halfHeight, halfLength],
                ['L', -halfHeight, -halfLength],
                ['Z'],
            ];
        }
        else {
            //up
            path = [
                ['M', 0, -halfHeight],
                ['L', halfLength, halfHeight],
                ['L', -halfLength, halfHeight],
                ['Z'],
            ];
        }
        return path;
    }
    /**
     * @description: add 'defaultOffsetX' and 'defaultOffsetY' making the icon align to the triangle center
     */
    drawIconShape(model, shapeMap, diffData, diffState) {
        const { keyShape: keyShapeStyle } = this.mergedStyles;
        const { iconShape: shapeStyle } = this.mergedStyles;
        let defaultOffsetX = 0;
        let defaultOffsetY = keyShapeStyle.r / 4;
        if (keyShapeStyle.direction === 'right') {
            defaultOffsetX = -keyShapeStyle.r / 4;
            defaultOffsetY = 0;
        }
        else if (keyShapeStyle.direction === 'left') {
            defaultOffsetX = keyShapeStyle.r / 4;
            defaultOffsetY = 0;
        }
        else if (keyShapeStyle.direction === 'down') {
            defaultOffsetX = 0;
            defaultOffsetY = -keyShapeStyle.r / 4;
        }
        const { width, height, fontSize, text, offsetX = 0, offsetY = 0, } = shapeStyle;
        const w = (width || fontSize);
        const h = (height || fontSize);
        const iconShapeType = text ? 'text' : 'image';
        if (iconShapeType === 'image') {
            shapeStyle.x = -w / 2 + offsetX + defaultOffsetX;
            shapeStyle.y = -h / 2 + offsetY + defaultOffsetY;
            shapeStyle.width = w;
            shapeStyle.height = h;
        }
        else {
            shapeStyle.textAlign = 'center';
            shapeStyle.textBaseline = 'middle';
            shapeStyle.x = offsetX + defaultOffsetX;
            shapeStyle.y = offsetY + defaultOffsetY;
            shapeStyle.fontSize = w;
        }
        return this.upsertShape(iconShapeType, 'iconShape', shapeStyle, shapeMap, model);
    }
    calculateAnchorPosition(keyShapeStyle) {
        const direction = keyShapeStyle.direction;
        const r = keyShapeStyle.r;
        const halfHeight = (3 * r) / 2;
        const halfLength = (3 * r) / Math.sin((1 / 3) * Math.PI) / 2;
        const anchorPositionMap = {};
        if (direction === 'down') {
            anchorPositionMap['bottom'] = [0, halfHeight];
            anchorPositionMap['right'] = anchorPositionMap['default'] = [
                halfLength,
                -halfHeight,
            ];
            anchorPositionMap['left'] = [-halfLength, -halfHeight];
        }
        else if (direction === 'left') {
            anchorPositionMap['top'] = [halfHeight, -halfLength];
            anchorPositionMap['bottom'] = [halfHeight, halfLength];
            anchorPositionMap['left'] = anchorPositionMap['default'] = [
                -halfHeight,
                0,
            ];
        }
        else if (direction === 'right') {
            anchorPositionMap['top'] = [-halfHeight, -halfLength];
            anchorPositionMap['bottom'] = [-halfHeight, halfLength];
            anchorPositionMap['right'] = anchorPositionMap['default'] = [
                halfHeight,
                0,
            ];
        }
        else {
            //up
            anchorPositionMap['left'] = [-halfLength, halfHeight];
            anchorPositionMap['top'] = anchorPositionMap['default'] = [
                0,
                -halfHeight,
            ];
            anchorPositionMap['right'] = [halfLength, halfHeight];
        }
        return anchorPositionMap;
    }
}
//# sourceMappingURL=triangle.js.map