import type { IGraph } from 'types';
import { StackCfg, StackType } from '../../../types/history';
import { Plugin as Base, IPluginBaseConfig } from '../../../types/plugin';
import { Command } from './command';
/**
 * The `HistoryConfig` interface contains the following properties:

- `enableStack`: An optional boolean value that indicates whether to enable the stack.
- `stackCfg`: A required object of type `StackCfg` representing the stack configuration.

The `StackCfg` type is defined as an object with the following properties:

- `stackSize`: An optional number representing the size of the stack.
- `stackActive`: An optional boolean value indicating whether the stack is active. If active, operations can be pushed onto the stack; otherwise, they cannot.
- `excludes`: An optional array of strings representing APIs that should be excluded from being put on the stack, even if their operation type is not ignored.
- `includes`: An optional array of strings representing APIs that should be included in being put on the stack.
- `ignoreAdd`: An optional boolean value indicating whether to ignore add operations.
- `ignoreRemove`: An optional boolean value indicating whether to ignore remove operations.
- `ignoreUpdate`: An optional boolean value indicating whether to ignore update operations.
- `ignoreStateChange`: An optional boolean value indicating whether to ignore state change operations.
- `ignoreLayerChange`: An optional boolean value indicating whether to ignore layer change operations.
- `ignoreDisplayChange`: An optional boolean value indicating whether to ignore display change operations.

 */
export interface HistoryConfig extends IPluginBaseConfig {
    enableStack?: boolean;
    stackCfg: StackCfg;
}
/**
 * Retrieve the category of a given API based on its name.
 * @param apiName name of API
 */
export declare const getCategoryByApiName: (apiName: any) => string | number | symbol;
declare class HistoryStack<T> {
    private items;
    private maxStep;
    constructor(maxStep?: number);
    push(item: T): void;
    pop(): T;
    size(): number;
    clear(): void;
    toArray(): T[];
}
export declare class History extends Base {
    readonly cfg: StackCfg;
    readonly enableStack: boolean;
    protected undoStack: HistoryStack<Command[]>;
    protected redoStack: HistoryStack<Command[]>;
    protected stackSize: number;
    protected stackActive: boolean;
    protected withoutStackingCounter: number;
    protected isBatching: boolean;
    protected batchCommands: Command[];
    constructor(options?: HistoryConfig);
    getDefaultCfgs(): HistoryConfig;
    init(graph: IGraph): void;
    getUndoStack(): Command[][];
    getRedoStack(): Command[][];
    getStack(): Record<string, Command[][]>;
    clear(): void;
    clearUndoStack(): void;
    clearRedoStack(): void;
    isEnable(): boolean;
    push(cmd: Command[], stackType?: StackType, isNew?: boolean): void;
    /**
     * Pause stacking operations.
     */
    pauseStacking(): void;
    /**
     * Resume stacking operations.
     */
    resumeStacking(): void;
    undo(): this;
    redo(): this;
    canUndo(): boolean;
    canRedo(): boolean;
    private initBatchCommands;
    /**
     * Begin a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    startBatch(): void;
    /**
     * End a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    stopBatch(): void;
    /**
     * Execute the provided function in a batch mode
     * @param callback
     */
    batch(callback: any): void;
    getEvents(): {
        afteritemchange: (props: any) => void;
        afteritemzindexchange: (props: any) => void;
        afteritemstatechange: (props: any) => void;
        afteritemvisibilitychange: (props: any) => void;
        aftercollapsecombo: (props: any) => void;
        afterexpandcombo: (props: any) => void;
    };
    private handleUpdateHistory;
    /**
     * Determine if a given operation should be pushed onto the history stack based on various configurations.
     * @param {string} apiName name of the API operation.
     * @param {boolean} stack Optional. whether push this operation into graph's stack.
     */
    private shouldPushToStack;
    notify(graph: any, eventName: any, ...data: any[]): void;
}
export {};
