import { DisplayObject } from '@antv/g';
import { ComboModel, ID, IGraph, NodeModel } from '../../../types';
import { ShapeStyle } from '../../../types/item';
import { BubblesetCfg } from './types';
type LabelPlacement = 'top' | 'bottom' | 'right' | 'left';
export interface HullComponentOptions {
    id: string;
    members?: ID[];
    nonMembers?: ID[];
    style?: ShapeStyle;
    padding?: number;
    type?: 'bubble' | 'round-convex' | 'smooth-convex';
    labelShape?: ShapeStyle & {
        position?: LabelPlacement;
        offsetX?: number;
        offsetY?: number;
        maxWidth?: string | number;
        /** Whether rotate automatically according to the slope of the hull edge. */
        autoRotate?: boolean;
    };
}
interface HullComponentFullOptions extends HullComponentOptions {
    /** Controlling the effect of the bubble more finely (the scaling degree of the point and edge design, setting the granularity), generally no configuration is required */
    bubbleCfg?: BubblesetCfg;
}
/**
 * The outline of the member used to wrap the interior.
 * convex hull(convex hull): http://geomalgorithms.com/a10-_hull-1.html#Monotone%20Chain
 * bubble: bubbleset algorithm, refer: http://vialab.science.uoit.ca/wp-content/papercite-data/pdf/col2009c.pdf
 * 通By configuring padding, you can adjust the tightness of the wrapping contour to the node.
 */
export default class Hull {
    graph: IGraph;
    path: any[][];
    members: (NodeModel | ComboModel)[];
    nonMembers: (NodeModel | ComboModel)[];
    padding: number;
    bubbleCfg: Partial<BubblesetCfg>;
    type: 'bubble' | 'round-convex' | 'smooth-convex';
    hullShape: DisplayObject;
    options: HullComponentFullOptions;
    constructor(graph: IGraph, options: HullComponentFullOptions);
    getDefaultCfg(): HullComponentFullOptions;
    setPadding(): void;
    setType(): void;
    getPath(members: (NodeModel | ComboModel)[], nonMembers: (NodeModel | ComboModel)[]): any;
    render(): void;
    private getLabelStyle;
    /**
     * Add a member of hull, and if the member was originally in nonMembers, remove it from nonMembers.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Whether successfully added.
     */
    addMember(itemIds: ID | ID[]): boolean;
    /**
     * Add the nodes that the hull needs to exclude, and if the member was originally in the members, remove it from the members.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Whether successfully added.
     */
    addNonMember(itemIds: ID | ID[]): boolean;
    /**
     * Remove members from the hull.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Returns true if the removal is successful, otherwise returns false.
     */
    removeMember(itemIds: ID | ID[]): boolean;
    /**
     * Remove non members from the hull.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Returns true if the removal is successful, otherwise returns false.
     */
    removeNonMember(itemIds: ID | ID[]): boolean;
    updateMembers(memberIds?: ID[], nonMemberIds?: ID[]): void;
    updateStyle(style: ShapeStyle): void;
    /**
     * Update the options
     * @param options hull options
     */
    updateOptions(options: Partial<HullComponentOptions>): void;
    /**
     * Whether a node/edge is inside the range of the hull.
     * @param itemId
     */
    contain(itemId: any): boolean;
    destroy(): void;
}
export {};
