import { Canvas } from '@antv/g';
import { IGraph } from '../../../types';
import { RendererName } from '../../../types/render';
import { Plugin as Base, IPluginBaseConfig } from '../../../types/plugin';
type ItemLegendConfig = {
    enable: boolean;
    typeField?: string;
    title?: string;
    rows?: number;
    cols?: number;
    colPadding?: number;
    rowPadding?: number;
    padding?: number | number[];
    /**
     * formatter for the label
     */
    labelFormatter?: (type: string) => string;
    /**
     * style for the label, will be passed in GUI Legend's style with `itemLabel` prefix
     */
    labelStyle?: {
        [shapeAttr: string]: unknown;
    };
    markerStyle?: {
        shape?: string | {
            [nodeType: string]: string;
        };
        size?: number | number[] | {
            [nodeType: string]: number | number[];
        };
        color?: string | {
            [nodeType: string]: string;
        };
        [styleName: string]: unknown | {
            [nodeType: string]: unknown;
        };
    };
};
export interface LegendConfig extends IPluginBaseConfig {
    /**
     * Container for the legend, using graph's container by default
     */
    container?: HTMLDivElement | string;
    /**
     * Renderer for the legend canvas, 'canvas' by default
     */
    renderer?: RendererName;
    /**
     * User-defined GCanvas
     */
    canvas?: Canvas;
    /**
     * The color used to clear the canvas when it is initialized.
     * @see https://g.antv.antgroup.com/api/canvas/options#background
     */
    background?: string;
    /**
     * ClassName for the DOM wrapper, "g6-category-legend" by default
     */
    className?: string;
    /**
     * Size for the legend canvas, 'fit-content', or an array of number(px) and string(percentage with %)
     */
    size?: 'fit-content' | [number | string, number | string];
    /**
     * Orientation for the legend layout.
     */
    orientation?: 'horizontal' | 'vertical';
    /**
     * Selected state name, triggered while clicking a legend item. Click will not take effect if selectedState is not assigned
     */
    selectedState?: string;
    activeState?: string;
    inactiveState?: string;
    node?: ItemLegendConfig;
    edge?: ItemLegendConfig;
}
export declare class Legend extends Base {
    /**
     * GUI instance for the node legend.
     */
    private nodeLegend;
    /**
     * GUI instance for the edge legend.
     */
    private edgeLegend;
    private wrapper;
    private canvas;
    private size;
    private selectedTypes;
    private activeType;
    private selectedIds;
    private activeIds;
    private styleCache;
    constructor(options?: LegendConfig);
    getDefaultCfgs(): LegendConfig;
    getEvents(): {
        afterrender: () => void;
    };
    init(graph: IGraph): void;
    /**
     * Update or create the legend including node and edge legend.
     */
    private updateLegend;
    /**
     * Creates a wrapper div element for the graph and appends it to the container.
     * @returns {HTMLDivElement} - The created wrapper div element.
     */
    private createWrapper;
    /**
     * Creates a legend for the graph based on the given item type (node or edge).
     * @param {string} itemType - The type of item for which the legend is being created (node or edge).
     * @returns {Category} - The created legend.
     */
    private upsertLegend;
    /**
     * Updates the layout by positioning the node and edge legends.
     */
    private updateLayout;
    /**
     * generate a function that takes an legend's item object and returns an object with Shape and style properties for the item marker rendering.
     * @param itemType specifies whether the item is a node or an edge.
     * @param markerShape
     * @returns
     */
    private getShapeConfig;
    /**
     * Calculates the size of the legend based on the number of rows and columns, padding, and the orientation of the legend.
     * @param rows
     * @param cols
     * @param padding
     * @returns
     */
    private getLegendSize;
    /**
     * Handles the mouse enter event for a given element.
     * @param {Object} ele - The element (legend item marker) that triggered the event.
     * @param {Array} types - An array of all the types on the legend.
     * @param {string} typeField - The field that contains the type information.
     * @param {string} itemType - The type of item, node or edge.
     * @param {Array} datas - An array of data models.
     */
    private handleMouseEnter;
    /**
     * Handles mouse leave event for a given element.
     * @param {HTMLElement} ele - The element (legend item marker) that triggered the event.
     * @param {Array} types - An array of all the types on the legend.
     * @param {string} typeField - The field that contains the type information.
     * @param {string} itemType - The type of the item, node or edge.
     */
    private handleMouseLeave;
    /**
     * Handles the click event on a legend category item.
     * @param {HTMLElement} ele - The legend category item element that was clicked.
     * @param {Array} types - An array of all the types on the legend.
     * @param {string} typeField - The field in the data model that contains the type of the data.
     * @param {string} itemType - The type of the item, node or edge.
     * @param {Array} datas - An array of data models.
     */
    private handleClick;
    destroy(): void;
}
export {};
