import { AABB } from '@antv/g';
import { PolyPoint } from 'types/common';
/**
 * Retrieve the width of a bounding box
 * @param bbox a bounding box
 * @returns width of bbox
 */
export declare const getBBoxWidth: (bbox: AABB) => number;
/**
 * Retrieve the height of a bounding box
 * @param bbox a bounding box
 * @returns height of box
 */
export declare const getBBoxHeight: (bbox: AABB) => number;
/**
 * Calculate the bounding box of point
 * @param point a 2d point
 * @returns the bounding box of the given point
 */
export declare const getBBoxFromPoint: (point: PolyPoint) => AABB;
/**
 * Calculate the bounding box for a list of 2d points
 * @param points a list of 2d points
 * @returns bbox encompassing all the given points
 */
export declare const getBBoxFromPoints: (points?: PolyPoint[]) => AABB;
/**
 * Expand the given bounding box by a specified offset
 * @param bbox
 * @param offset
 * @returns
 */
export declare const getExpandedBBox: (bbox: AABB, offset: number) => AABB;
/**
 * Merge two bboxes into a single bbox that encompasses both
 * @param b1 the first bbox
 * @param b2 the second bbox
 * @returns a whole bbox
 */
export declare const mergeBBox: (b1: AABB, b2: AABB) => AABB;
/**
 * Retrieve the four corners of a bounding box
 * @param bbox a bounding box
 * @returns a list containing four tuples representing the corners in anticlockwise order
 */
export declare const getBBoxPoints: (bbox: AABB) => PolyPoint[];
/**
 * Determine if a given point is outside of the bounding box
 * @param point 2D point
 * @param bbox a bounding box
 * @returns True if the point is outside of box. False otherwise.
 */
export declare const isPointOutsideBBox: (point: PolyPoint, bbox: AABB) => boolean;
/**
 * Retrieve two points where a vertical line at a specific x-coordinate intersects with a bbox
 * @param bbox a bounding box
 * @param x x-coord of the vertical line
 * @returns two intersection points or none
 */
export declare const getBBoxXCrossPoints: (bbox: AABB, x: number) => PolyPoint[];
/**
 * Retrieve two points where a horizontal line at a specific x-coordinate intersects with a bbox
 * @param bbox a bounding box
 * @param y y-coord of the horizontal line
 * @returns two intersection points or none
 */
export declare const getBBoxYCrossPoints: (bbox: AABB, y: number) => PolyPoint[];
/**
 * Retrieve points where the horizontal and vertical line passing through a specific 2d point intersects with a bbox
 * @param bbox a bounding box
 * @param point a 2d point
 * @returns intersection points
 */
export declare const getBBoxCrossPointsByPoint: (bbox: AABB, point: PolyPoint) => PolyPoint[];
/**
 * Determine whether the given point is a point in the horizontal direction
 * @param point a 2d point
 * @param bbox a bounding box
 * @returns 0 if the point is in the center; True if the point is closer to the horizontal axis; False otherwise.
 */
export declare const isHorizontalPoint: (point: PolyPoint, bbox: AABB) => boolean | number;
/**
 * Retrieve the anchor point of bbox
 * @param bbox the expanded bbox of original node bbox by offset, and formatted by gridSize
 * @param point anchorPoint of position formatted by gridSize
 * @param anotherPoint anotherPoint
 * @returns the anchor point of bbox
 */
export declare const getExpandedBBoxPoint: (bbox: AABB, point: PolyPoint, anotherPoint: PolyPoint) => PolyPoint;
/**
 * Determine if two line segments defined by p0 p1 and p2 p3 intersect
 * @param p0 Endpoint of the first line segment
 * @param p1 Endpoint of the first line segment
 * @param p2 Endpoint of the second line segment
 * @param p3 Endpoint of the second line segment
 * @returns True if the line segments intersect, False otherwise
 */
export declare const isSegmentsIntersected: (p0: PolyPoint, p1: PolyPoint, p2: PolyPoint, p3: PolyPoint) => boolean;
/**
 * Determine if a line segment intersects with a bounding box
 * @param p1 endpoint of the line segment
 * @param p2 endpoint of the line segment
 * @param bbox a bounding box
 * @returns True if the line segment intersects with bbox, False otherwise
 */
export declare const isSegmentCrossingBBox: (p1: PolyPoint, p2: PolyPoint, bbox: AABB) => boolean;
