import { AABB } from '@antv/g';
import { ID } from '@antv/graphlib';
import Node from '../item/node';
import { Point, PolyPoint } from '../types/common';
import { RouterCfg } from './router';
/**
 * Simplify points of polyline by removing duplicated points
 * @param points list of 2d points
 * @returns unique set of points
 */
export declare const simplifyPolylinePoints: (points: PolyPoint[]) => PolyPoint[];
/**
 * Use the A* path-finding algorithm to obtain the shortest path. Considering that there is no need to implement automatic obstacle avoidance, use waypoints to plan the path
 */
export declare const simplePathFinder: (start: PolyPoint, end: PolyPoint, startNode: Node, endNode: Node, cfg: RouterCfg) => PolyPoint[];
/**
 * 如果 points 中的一个节点 x 与 p 相等，则消耗 -2。y 同
 * 即优先选择和 points 在同一水平线 / 垂直线上的点
 */
export declare const _costByPoints: (p: PolyPoint, points: PolyPoint[]) => number;
/**
 * ps 经过 p 到 pt 的距离，减去其他路过节点造成的消耗
 */
export declare const heuristicCostEstimate: (p: PolyPoint, ps: PolyPoint, pt: PolyPoint, source?: PolyPoint, target?: PolyPoint) => number;
export declare const reconstructPath: (pathPoints: PolyPoint[], pointById: any, cameFrom: any, currentId: ID, iterator?: number) => void;
/**
 * 从 arr 中删去 item
 */
export declare const removeFrom: (arr: PolyPoint[], item: PolyPoint) => void;
/**
 * 在 points 中找到满足 x 或 y 和 point 的 x 或 y 相等，且与 point 连线不经过 bbox1 与 bbox2 的点
 */
export declare const getNeighborPoints: (points: PolyPoint[], point: PolyPoint, bbox1: AABB, bbox2: AABB) => PolyPoint[];
/** A-Star Algorithm using waypoints */
export declare const pathFinderUsingWaypoints: (points: PolyPoint[], start: PolyPoint, goal: PolyPoint, sBBox: AABB, tBBox: AABB, os: any, ot: any) => PolyPoint[];
/**
 * Calculate the two points necessary to draw a rounded corner between three points, given a radius.
 */
export declare const getBorderRadiusPoints: (p0: Point, p1: Point, p2: Point, r: number) => Point[];
/** Draw a polyline path */
export declare const pointsToPolyline: (points: Point[], borderRadius?: number, z?: boolean) => string;
/**
 * Get polyline path
 * @param id current edge id
 * @param points list of vertices
 * @param radius radius of the corner
 * @param z  whether is closed from the last point to the first point
 * @returns
 */
export declare const getPolylinePath: (id: ID, points: Point[], radius?: number, z?: boolean) => string;
/**
 * Get key waypoints between two points
 * @param start 2d start point
 * @param end 2d end point
 * @param offset offset
 * @param sNode source node
 * @param tNode target node
 * @returns key waypoints between two points
 */
export declare const getPolylinePoints: (start: PolyPoint, end: PolyPoint, offset: number, sNode?: Node, tNode?: Node) => PolyPoint[];
/**
 * Remove the intermediate points that are consecutively the same as x and different from y; remove the intermediate points that are consecutively the same as y but different from x
 * @param points list of 2d points
 * @returns
 */
export declare const removeRedundantPoint: (points: Point[]) => Point[];
/**
 * sorted array ascendly
 * add new item to proper index when calling add
 */
export declare class SortedArray {
    arr: {
        id: string;
        value: number;
    }[];
    private map;
    constructor();
    private _innerAdd;
    add(item: any): void;
    remove(id: any): void;
    private _clearAndGetMinId;
    private _findFirstId;
    minId(clear: any): any;
}
interface ComputedEdge {
    id: ID;
    p1: Point;
    p2: Point;
    bbox: AABB;
}
/** Define a quadtree */
export declare class QuadTree {
    boundary: AABB;
    private readonly capacity;
    private edges;
    private divided;
    private northwest?;
    private northeast?;
    private southwest?;
    private southeast?;
    constructor(boundary: AABB, capacity: number);
    insert(edge: ComputedEdge): any;
    subdivide(): void;
    queryRange(range: AABB): ComputedEdge[];
}
/** Collision detector */
export declare class EdgeCollisionChecker {
    private quadTree;
    constructor(quadTree: QuadTree);
    /** Check if node and edge intersect */
    getCollidingEdges(nodeBBox: AABB): ComputedEdge[];
}
export {};
