"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const g_1 = require("@antv/g");
const util_1 = require("@antv/util");
const point_1 = require("../util/point");
const node_1 = __importDefault(require("./node"));
class Combo extends node_1.default {
    constructor(props) {
        super(Object.assign(Object.assign({}, props), { type: 'combo' }));
        /**
         * Sometimes no changes on edge data, but need to re-draw it
         * e.g. source and target nodes' position changed
         */
        this.forceUpdate = (0, util_1.throttle)(() => {
            this.updateModelByBounds(this.displayModel);
            if (!this.destroyed)
                this.draw(this.displayModel);
        }, 16, {
            leading: true,
            trailing: true,
        });
    }
    init(props) {
        this.getCombinedBounds = props.getCombinedBounds;
        this.getChildren = props.getChildren;
        super.init(props);
    }
    draw(displayModel, diffData, diffState, animate = true, onfinish = () => { }) {
        if (displayModel.data.collapsed) {
            Object.keys(this.themeStyles.collapsed).forEach((shapeId) => {
                displayModel.data[shapeId] = displayModel.data[shapeId] || {};
                displayModel.data[shapeId] = Object.assign(Object.assign({}, displayModel.data[shapeId]), this.themeStyles.collapsed[shapeId]);
                if (this.themeStyles.collapsed[shapeId].contentType === 'childCount') {
                    displayModel.data[shapeId].text = `${this.getChildren().length || 0}`;
                }
            });
        }
        super.draw(displayModel, diffData, diffState, animate, onfinish);
    }
    /**
     * Maps (mapper will be function, value, or encode format) model to displayModel and find out the shapes to be update for incremental updating.
     * @param model inner model
     * @param diffData changes from graphCore changed event
     * @param isReplace whether replace the whole data or partial update
     * @returns
     */
    getDisplayModelAndChanges(innerModel, diffData, isReplace) {
        const superResult = super.getDisplayModelAndChanges(innerModel, diffData, isReplace);
        const model = superResult.model;
        return {
            model: this.updateModelByBounds(model),
            typeChange: superResult.typeChange,
        };
    }
    updateModelByBounds(model) {
        const bounds = this.getCombinedBounds();
        this.cacheCombinedBounds = bounds;
        if (bounds) {
            const { size, center } = bounds;
            model.data.x = center[0];
            model.data.y = center[1];
            model.data.z = center[2];
            if (!model.data.collapsed) {
                model.data.keyShape = model.data.keyShape || {};
                const { padding = [0, 0, 0, 0] } = model.data.keyShape;
                const width = size[0] + padding[1] + padding[3];
                const height = size[1] + padding[0] + padding[2];
                model.data.keyShape.width = width;
                model.data.keyShape.height = height;
                model.data.keyShape.r = Math.sqrt(width * width + height * height) / 2;
            }
        }
        return model;
    }
    getPosition() {
        const { x = 0, y = 0, z = 0 } = this.model.data;
        this.cacheCombinedBounds =
            this.cacheCombinedBounds || this.getCombinedBounds();
        const { center } = this.cacheCombinedBounds || {};
        return center
            ? {
                x: center[0],
                y: center[1],
                z: center[2],
            }
            : {
                x: x,
                y: y,
                z: z,
            };
    }
    getIntersectPoint(point, innerPoint, anchorPoints) {
        var _a, _b;
        const { keyShape } = this.shapeMap;
        const shapeType = keyShape.nodeName;
        const { collapsed, keyShape: keyShapeStyle } = this.displayModel.data;
        const { x, y, z } = innerPoint;
        let intersectPoint;
        switch (shapeType) {
            case 'circle':
                intersectPoint = (0, point_1.getCircleIntersectByPoint)({
                    x,
                    y,
                    r: (collapsed ? keyShapeStyle === null || keyShapeStyle === void 0 ? void 0 : keyShapeStyle.r : keyShape.attributes.r) ||
                        keyShape.attributes.r,
                }, point);
                break;
            default: {
                const bbox = ((_a = this.renderExt.boundsCache) === null || _a === void 0 ? void 0 : _a.keyShapeLocal) ||
                    keyShape.getLocalBounds();
                intersectPoint = (0, point_1.getRectIntersectByPoint)(collapsed
                    ? {
                        x: x - keyShapeStyle.width / 2,
                        y: y - keyShapeStyle.height / 2,
                        width: keyShapeStyle.width,
                        height: keyShapeStyle.height,
                    }
                    : {
                        x: x + bbox.min[0],
                        y: y + bbox.min[1],
                        width: bbox.max[0] - bbox.min[0],
                        height: bbox.max[1] - bbox.min[1],
                    }, point);
            }
        }
        let anchorPointsPositions = this.anchorPointsCache;
        if (!anchorPointsPositions) {
            const keyShapeBBox = ((_b = this.renderExt.boundsCache) === null || _b === void 0 ? void 0 : _b.keyShapeLocal) ||
                this.shapeMap.keyShape.getLocalBounds();
            const keyShapeWidth = keyShapeBBox.max[0] - keyShapeBBox.min[0];
            const keyShapeHeight = keyShapeBBox.max[1] - keyShapeBBox.min[1];
            anchorPointsPositions = anchorPoints.map((pointRatio) => {
                const [xRatio, yRatio] = pointRatio;
                return {
                    x: keyShapeWidth * (xRatio - 0.5) + x,
                    y: keyShapeHeight * (yRatio - 0.5) + y,
                };
            });
            this.anchorPointsCache = anchorPointsPositions;
        }
        let linkPoint = intersectPoint;
        // If the node has anchorPoints in the data, find the nearest anchor point.
        if (anchorPoints.length) {
            if (!linkPoint) {
                // If the linkPoint is failed to calculate.
                linkPoint = point;
            }
            linkPoint = (0, point_1.getNearestPoint)(anchorPointsPositions, linkPoint).nearestPoint;
        }
        if (!linkPoint) {
            // If the calculations above are all failed, return the data's position
            return { x, y };
        }
        return linkPoint;
    }
    // @ts-ignore
    clone(containerGroup, onlyKeyShape, disableAnimate, getCombinedBounds, getChildren) {
        if (onlyKeyShape) {
            const clonedKeyShape = this.shapeMap.keyShape.cloneNode();
            const pos = this.group.getPosition();
            const clonedGroup = new g_1.Group();
            clonedGroup.setPosition(pos);
            clonedGroup.appendChild(clonedKeyShape);
            containerGroup.appendChild(clonedGroup);
            return clonedGroup;
        }
        const clonedModel = (0, util_1.clone)(this.model);
        clonedModel.data.disableAnimate = disableAnimate;
        const clonedNode = new Combo({
            model: clonedModel,
            renderExtensions: this.renderExtensions,
            containerGroup,
            mapper: this.mapper,
            stateMapper: this.stateMapper,
            zoom: this.zoom,
            theme: {
                styles: this.themeStyles,
                lodStrategy: this.lodStrategy,
            },
            getCombinedBounds: getCombinedBounds || this.getCombinedBounds,
            getChildren: getChildren || this.getChildren,
        });
        Object.keys(this.shapeMap).forEach((shapeId) => {
            if (!this.shapeMap[shapeId].isVisible())
                clonedNode.shapeMap[shapeId].hide();
        });
        return clonedNode;
    }
}
exports.default = Combo;
//# sourceMappingURL=combo.js.map