"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.InteractionController = void 0;
const stdlib_1 = require("../../stdlib");
const extension_1 = require("../../util/extension");
const event_1 = require("../../types/event");
const event_2 = require("../../util/event");
/**
 * Wraps the listener with error logging.
 * @returns a new listener with error logging.
 */
const wrapListener = (type, eventName, listener) => {
    return (event) => {
        try {
            listener(event);
        }
        catch (error) {
            console.error(`G6: Error occurred in "${eventName}" phase of the behavior "${type}"!`);
            throw error;
        }
    };
};
/**
 * Manages the interaction extensions and graph modes;
 * Storage related data.
 */
class InteractionController {
    constructor(graph) {
        /**
         * Available behaviors of current mode.
         * @example
         * { 'drag-node': DragNode, 'drag-canvas': DragCanvas }
         */
        this.behaviorMap = new Map();
        /**
         * Listeners added by all current behaviors.
         * @example
         * {
         *   'drag-node': { 'node:pointerEnter': function },
         * }
         */
        this.listenersMap = {};
        /**
         * Subscribe the lifecycle of graph.
         */
        this.tap = () => {
            this.graph.hooks.init.tap(() => this.onModeChange({ mode: 'default' }));
            this.onModeChange({ mode: 'default' });
            this.graph.hooks.modechange.tap(this.onModeChange);
            this.graph.hooks.behaviorchange.tap(this.onBehaviorChange);
        };
        this.validateMode = (mode) => {
            if (mode === 'default')
                return true;
            const modes = this.graph.getSpecification().modes || {};
            return Object.keys(modes).includes(mode);
        };
        this.initBehavior = (config) => {
            const key = typeof config === 'string'
                ? config
                : config.key || config.type;
            if (this.behaviorMap.has(key)) {
                console.error(`G6: Failed to add behavior with key or type "${key}"! It was already added.`);
                return;
            }
            try {
                // Get behavior extensions from useLib.
                const BehaviorClass = (0, extension_1.getExtension)(config, stdlib_1.registery.useLib, 'behavior');
                const options = typeof config === 'string' ? {} : config;
                const behavior = new BehaviorClass(options);
                behavior.graph = this.graph;
                if (behavior) {
                    this.behaviorMap.set(key, behavior);
                }
                return behavior;
            }
            catch (error) {
                console.error(`G6: Failed to initialize behavior with key or type "${key}"!`, error);
                return null;
            }
        };
        this.destroyBehavior = (key, behavior) => {
            try {
                behavior.destroy();
            }
            catch (error) {
                console.error(`G6: Failed to destroy behavior with key or type "${key}"!`, error);
            }
        };
        this.addListeners = (key, behavior) => {
            const events = behavior.getEvents();
            this.listenersMap[key] = {};
            Object.keys(events).forEach((eventName) => {
                // Wrap the listener with error logging.
                const listener = wrapListener(key, eventName, events[eventName].bind(behavior));
                this.graph.on(eventName, listener);
                this.listenersMap[key][eventName] = listener;
            });
        };
        this.removeListeners = (key) => {
            Object.keys(this.listenersMap[key] || {}).forEach((eventName) => {
                const listener = this.listenersMap[key][eventName];
                this.graph.off(eventName, listener);
            });
        };
        /**
         * Listener of graph's init hook. Add listeners from behaviors to graph.
         * @param param contains the mode to switch to
         */
        this.onModeChange = (param) => {
            var _a;
            const { mode } = param;
            // Skip if set to same mode.
            if (this.mode === mode) {
                return;
            }
            if (!this.validateMode(mode)) {
                console.warn(`G6: Mode "${mode}" was not specified in current graph.`);
            }
            this.mode = mode;
            // 1. Remove listeners && destroy current behaviors.
            this.behaviorMap.forEach((behavior, key) => {
                this.removeListeners(key);
                this.destroyBehavior(key, behavior);
            });
            // 2. Initialize new behaviors.
            this.behaviorMap.clear();
            const behaviorConfigs = ((_a = this.graph.getSpecification().modes) === null || _a === void 0 ? void 0 : _a[mode]) || [];
            behaviorConfigs.forEach((config) => {
                this.initBehavior(config);
            });
            // 3. Add listeners for each behavior.
            this.listenersMap = {};
            this.behaviorMap.forEach((behavior, key) => {
                this.addListeners(key, behavior);
            });
        };
        /**
         * Listener of graph's behaviorchange hook. Update, add, or remove behaviors from modes.
         * @param param contains action, modes, and behaviors
         */
        this.onBehaviorChange = (param) => {
            const { action, modes, behaviors } = param;
            modes.forEach((mode) => {
                // Do nothing if it's not the current mode.
                // Changes are recorded in `graph.specification`. They are applied in onModeChange.
                if (mode !== this.mode) {
                    return;
                }
                if (action === 'add') {
                    behaviors.forEach((config) => {
                        const key = typeof config === 'string'
                            ? config
                            : config.key || config.type;
                        const behavior = this.initBehavior(config);
                        if (behavior) {
                            this.addListeners(key, behavior);
                        }
                    });
                    return;
                }
                if (action === 'remove') {
                    behaviors.forEach((config) => {
                        const key = typeof config === 'string'
                            ? config
                            : config.key || config.type;
                        const behavior = this.behaviorMap.get(key);
                        if (behavior) {
                            this.removeListeners(key);
                            this.destroyBehavior(key, behavior);
                        }
                    });
                    return;
                }
                if (action === 'update') {
                    behaviors.forEach((config) => {
                        const key = typeof config === 'string'
                            ? config
                            : config.key || config.type;
                        const behavior = this.behaviorMap.get(key);
                        if (behavior) {
                            behavior.updateConfig(config);
                        }
                    });
                }
            });
        };
        this.initEvents = () => {
            Object.values(event_1.CANVAS_EVENT_TYPE).forEach((eventName) => {
                this.graph.canvas.document.addEventListener(eventName, this.handleCanvasEvent);
            });
            const $dom = this.graph.canvas.getContextService().getDomElement();
            Object.values(event_1.DOM_EVENT_TYPE).forEach((eventName) => {
                if ($dom && $dom.addEventListener) {
                    $dom.addEventListener(eventName, this.handleDOMEvent);
                }
            });
        };
        this.handleCanvasEvent = (gEvent) => {
            const itemInfo = (0, event_2.getItemInfoFromElement)(gEvent.target);
            if (!itemInfo) {
                // This event was triggered from an element which belongs to none of the nodes/edges/canvas.
                return;
            }
            // GEvent => G6Event
            const { itemType, itemId } = itemInfo;
            const event = Object.assign(Object.assign({}, gEvent), { itemType,
                itemId,
                gEvent, stopPropagation: gEvent.stopPropagation, 
                // Set currentTarget to this.graph instead of canvas.
                // Because we add listeners like this: `graph.on('node:click', listener)`.
                currentTarget: this.graph });
            // Trigger a dblclick event.
            if (event.type === 'click' && event.detail === 2) {
                this.handleCanvasEvent(Object.assign(Object.assign({}, gEvent), { type: 'dblclick' }));
            }
            // Canvas Events
            if (event.itemType === 'canvas') {
                if (event.type === 'pointermove') {
                    this.handlePointerMove(event);
                }
                this.graph.emit(`canvas:${gEvent.type}`, event);
                this.graph.emit(`${gEvent.type}`, event);
                // contextmenu event for canvas
                if (event.type === 'pointerdown' && event.button === 2) {
                    const contextMenuEvent = (0, event_2.getContextMenuEventProps)(event, this.graph);
                    this.graph.emit(`canvas:contextmenu`, contextMenuEvent);
                    this.graph.emit('contextmenu', contextMenuEvent);
                }
                // debug(`Canvas ${event.type} :`, event);
                return;
            }
            // Item Events (Node/Edge/Combo)
            if (itemType === 'node' || itemType === 'edge' || itemType === 'combo') {
                if (event.type === 'pointermove' || event.type === 'pointerleave') {
                    this.handlePointerMove(event);
                }
                this.graph.emit(`${itemType}:${gEvent.type}`, event);
                this.graph.emit(`${gEvent.type}`, event);
                // contextmenu event for node, edge, or combo
                if (event.type === 'pointerdown' && event.button === 2) {
                    const contextMenuEvent = (0, event_2.getContextMenuEventProps)(event, this.graph);
                    this.graph.emit(`${itemType}:contextmenu`, contextMenuEvent);
                    this.graph.emit('contextmenu', contextMenuEvent);
                }
            }
        };
        /**
         * Emit item's pointerleave/pointerenter events when pointer moves on the canvas.
         */
        this.handlePointerMove = (event) => {
            const prevItemInfo = this.prevItemInfo;
            const curItemInfo = (0, event_2.getItemInfoFromElement)(event.target);
            if ((prevItemInfo === null || prevItemInfo === void 0 ? void 0 : prevItemInfo.itemId) !== (curItemInfo === null || curItemInfo === void 0 ? void 0 : curItemInfo.itemId)) {
                if (prevItemInfo) {
                    const preType = prevItemInfo.itemType;
                    this.graph.emit(`${preType}:pointerleave`, Object.assign(Object.assign({}, event), { itemId: prevItemInfo.itemId, itemType: prevItemInfo.itemType, type: 'pointerleave', target: prevItemInfo.groupElement }));
                }
                if (curItemInfo) {
                    const curType = curItemInfo.itemType;
                    this.graph.emit(`${curType}:pointerenter`, Object.assign(Object.assign({}, event), { type: 'pointerenter', target: curItemInfo.groupElement }));
                }
            }
            this.prevItemInfo = curItemInfo;
        };
        this.handleDOMEvent = (event) => {
            this.graph.emit(event.type, event);
        };
        this.graph = graph;
        this.initEvents();
        this.tap();
    }
    getMode() {
        return this.mode;
    }
}
exports.InteractionController = InteractionController;
//# sourceMappingURL=interaction.js.map