import EventEmitter from '@antv/event-emitter';
import { AABB, Canvas, DisplayObject, PointLike } from '@antv/g';
import { ID } from '@antv/graphlib';
import { Command } from '../stdlib/plugin/history/command';
import type { ComboUserModel, EdgeUserModel, IGraph, NodeUserModel, Specification } from '../types';
import type { CameraAnimationOptions } from '../types/animate';
import type { BehaviorOptionsOf, BehaviorRegistry } from '../types/behavior';
import type { ComboDisplayModel, ComboModel } from '../types/combo';
import type { Padding, Point } from '../types/common';
import type { DataConfig } from '../types/data';
import type { EdgeDisplayModel, EdgeModel } from '../types/edge';
import type { StackType } from '../types/history';
import type { Hooks } from '../types/hook';
import type { ITEM_TYPE, SHAPE_TYPE, ShapeStyle } from '../types/item';
import type { LayoutOptions } from '../types/layout';
import type { NodeDisplayModel, NodeModel } from '../types/node';
import type { RendererName } from '../types/render';
import type { ThemeOptionsOf, ThemeRegistry } from '../types/theme';
import { FitViewRules, GraphTransformOptions } from '../types/view';
import { Plugin as PluginBase } from '../types/plugin';
export default class Graph<B extends BehaviorRegistry, T extends ThemeRegistry> extends EventEmitter implements IGraph<B, T> {
    hooks: Hooks;
    canvas: Canvas;
    container: HTMLElement;
    destroyed: boolean;
    rendererType: RendererName;
    transientCanvas: Canvas;
    backgroundCanvas: Canvas;
    private canvasReady;
    private specification;
    private dataController;
    private interactionController;
    private layoutController;
    private viewportController;
    private itemController;
    private extensionController;
    private themeController;
    private pluginController;
    private defaultSpecification;
    constructor(spec: Specification<B, T>);
    /**
     * Initialize the controllers for different plugins.
     */
    private initControllers;
    private initCanvas;
    private initHistory;
    /**
     * Change the renderer at runtime.
     * @param type renderer name
     * @returns
     */
    changeRenderer(type: any): void;
    /**
     * Initialize the hooks for graph's lifecycles.
     */
    private initHooks;
    /**
     * Update the specs(configurations).
     */
    updateSpecification(spec: Specification<B, T>): Specification<B, T>;
    /**
     * Update the theme specs (configurations).
     */
    updateTheme(theme: ThemeOptionsOf<T>): void;
    /**
     * Get the copy of specs(configurations).
     * @returns graph specs
     */
    getSpecification(): Specification<B, T>;
    /**
     * Input data and render the graph.
     * If there is old data, diffs and changes it.
     * @param data
     * @returns
     * @group Data
     */
    read(data: DataConfig): Promise<void>;
    /**
     * Change graph data.
     * @param data new data
     * @param type the way to change data, 'replace' means discard the old data and use the new one; 'mergeReplace' means merge the common part, remove (old - new), add (new - old)
     * @returns
     * @group Data
     */
    changeData(data: DataConfig, type?: 'replace' | 'mergeReplace'): Promise<void>;
    /**
     * Clear the graph, means remove all the items on the graph.
     * @returns
     */
    clear(): void;
    getViewportCenter(): PointLike;
    transform(options: GraphTransformOptions, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Stop the current transition of transform immediately.
     */
    stopTransformTransition(): void;
    /**
     * Move the graph with a relative distance under viewport coordinates.
     * @param dx x of the relative distance
     * @param dy y of the relative distance
     * @param effectTiming animation configurations
     */
    translate(distance: Partial<{
        dx: number;
        dy: number;
        dz: number;
    }>, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Move the graph to destination under viewport coordinates.
     * @param destination destination under viewport coordinates.
     * @param effectTiming animation configurations
     */
    translateTo({ x, y }: Point, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Zoom the graph with a relative ratio.
     * @param ratio relative ratio to zoom
     * @param origin origin under viewport coordinates.
     * @param effectTiming animation configurations
     */
    zoom(ratio: number, origin?: Point, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Zoom the graph to a specified ratio.
     * @param zoom specified ratio
     * @param origin zoom center
     * @param effectTiming animation configurations
     */
    zoomTo(zoom: number, origin?: PointLike, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Return the current zoom level of camera.
     * @returns current zoom
     */
    getZoom(): number;
    /**
     * Rotate the graph with a relative angle.
     * @param angle
     * @param origin
     * @param effectTiming
     */
    rotate(angle: number, origin?: PointLike, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Rotate the graph to an absolute angle.
     * @param angle
     * @param origin
     * @param effectTiming
     */
    rotateTo(angle: number, origin?: PointLike, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Fit the graph content to the view.
     * @param options.padding padding while fitting
     * @param options.rules rules for fitting
     * @param effectTiming animation configurations
     */
    fitView(options?: {
        padding: Padding;
        rules: FitViewRules;
    }, effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Fit the graph center to the view center.
     * @param effectTiming animation configurations
     */
    fitCenter(effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Move the graph to make the item align the view center.
     * @param item node/edge/combo item or its id
     * @param effectTiming animation configurations
     */
    focusItem(id: ID | ID[], effectTiming?: CameraAnimationOptions): Promise<void>;
    /**
     * Get item by id. We don't want to
     * @param id
     * @returns Node | Edge | Combo
     */
    private getItemById;
    /**
     * Get the size of the graph canvas.
     * @returns [width, height]
     * @group View
     */
    getSize(): number[];
    /**
     * Set the size for the graph canvas.
     * @param number[] [width, height]
     * @group View
     */
    setSize(size: number[]): void;
    /**
     * Get the rendering coordinate according to the canvas dom (viewport) coordinate.
     * @param Point rendering coordinate
     * @returns canvas dom (viewport) coordinate
     * @group View
     */
    getCanvasByViewport(viewportPoint: Point): Point;
    /**
     * Get the canvas dom (viewport) coordinate according to the rendering coordinate.
     * @param Point canvas dom (viewport) coordinate
     * @returns rendering coordinate
     * @group View
     */
    getViewportByCanvas(canvasPoint: Point): Point;
    /**
     * Get the browser coordinate according to the rendering coordinate.
     * @param Point rendering coordinate
     * @returns browser coordinate
     * @group View
     */
    getClientByCanvas(canvasPoint: Point): Point;
    /**
     * Get the rendering coordinate according to the browser coordinate.
     * @param Point browser coordinate
     * @returns rendering coordinate
     * @group View
     */
    getCanvasByClient(clientPoint: Point): Point;
    /**
     * Find a node's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result node's inner data
     * @group Data
     */
    getNodeData(condition: ID | Function): NodeModel | undefined;
    /**
     * Find an edge's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result edge's inner data
     * @group Data
     */
    getEdgeData(condition: ID | Function): EdgeModel | undefined;
    /**
     * Find an combo's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result combo's inner data
     * @group Data
     */
    getComboData(condition: ID | Function): ComboModel | undefined;
    /**
     * Get all the nodes' inner data
     * @returns all nodes' inner data on the graph
     * @group Data
     */
    getAllNodesData(): NodeModel[];
    /**
     * Get all the edges' inner data
     * @returns all edges' inner data on the graph
     * @group Data
     */
    getAllEdgesData(): EdgeModel[];
    /**
     * Get all the combos' inner data
     * @returns all combos' inner data on the graph
     * @group Data
     */
    getAllCombosData(): ComboModel[];
    /**
     * Get one-hop edge ids from a start node.
     * @param nodeId id of the start node
     * @returns one-hop edges' data array
     * @group Data
     */
    getRelatedEdgesData(nodeId: ID, direction?: 'in' | 'out' | 'both'): EdgeModel[];
    /**
     * Get one-hop node ids from a start node.
     * @param nodeId id of the start node
     * @returns one-hop nodes' data array
     * @group Data
     */
    getNeighborNodesData(nodeId: ID, direction?: 'in' | 'out' | 'both'): NodeModel[];
    /**
     * Get the children's data of a combo.
     * @param comboId combo id
     * @returns children's data array
     * @group Data
     */
    getComboChildrenData(comboId: ID): (ComboModel | NodeModel)[];
    protected getDisplayModel(id: ID): NodeDisplayModel | EdgeDisplayModel | ComboDisplayModel;
    /**
     * Retrieve the nearby edges for a given node using quadtree collision detection.
     * @param nodeId node id
     * @group Data
     */
    getNearEdgesForNode(nodeId: ID): EdgeModel[];
    /**
     * Find items which has the state.
     * @param itemType item type
     * @param state state name
     * @param additionalFilter additional filter function
     * @returns items that is the type and has the state
     * @group Item
     */
    findIdByState(itemType: ITEM_TYPE, state: string, value?: string | boolean, additionalFilter?: (item: NodeModel | EdgeModel | ComboModel) => boolean): ID[];
    /**
     * Add one or more node/edge/combo data to the graph.
     * @param itemType item type
     * @param model user data
     * @returns whether success
     * @group Data
     */
    addData(itemType: ITEM_TYPE, models: NodeUserModel | EdgeUserModel | ComboUserModel | NodeUserModel[] | EdgeUserModel[] | ComboUserModel[]): NodeModel | EdgeModel | ComboModel | NodeModel[] | EdgeModel[] | ComboModel[];
    /**
     * Remove one or more node/edge/combo data from the graph.
     * @param item the item to be removed
     * @returns whether success
     * @group Data
     */
    removeData(itemType: ITEM_TYPE, ids: ID | ID[]): void;
    /**
     * Extends the fields of oldValue and newValue
     * to make sure they both contain the union set of their original fields.
     * The missing fields' values are filled using data from displayModel data.
     * @param changes
     * @returns
     */
    private extendChanges;
    /**
     * Update one or more node/edge/combo data on the graph.
     * @param {ITEM_TYPE} itemType 'node' | 'edge' | 'combo'
     * @param models new configurations for every node/edge/combo, which has id field to indicate the specific item
     * @group Data
     */
    updateData(itemType: ITEM_TYPE, models: Partial<NodeUserModel> | Partial<EdgeUserModel> | Partial<ComboUserModel | Partial<NodeUserModel>[] | Partial<EdgeUserModel>[] | Partial<ComboUserModel>[]>): NodeModel | EdgeModel | ComboModel | NodeModel[] | EdgeModel[] | ComboModel[];
    /**
     * Update one or more nodes' positions,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * @param models new configurations with x and y for every node, which has id field to indicate the specific item
     * @group Data
     */
    updateNodePosition(models: Partial<NodeUserModel> | Partial<ComboUserModel | Partial<NodeUserModel>[] | Partial<ComboUserModel>[]>, upsertAncestors?: boolean, disableAnimate?: boolean, callback?: (model: NodeModel | EdgeModel | ComboModel, canceled?: boolean) => void, stack?: boolean): NodeModel | EdgeModel | ComboModel | NodeModel[] | EdgeModel[] | ComboModel[];
    /**
     * Update one or more combos' positions,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * In fact, it changes the succeed nodes positions to affect the combo's position, but not modify the combo's position directly.
     * @param models new configurations with x and y for every combo, which has id field to indicate the specific item
     * @group Data
     */
    updateComboPosition(models: Partial<NodeUserModel> | Partial<ComboUserModel | Partial<NodeUserModel>[] | Partial<ComboUserModel>[]>, upsertAncestors?: boolean, disableAnimate?: boolean, callback?: (model: NodeModel | EdgeModel | ComboModel) => void, stack?: boolean): NodeModel | EdgeModel | ComboModel | NodeModel[] | EdgeModel[] | ComboModel[];
    /**
     * Get history plugin instance
     */
    private getHistoryPlugin;
    private updatePosition;
    private getItemPreviousVisibility;
    /**
     * Show the item(s).
     * @param item the item to be shown
     * @returns
     * @group Item
     */
    showItem(ids: ID | ID[], disableAnimate?: boolean): void;
    /**
     * Hide the item(s).
     * @param item the item to be hidden
     * @returns
     * @group Item
     */
    hideItem(ids: ID | ID[], disableAnimate?: boolean): void;
    /**
     * Make the item(s) to the front.
     * @param ids
     * @returns
     * @group Item
     */
    frontItem(ids: ID | ID[]): void;
    /**
     * Make the item(s) to the back.
     * @param ids
     * @returns
     * @group Item
     */
    backItem(ids: ID | ID[]): void;
    private getItemPreviousStates;
    /**
     * Set state for the item.
     * @param item the item to be set
     * @param state the state name
     * @param value state value
     * @returns
     * @group Item
     */
    setItemState(ids: ID | ID[], states: string | string[], value: boolean): void;
    /**
     * Get the state value for an item.
     * @param id the id for the item
     * @param states the state name
     * @returns {boolean} the state value
     * @group Item
     */
    getItemState(id: ID, state: string): string | boolean;
    /**
     * Get all the state names with value true for an item.
     * @param id the id for the item
     * @returns {string[]} the state names with value true
     * @group Item
     */
    getItemAllStates(id: ID): string[];
    /**
     * Clear all the states for item(s).
     * @param ids the id(s) for the item(s) to be clear
     * @param states the states' names, all the states wil be cleared if states is not assigned
     * @returns
     * @group Item
     */
    clearItemState(ids: ID | ID[], states?: string[]): void;
    /**
     * Get the rendering bbox for a node / edge / combo, or the graph (when the id is not assigned).
     * @param id the id for the node / edge / combo, undefined for the whole graph
     * @returns rendering bounding box. returns false if the item is not exist
     * @group Item
     */
    getRenderBBox(id: ID | undefined, onlyKeyShape?: boolean, isTransient?: boolean): AABB | false;
    /**
     * Get the visibility for a node / edge / combo.
     * @param id the id for the node / edge / combo
     * @returns visibility for the item, false for invisible or unexistence for the item
     * @group Item
     */
    getItemVisible(id: ID): boolean;
    /**
     * Add a new combo to the graph, and update the structure of the existed child in childrenIds to be the children of the new combo.
     * Different from addData with combo type, this API update the succeeds' combo tree strucutres in the same time.
     * @param model combo user data
     * @returns whether success
     * @group Combo
     */
    addCombo(model: ComboUserModel, childrenIds: ID[]): ComboModel;
    /**
     * Collapse a combo.
     * @param comboId combo ids
     * @group Combo
     */
    collapseCombo(comboIds: ID | ID[]): void;
    /**
     * Expand a combo.
     * @param combo combo ids
     * @group Combo
     */
    expandCombo(comboIds: ID | ID[]): void;
    /**
     * Move one or more combos a distance (dx, dy) relatively,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * In fact, it changes the succeed nodes positions to affect the combo's position, but not modify the combo's position directly.
     * @param models new configurations with x and y for every combo, which has id field to indicate the specific item
     * @group Combo
     */
    moveCombo(ids: ID | ID[], dx: number, dy: number, upsertAncestors?: boolean, callback?: (model: NodeModel | EdgeModel | ComboModel, canceled?: boolean) => void): ComboModel[];
    /**
     * Layout the graph (with current configurations if cfg is not assigned).
     */
    layout(options?: LayoutOptions, disableAnimate?: boolean): Promise<void>;
    /**
     * Some layout algorithms has many iterations which can be stopped at any time.
     */
    stopLayout(): void;
    /**
     *
     * @returns
     */
    getLayoutCurrentAnimation(): import("@antv/g").Animation;
    /**
     * Switch mode.
     * @param mode mode name
     * @returns
     * @group Interaction
     */
    setMode(mode: string): void;
    /**
     * Get current mode.
     * @returns mode name
     * @group Interaction
     */
    getMode(): string;
    /**
     * Add behavior(s) to mode(s).
     * @param behaviors behavior names or configs
     * @param modes mode names
     * @returns
     * @group Interaction
     */
    addBehaviors(behaviors: BehaviorOptionsOf<B> | BehaviorOptionsOf<B>[], modes: string | string[]): void;
    /**
     * Remove behavior(s) from mode(s).
     * @param behaviors behavior configs with unique key
     * @param modes mode names
     * @returns
     * @group Interaction
     */
    removeBehaviors(behaviorKeys: string[], modes: string | string[]): void;
    /**
     * Update a behavior on a mode.
     * @param behavior behavior configs, whose name indicates the behavior to be updated
     * @param mode mode name
     * @returns
     * @group Interaction
     */
    updateBehavior(behavior: BehaviorOptionsOf<B>, mode?: string): void;
    /**
     * Add plugin(s) to graph.
     * @param pluginCfgs
     * @returns
     * @group Plugin
     */
    addPlugins(pluginCfgs: ({
        key: string;
        type: string;
        [cfgName: string]: unknown;
    } | string)[]): void;
    /**
     * Remove plugin(s) from graph.
     * @param pluginKeys
     * @returns
     * @group Plugin
     */
    removePlugins(pluginKeys: (PluginBase | string)[]): void;
    /**
     * Update a plugin of the graph.
     * @param plugin plugin configs, whose key indicates the behavior to be updated
     * @returns
     * @group Interaction
     */
    updatePlugin(plugin: {
        key: string;
        type: string;
        [cfg: string]: unknown;
    } | PluginBase): void;
    /**
     * Draw or update a G shape or group to the transient canvas.
     * @param type shape type or item type
     * @param id new shape id or updated shape id for a interation shape, node/edge/combo id for item interaction group drawing
     * @returns upserted shape or group
     * @group Interaction
     */
    drawTransient(type: ITEM_TYPE | SHAPE_TYPE, id: ID, config: {
        action: 'remove' | 'add' | 'update' | undefined;
        style: ShapeStyle;
        onlyDrawKeyShape?: boolean;
        upsertAncestors?: boolean;
    }, canvas?: Canvas): DisplayObject;
    /**
     * Determine if history (redo/undo) is enabled.
     */
    isHistoryEnabled(): boolean;
    /**
     * Push the operation(s) onto the specified stack
     * @param cmd commands to be pushed
     * @param stackType undo/redo stack
     * @param isNew
     */
    pushStack(cmd: Command[], stackType: StackType, isNew?: boolean): void;
    /**
     * Pause stacking operation.
     */
    pauseStacking(): void;
    /**
     * Resume stacking operation.
     */
    resumeStacking(): void;
    /**
     * Execute a callback without allowing any stacking operations.
     * @param callback
     */
    executeWithoutStacking: (callback: () => void) => void;
    /**
     * Retrieve the current redo stack which consists of operations that could be undone
     */
    getUndoStack(): Command[][];
    /**
     * Retrieve the current undo stack which consists of operations that were undone
     */
    getRedoStack(): Command[][];
    /**
     * Retrieve the complete history stack
     * @returns
     */
    getStack(): Record<string, Command[][]>;
    /**
     * Restore n operations that were last n reverted on the graph.
     * @param steps The number of operations to undo. Default to 1.
     * @returns
     */
    undo(): void;
    /**
     * Revert recent n operation(s) performed on the graph.
     * @param steps The number of operations to redo. Default to 1.
     * @returns
     */
    redo(): void;
    /**
     * Indicate whether there are any actions available in the undo stack.
     */
    canUndo(): boolean;
    /**
     * Indicate whether there are any actions available in the redo stack.
     */
    canRedo(): boolean;
    /**
     * Begin a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    startBatch(): void;
    /**
     * End a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    stopBatch(): void;
    /**
     * Execute a provided function within a batched context
     * All operations performed inside callback will be treated as a composite operation
     * more convenient way without manually invoking `startBatch` and `stopBatch`.
     * @param callback The func containing operations to be batched together.
     */
    batch(callback: () => void): void;
    /**
     * Clear history stack
     * @param {StackType} stackType undo/redo stack
     */
    clearStack(stackType?: StackType): void;
    /**
     * Collapse sub tree(s).
     * @param ids Root id(s) of the sub trees.
     * @param disableAnimate Whether disable the animations for this operation.
     * @param stack Whether push this operation to stack.
     * @returns
     * @group Tree
     */
    collapse(ids: ID | ID[], disableAnimate?: boolean, stack?: boolean): void;
    /**
     * Expand sub tree(s).
     * @param ids Root id(s) of the sub trees.
     * @param disableAnimate Whether disable the animations for this operation.
     * @param stack Whether push this operation to stack.
     * @returns
     * @group Tree
     */
    expand(ids: ID | ID[], disableAnimate?: boolean, stack?: boolean): void;
    /**
     * Destroy the graph instance and remove the related canvases.
     * @returns
     * @group Graph Instance
     */
    destroy(callback?: Function): void;
}
