"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const event_emitter_1 = __importDefault(require("@antv/event-emitter"));
const util_1 = require("@antv/util");
const canvas_1 = require("../util/canvas");
const shape_1 = require("../util/shape");
const controller_1 = require("./controller");
const plugin_1 = require("./controller/plugin");
const hooks_1 = __importDefault(require("./hooks"));
class Graph extends event_emitter_1.default {
    constructor(spec) {
        super();
        this.defaultSpecification = {
            theme: {
                type: 'spec',
                base: 'light',
            },
            enableStack: true,
        };
        /**
         * Execute a callback without allowing any stacking operations.
         * @param callback
         */
        this.executeWithoutStacking = (callback) => {
            const history = this.getHistoryPlugin();
            history === null || history === void 0 ? void 0 : history.pauseStacking();
            try {
                callback();
            }
            finally {
                history === null || history === void 0 ? void 0 : history.resumeStacking();
            }
        };
        this.specification = Object.assign({}, this.defaultSpecification, spec);
        this.initHooks();
        this.initCanvas();
        this.initControllers();
        this.initHistory();
        this.hooks.init.emit({
            canvases: {
                background: this.backgroundCanvas,
                main: this.canvas,
                transient: this.transientCanvas,
            },
        });
        const { data } = spec;
        if (data) {
            // TODO: handle multiple type data configs
            this.read(data);
        }
    }
    /**
     * Initialize the controllers for different plugins.
     */
    initControllers() {
        this.dataController = new controller_1.DataController(this);
        this.interactionController = new controller_1.InteractionController(this);
        this.layoutController = new controller_1.LayoutController(this);
        this.themeController = new controller_1.ThemeController(this);
        this.itemController = new controller_1.ItemController(this);
        this.viewportController = new controller_1.ViewportController(this);
        this.extensionController = new controller_1.ExtensionController(this);
        this.pluginController = new plugin_1.PluginController(this);
    }
    initCanvas() {
        const { renderer, container, canvas, backgroundCanvas, transientCanvas, width, height, } = this.specification;
        let pixelRatio;
        if (renderer && !(0, util_1.isString)(renderer)) {
            // @ts-ignore
            this.rendererType = renderer.type || 'canvas';
            // @ts-ignore
            pixelRatio = renderer.pixelRatio;
        }
        else {
            // @ts-ignore
            this.rendererType = renderer || 'canvas';
        }
        /**
         * These 3 canvases can be passed in by users, e.g. when doing serverside rendering we can't use DOM API.
         */
        if (canvas) {
            this.canvas = canvas;
            this.backgroundCanvas = backgroundCanvas;
            this.transientCanvas = transientCanvas;
            this.container = container;
        }
        else {
            const containerDOM = (0, util_1.isString)(container)
                ? document.getElementById(container)
                : container;
            if (!containerDOM) {
                console.error(`Create graph failed. The container for graph ${containerDOM} is not exist.`);
                this.destroy();
                return;
            }
            this.container = containerDOM;
            const size = [width, height];
            if (size[0] === undefined) {
                size[0] = containerDOM.scrollWidth;
            }
            if (size[1] === undefined) {
                size[1] = containerDOM.scrollHeight;
            }
            this.backgroundCanvas = (0, canvas_1.createCanvas)(this.rendererType, containerDOM, size[0], size[1], pixelRatio);
            this.canvas = (0, canvas_1.createCanvas)(this.rendererType, containerDOM, size[0], size[1], pixelRatio);
            this.transientCanvas = (0, canvas_1.createCanvas)(this.rendererType, containerDOM, size[0], size[1], pixelRatio);
        }
        Promise.all([this.backgroundCanvas, this.canvas, this.transientCanvas].map((canvas) => canvas.ready)).then(() => {
            [this.backgroundCanvas, this.canvas, this.transientCanvas].forEach((canvas, i) => {
                const $domElement = canvas
                    .getContextService()
                    .getDomElement();
                if ($domElement && $domElement.style) {
                    $domElement.style.position = 'fixed';
                    $domElement.style.outline = 'none';
                    $domElement.tabIndex = 1; // Enable keyboard events
                    // Transient canvas should let interactive events go through.
                    if (i === 2) {
                        $domElement.style.pointerEvents = 'none';
                    }
                }
            });
            this.canvasReady = true;
        });
    }
    initHistory() {
        const { enableStack, stackCfg } = this.specification;
        if (enableStack) {
            const history = {
                type: 'history',
                key: 'history',
                enableStack,
                stackCfg,
            };
            if (!this.specification.plugins) {
                this.specification.plugins = [];
            }
            this.specification.plugins.push(history);
        }
    }
    /**
     * Change the renderer at runtime.
     * @param type renderer name
     * @returns
     */
    changeRenderer(type) {
        this.rendererType = type || 'canvas';
        (0, canvas_1.changeRenderer)(this.rendererType, this.canvas);
    }
    /**
     * Initialize the hooks for graph's lifecycles.
     */
    initHooks() {
        this.hooks = {
            init: new hooks_1.default({ name: 'init' }),
            datachange: new hooks_1.default({
                name: 'datachange',
            }),
            itemchange: new hooks_1.default({ name: 'itemchange' }),
            render: new hooks_1.default({
                name: 'render',
            }),
            layout: new hooks_1.default({ name: 'layout' }),
            viewportchange: new hooks_1.default({ name: 'viewport' }),
            modechange: new hooks_1.default({ name: 'modechange' }),
            behaviorchange: new hooks_1.default({ name: 'behaviorchange' }),
            itemstatechange: new hooks_1.default({
                name: 'itemstatechange',
            }),
            itemvisibilitychange: new hooks_1.default({
                name: 'itemvisibilitychange',
            }),
            itemzindexchange: new hooks_1.default({
                name: 'itemzindexchange',
            }),
            transientupdate: new hooks_1.default({ name: 'transientupdate' }),
            pluginchange: new hooks_1.default({ name: 'pluginchange' }),
            themechange: new hooks_1.default({ name: 'init' }),
            treecollapseexpand: new hooks_1.default({ name: 'treecollapseexpand' }),
            destroy: new hooks_1.default({ name: 'destroy' }),
        };
    }
    /**
     * Update the specs(configurations).
     */
    updateSpecification(spec) {
        return Object.assign(this.specification, spec);
    }
    /**
     * Update the theme specs (configurations).
     */
    updateTheme(theme) {
        this.specification.theme = theme;
        // const { specification } = this.themeController;
        // notifiying the themeController
        this.hooks.themechange.emit({
            canvases: {
                background: this.backgroundCanvas,
                main: this.canvas,
                transient: this.transientCanvas,
            },
        });
        // theme is formatted by themeController, notify the item controller to update the items
        this.hooks.themechange.emit({
            theme: this.themeController.specification,
        });
    }
    /**
     * Get the copy of specs(configurations).
     * @returns graph specs
     */
    getSpecification() {
        return this.specification;
    }
    /**
     * Input data and render the graph.
     * If there is old data, diffs and changes it.
     * @param data
     * @returns
     * @group Data
     */
    read(data) {
        return __awaiter(this, void 0, void 0, function* () {
            this.hooks.datachange.emit({ data, type: 'replace' });
            const emitRender = () => __awaiter(this, void 0, void 0, function* () {
                this.hooks.render.emit({
                    graphCore: this.dataController.graphCore,
                    theme: this.themeController.specification,
                    transientCanvas: this.transientCanvas,
                });
                this.emit('afterrender');
                this.once('afterlayout', () => __awaiter(this, void 0, void 0, function* () {
                    const { autoFit } = this.specification;
                    if (autoFit) {
                        if (autoFit === 'view') {
                            yield this.fitView();
                        }
                        else if (autoFit === 'center') {
                            yield this.fitCenter();
                        }
                        else {
                            const { type, effectTiming } = autoFit, others = __rest(autoFit, ["type", "effectTiming"]);
                            if (type === 'view') {
                                yield this.fitView(others, effectTiming);
                            }
                            else if (type === 'center') {
                                yield this.fitCenter(effectTiming);
                            }
                            else if (type === 'position') {
                                // TODO: align
                                yield this.translateTo(others.position, effectTiming);
                            }
                        }
                    }
                }));
                yield this.layout();
            });
            if (this.canvasReady) {
                yield emitRender();
            }
            else {
                yield Promise.all([this.backgroundCanvas, this.canvas, this.transientCanvas].map((canvas) => canvas.ready));
                yield emitRender();
            }
        });
    }
    /**
     * Change graph data.
     * @param data new data
     * @param type the way to change data, 'replace' means discard the old data and use the new one; 'mergeReplace' means merge the common part, remove (old - new), add (new - old)
     * @returns
     * @group Data
     */
    changeData(data, type = 'mergeReplace') {
        return __awaiter(this, void 0, void 0, function* () {
            this.hooks.datachange.emit({ data, type });
            this.hooks.render.emit({
                graphCore: this.dataController.graphCore,
                theme: this.themeController.specification,
                transientCanvas: this.transientCanvas,
            });
            this.emit('afterrender');
            yield this.layout();
        });
    }
    /**
     * Clear the graph, means remove all the items on the graph.
     * @returns
     */
    clear() {
        this.startBatch();
        this.removeData('edge', this.getAllEdgesData().map((edge) => edge.id));
        this.removeData('node', this.getAllNodesData().map((node) => node.id));
        this.removeData('combo', this.getAllCombosData().map((combo) => combo.id));
        this.stopBatch();
    }
    getViewportCenter() {
        const { width, height } = this.canvas.getConfig();
        return { x: width / 2, y: height / 2 };
    }
    transform(options, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.hooks.viewportchange.emitLinearAsync({
                transform: options,
                effectTiming,
            });
            this.emit('viewportchange', options);
        });
    }
    /**
     * Stop the current transition of transform immediately.
     */
    stopTransformTransition() {
        this.canvas.getCamera().cancelLandmarkAnimation();
    }
    /**
     * Move the graph with a relative distance under viewport coordinates.
     * @param dx x of the relative distance
     * @param dy y of the relative distance
     * @param effectTiming animation configurations
     */
    translate(distance, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.transform({
                translate: distance,
            }, effectTiming);
        });
    }
    /**
     * Move the graph to destination under viewport coordinates.
     * @param destination destination under viewport coordinates.
     * @param effectTiming animation configurations
     */
    translateTo({ x, y }, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            const { x: cx, y: cy } = this.getViewportCenter();
            yield this.translate({ dx: cx - x, dy: cy - y }, effectTiming);
        });
    }
    /**
     * Zoom the graph with a relative ratio.
     * @param ratio relative ratio to zoom
     * @param origin origin under viewport coordinates.
     * @param effectTiming animation configurations
     */
    zoom(ratio, origin, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.transform({
                zoom: {
                    ratio,
                },
                origin,
            }, effectTiming);
        });
    }
    /**
     * Zoom the graph to a specified ratio.
     * @param zoom specified ratio
     * @param origin zoom center
     * @param effectTiming animation configurations
     */
    zoomTo(zoom, origin, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.zoom(zoom / this.canvas.getCamera().getZoom(), origin, effectTiming);
        });
    }
    /**
     * Return the current zoom level of camera.
     * @returns current zoom
     */
    getZoom() {
        return this.canvas.getCamera().getZoom();
    }
    /**
     * Rotate the graph with a relative angle.
     * @param angle
     * @param origin
     * @param effectTiming
     */
    rotate(angle, origin, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.transform({
                rotate: {
                    angle,
                },
                origin,
            }, effectTiming);
        });
    }
    /**
     * Rotate the graph to an absolute angle.
     * @param angle
     * @param origin
     * @param effectTiming
     */
    rotateTo(angle, origin, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.rotate(angle - this.canvas.getCamera().getRoll(), origin, effectTiming);
        });
    }
    /**
     * Fit the graph content to the view.
     * @param options.padding padding while fitting
     * @param options.rules rules for fitting
     * @param effectTiming animation configurations
     */
    fitView(options, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            const { padding, rules } = options || {};
            const [top, right, bottom, left] = padding
                ? (0, shape_1.formatPadding)(padding)
                : [0, 0, 0, 0];
            const { direction = 'both', ratioRule = 'min' } = rules || {};
            // Get the bounds of the whole graph.
            const { center: [graphCenterX, graphCenterY], halfExtents, } = this.canvas.document.documentElement.getBounds();
            const origin = this.canvas.canvas2Viewport({
                x: graphCenterX,
                y: graphCenterY,
            });
            const { width: viewportWidth, height: viewportHeight } = this.canvas.getConfig();
            const graphWidth = halfExtents[0] * 2;
            const graphHeight = halfExtents[1] * 2;
            const tlInCanvas = this.canvas.viewport2Canvas({ x: left, y: top });
            const brInCanvas = this.canvas.viewport2Canvas({
                x: viewportWidth - right,
                y: viewportHeight - bottom,
            });
            const targetViewWidth = brInCanvas.x - tlInCanvas.x;
            const targetViewHeight = brInCanvas.y - tlInCanvas.y;
            const wRatio = targetViewWidth / graphWidth;
            const hRatio = targetViewHeight / graphHeight;
            let ratio;
            if (direction === 'x') {
                ratio = wRatio;
            }
            else if (direction === 'y') {
                ratio = hRatio;
            }
            else {
                ratio =
                    ratioRule === 'max'
                        ? Math.max(wRatio, hRatio)
                        : Math.min(wRatio, hRatio);
            }
            yield this.transform({
                translate: {
                    dx: viewportWidth / 2 - origin.x,
                    dy: viewportHeight / 2 - origin.y,
                },
                zoom: {
                    ratio,
                },
            }, effectTiming);
        });
    }
    /**
     * Fit the graph center to the view center.
     * @param effectTiming animation configurations
     */
    fitCenter(effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            // Get the bounds of the whole graph.
            const { center: [graphCenterX, graphCenterY], } = this.canvas.document.documentElement.getBounds();
            yield this.translateTo(this.canvas.canvas2Viewport({ x: graphCenterX, y: graphCenterY }), effectTiming);
        });
    }
    /**
     * Move the graph to make the item align the view center.
     * @param item node/edge/combo item or its id
     * @param effectTiming animation configurations
     */
    focusItem(id, effectTiming) {
        return __awaiter(this, void 0, void 0, function* () {
            let bounds = null;
            for (const itemId of !(0, util_1.isArray)(id) ? [id] : id) {
                const item = this.getItemById(itemId);
                if (item) {
                    const itemBounds = item.group.getBounds();
                    if (!bounds) {
                        bounds = itemBounds;
                    }
                    else {
                        bounds.add(itemBounds);
                    }
                }
            }
            if (bounds) {
                const { center: [itemCenterX, itemCenterY], } = bounds;
                yield this.translateTo(this.canvas.canvas2Viewport({ x: itemCenterX, y: itemCenterY }), effectTiming);
            }
        });
    }
    /**
     * Get item by id. We don't want to
     * @param id
     * @returns Node | Edge | Combo
     */
    getItemById(id) {
        return this.itemController.getItemById(id);
    }
    /**
     * Get the size of the graph canvas.
     * @returns [width, height]
     * @group View
     */
    getSize() {
        const { width = 500, height = 500 } = this.specification;
        return [width, height];
    }
    /**
     * Set the size for the graph canvas.
     * @param number[] [width, height]
     * @group View
     */
    setSize(size) {
        if (!(0, util_1.isArray)(size) || size.length < 2) {
            console.warn(`Failed to setSize. The parameter size: ${size} is invalid. It must be an array with 2 number elements.`);
            return;
        }
        this.specification.width = size[0];
        this.specification.height = size[1];
        this.canvas.resize(size[0], size[1]);
    }
    /**
     * Get the rendering coordinate according to the canvas dom (viewport) coordinate.
     * @param Point rendering coordinate
     * @returns canvas dom (viewport) coordinate
     * @group View
     */
    getCanvasByViewport(viewportPoint) {
        return this.canvas.viewport2Canvas(viewportPoint);
    }
    /**
     * Get the canvas dom (viewport) coordinate according to the rendering coordinate.
     * @param Point canvas dom (viewport) coordinate
     * @returns rendering coordinate
     * @group View
     */
    getViewportByCanvas(canvasPoint) {
        return this.canvas.canvas2Viewport(canvasPoint);
    }
    /**
     * Get the browser coordinate according to the rendering coordinate.
     * @param Point rendering coordinate
     * @returns browser coordinate
     * @group View
     */
    getClientByCanvas(canvasPoint) {
        const viewportPoint = this.canvas.canvas2Viewport(canvasPoint);
        return this.canvas.viewport2Client(viewportPoint);
    }
    /**
     * Get the rendering coordinate according to the browser coordinate.
     * @param Point browser coordinate
     * @returns rendering coordinate
     * @group View
     */
    getCanvasByClient(clientPoint) {
        const viewportPoint = this.canvas.client2Viewport(clientPoint);
        return this.canvas.viewport2Canvas(viewportPoint);
    }
    // ===== item operations =====
    /**
     * Find a node's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result node's inner data
     * @group Data
     */
    getNodeData(condition) {
        var _a;
        const conds = (0, util_1.isString)(condition) || (0, util_1.isNumber)(condition) ? [condition] : condition;
        return (_a = this.dataController.findData('node', conds)) === null || _a === void 0 ? void 0 : _a[0];
    }
    /**
     * Find an edge's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result edge's inner data
     * @group Data
     */
    getEdgeData(condition) {
        var _a;
        const conds = (0, util_1.isString)(condition) || (0, util_1.isNumber)(condition) || (0, util_1.isNumber)(condition)
            ? [condition]
            : condition;
        return (_a = this.dataController.findData('edge', conds)) === null || _a === void 0 ? void 0 : _a[0];
    }
    /**
     * Find an combo's inner data according to id or function.
     * @param { ID | Function } condition id or condition function
     * @returns result combo's inner data
     * @group Data
     */
    getComboData(condition) {
        var _a;
        const conds = (0, util_1.isString)(condition) || (0, util_1.isNumber)(condition) ? [condition] : condition;
        return (_a = this.dataController.findData('combo', conds)) === null || _a === void 0 ? void 0 : _a[0];
    }
    /**
     * Get all the nodes' inner data
     * @returns all nodes' inner data on the graph
     * @group Data
     */
    getAllNodesData() {
        return this.dataController.findAllData('node');
    }
    /**
     * Get all the edges' inner data
     * @returns all edges' inner data on the graph
     * @group Data
     */
    getAllEdgesData() {
        return this.dataController.findAllData('edge');
    }
    /**
     * Get all the combos' inner data
     * @returns all combos' inner data on the graph
     * @group Data
     */
    getAllCombosData() {
        return this.dataController.findAllData('combo');
    }
    /**
     * Get one-hop edge ids from a start node.
     * @param nodeId id of the start node
     * @returns one-hop edges' data array
     * @group Data
     */
    getRelatedEdgesData(nodeId, direction = 'both') {
        return this.dataController.findRelatedEdges(nodeId, direction);
    }
    /**
     * Get one-hop node ids from a start node.
     * @param nodeId id of the start node
     * @returns one-hop nodes' data array
     * @group Data
     */
    getNeighborNodesData(nodeId, direction = 'both') {
        return this.dataController.findNeighborNodes(nodeId, direction);
    }
    /**
     * Get the children's data of a combo.
     * @param comboId combo id
     * @returns children's data array
     * @group Data
     */
    getComboChildrenData(comboId) {
        return this.dataController.findChildren(comboId, 'combo');
    }
    /*
     * Get the display model of a node / edge / combo.
     * @param id item id
     * @returns display model
     * @group Data
     */
    getDisplayModel(id) {
        return this.itemController.findDisplayModel(id);
    }
    /**
     * Retrieve the nearby edges for a given node using quadtree collision detection.
     * @param nodeId node id
     * @group Data
     */
    getNearEdgesForNode(nodeId) {
        const { graphCore } = this.dataController;
        return this.itemController.findNearEdgesByNode(nodeId, graphCore);
    }
    /**
     * Find items which has the state.
     * @param itemType item type
     * @param state state name
     * @param additionalFilter additional filter function
     * @returns items that is the type and has the state
     * @group Item
     */
    findIdByState(itemType, state, value = true, additionalFilter) {
        let ids = this.itemController.findIdByState(itemType, state, value);
        if (additionalFilter) {
            let getDataFunc = this.getEdgeData;
            if (itemType === 'node')
                getDataFunc = this.getNodeData;
            else if (itemType === 'combo')
                getDataFunc = this.getComboData;
            ids = ids.filter((id) => additionalFilter(getDataFunc(id)));
        }
        return ids;
    }
    /**
     * Add one or more node/edge/combo data to the graph.
     * @param itemType item type
     * @param model user data
     * @returns whether success
     * @group Data
     */
    addData(itemType, models) {
        // data controller and item controller subscribe additem in order
        const { graphCore } = this.dataController;
        const { specification } = this.themeController;
        graphCore.once('changed', (event) => {
            if (!event.changes.length)
                return;
            const changes = event.changes;
            this.hooks.itemchange.emit({
                type: itemType,
                changes: graphCore.reduceChanges(event.changes),
                graphCore,
                theme: specification,
            });
            this.emit('afteritemchange', {
                type: itemType,
                action: 'add',
                models,
                apiName: 'addData',
                changes,
            });
        });
        const modelArr = (0, util_1.isArray)(models) ? models : [models];
        const data = { nodes: [], edges: [], combos: [] };
        data[`${itemType}s`] = modelArr;
        this.hooks.datachange.emit({
            data,
            type: 'union',
        });
        const dataList = this.dataController.findData(itemType, modelArr.map((model) => model.id));
        return (0, util_1.isArray)(models) ? dataList : dataList[0];
    }
    /**
     * Remove one or more node/edge/combo data from the graph.
     * @param item the item to be removed
     * @returns whether success
     * @group Data
     */
    removeData(itemType, ids) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        const data = { nodes: [], edges: [], combos: [] };
        const { userGraphCore, graphCore } = this.dataController;
        const { specification } = this.themeController;
        const getItem = itemType === 'edge' ? userGraphCore.getEdge : userGraphCore.getNode;
        data[`${itemType}s`] = idArr.map((id) => getItem.bind(userGraphCore)(id));
        graphCore.once('changed', (event) => {
            if (!event.changes.length)
                return;
            const changes = event.changes;
            this.hooks.itemchange.emit({
                type: itemType,
                changes: event.changes,
                graphCore,
                theme: specification,
            });
            this.emit('afteritemchange', {
                type: itemType,
                action: 'remove',
                ids: idArr,
                apiName: 'removeData',
                changes,
            });
        });
        this.hooks.datachange.emit({
            data,
            type: 'remove',
        });
    }
    /**
     * Extends the fields of oldValue and newValue
     * to make sure they both contain the union set of their original fields.
     * The missing fields' values are filled using data from displayModel data.
     * @param changes
     * @returns
     */
    extendChanges(changes) {
        return changes
            .map((change) => {
            const { id, propertyName, oldValue, newValue } = change;
            if (!oldValue || !newValue)
                return change;
            if (propertyName) {
                if (oldValue !== newValue)
                    return change;
                return false;
            }
            const displayData = this.getItemById(id).displayModel.data;
            const oldFields = Object.keys(oldValue);
            const newFields = Object.keys(newValue);
            const commonFields = [...new Set([...oldFields, ...newFields])];
            commonFields.forEach((field) => {
                if (!Object.prototype.hasOwnProperty.call(oldValue, field)) {
                    oldValue[field] = displayData[field];
                }
                if (!Object.prototype.hasOwnProperty.call(newValue, field)) {
                    newValue[field] = displayData[field];
                }
            });
            if ((oldValue.x === undefined || Number.isNaN(oldValue.x)) &&
                !oldValue._isCombo)
                oldValue.x = 0;
            if ((oldValue.y === undefined || Number.isNaN(oldValue.x)) &&
                !oldValue._isCombo)
                oldValue.y = 0;
            if (Number.isNaN(newValue.x))
                delete newValue.x;
            if (Number.isNaN(newValue.y))
                delete newValue.y;
            if ((0, util_1.isEqual)(newValue, oldValue))
                return false;
            return Object.assign(Object.assign({}, change), { newValue,
                oldValue });
        })
            .filter(Boolean);
    }
    /**
     * Update one or more node/edge/combo data on the graph.
     * @param {ITEM_TYPE} itemType 'node' | 'edge' | 'combo'
     * @param models new configurations for every node/edge/combo, which has id field to indicate the specific item
     * @group Data
     */
    updateData(itemType, models) {
        const modelArr = (0, util_1.isArray)(models) ? models : [models];
        const data = { nodes: [], edges: [], combos: [] };
        data[`${itemType}s`] = modelArr;
        const { graphCore } = this.dataController;
        const { specification } = this.themeController;
        graphCore.once('changed', (event) => {
            const changes = this.extendChanges((0, util_1.clone)(event.changes));
            this.hooks.itemchange.emit({
                type: itemType,
                changes: event.changes,
                graphCore,
                theme: specification,
            });
            this.emit('afteritemchange', {
                type: itemType,
                action: 'update',
                models,
                apiName: 'updateData',
                changes,
            });
        });
        this.hooks.datachange.emit({
            data,
            type: 'update',
        });
        const dataList = this.dataController.findData(itemType, modelArr.map((model) => model.id));
        return (0, util_1.isArray)(models) ? dataList : dataList[0];
    }
    /**
     * Update one or more nodes' positions,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * @param models new configurations with x and y for every node, which has id field to indicate the specific item
     * @group Data
     */
    updateNodePosition(models, upsertAncestors, disableAnimate = false, callback, stack) {
        return this.updatePosition('node', models, upsertAncestors, disableAnimate, callback, stack);
    }
    /**
     * Update one or more combos' positions,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * In fact, it changes the succeed nodes positions to affect the combo's position, but not modify the combo's position directly.
     * @param models new configurations with x and y for every combo, which has id field to indicate the specific item
     * @group Data
     */
    updateComboPosition(models, upsertAncestors, disableAnimate = false, callback, stack) {
        return this.updatePosition('combo', models, upsertAncestors, disableAnimate, callback, stack);
    }
    /**
     * Get history plugin instance
     */
    getHistoryPlugin() {
        if (!this.specification.enableStack ||
            !this.pluginController.hasPlugin('history')) {
            console.error('The history plugin is not loaded or initialized.');
            return;
        }
        return this.pluginController.getPlugin('history');
    }
    updatePosition(type, models, upsertAncestors, disableAnimate = false, callback, stack) {
        const modelArr = (0, util_1.isArray)(models) ? models : [models];
        const { graphCore } = this.dataController;
        const { specification } = this.themeController;
        graphCore.once('changed', (event) => {
            if (!event.changes.length)
                return;
            const changes = event.changes.filter((change) => !(0, util_1.isEqual)(change.newValue, change.oldValue));
            this.hooks.itemchange.emit({
                type,
                changes: event.changes,
                graphCore,
                theme: specification,
                upsertAncestors,
                action: 'updatePosition',
                animate: !disableAnimate,
                callback,
            });
            this.emit('afteritemchange', {
                type,
                action: 'updatePosition',
                upsertAncestors,
                models,
                apiName: 'updatePosition',
                changes,
            });
        });
        this.hooks.datachange.emit({
            data: {
                nodes: type === 'node' ? modelArr : [],
                edges: [],
                combos: type === 'combo' ? modelArr : [],
            },
            type: 'update',
        });
        const dataList = this.dataController.findData(type, modelArr.map((model) => model.id));
        return (0, util_1.isArray)(models) ? dataList : dataList[0];
    }
    getItemPreviousVisibility(ids) {
        const objs = ids.map((id) => ({ id, visible: this.getItemVisible(id) }));
        const groupedByVisible = (0, util_1.groupBy)(objs, 'visible');
        const values = [];
        for (const visible in groupedByVisible) {
            values.push({
                ids: (0, util_1.map)(groupedByVisible[visible], (item) => item.id),
                visible: visible === 'true',
            });
        }
        return values;
    }
    /**
     * Show the item(s).
     * @param item the item to be shown
     * @returns
     * @group Item
     */
    showItem(ids, disableAnimate = false) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        if ((0, util_1.isEmpty)(idArr))
            return;
        const changes = {
            newValue: [{ ids: idArr, visible: true }],
            oldValue: this.getItemPreviousVisibility(idArr),
        };
        this.hooks.itemvisibilitychange.emit({
            ids: idArr,
            value: true,
            graphCore: this.dataController.graphCore,
            animate: !disableAnimate,
        });
        this.emit('afteritemvisibilitychange', {
            ids,
            value: true,
            animate: !disableAnimate,
            action: 'updateVisibility',
            apiName: 'showItem',
            changes,
        });
    }
    /**
     * Hide the item(s).
     * @param item the item to be hidden
     * @returns
     * @group Item
     */
    hideItem(ids, disableAnimate = false) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        if ((0, util_1.isEmpty)(idArr))
            return;
        const changes = {
            newValue: [{ ids: idArr, visible: false }],
            oldValue: this.getItemPreviousVisibility(idArr),
        };
        this.hooks.itemvisibilitychange.emit({
            ids: idArr,
            value: false,
            graphCore: this.dataController.graphCore,
            animate: !disableAnimate,
        });
        this.emit('afteritemvisibilitychange', {
            ids,
            value: false,
            animate: !disableAnimate,
            action: 'updateVisibility',
            apiName: 'hideItem',
            changes,
        });
    }
    /**
     * Make the item(s) to the front.
     * @param ids
     * @returns
     * @group Item
     */
    frontItem(ids) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        this.hooks.itemzindexchange.emit({
            ids: idArr,
            action: 'front',
            graphCore: this.dataController.graphCore,
        });
        this.emit('afteritemzindexchange', {
            ids: idArr,
            action: 'front',
            apiName: 'frontItem',
        });
    }
    /**
     * Make the item(s) to the back.
     * @param ids
     * @returns
     * @group Item
     */
    backItem(ids) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        this.hooks.itemzindexchange.emit({
            ids: idArr,
            action: 'back',
            graphCore: this.dataController.graphCore,
        });
        this.emit('afteritemzindexchange', {
            ids: idArr,
            action: 'back',
            apiName: 'backItem',
        });
    }
    getItemPreviousStates(stateOption) {
        const { ids, states } = stateOption;
        const idArr = Array.isArray(ids) ? ids : [ids];
        const stateArr = Array.isArray(states) ? states : [states];
        if ((0, util_1.isEmpty)(idArr))
            return;
        return idArr.flatMap((id) => {
            return stateArr.map((state) => ({
                ids: id,
                states: state,
                value: this.getItemState(id, state),
            }));
        });
    }
    /**
     * Set state for the item.
     * @param item the item to be set
     * @param state the state name
     * @param value state value
     * @returns
     * @group Item
     */
    setItemState(ids, states, value) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        const stateArr = (0, util_1.isArray)(states) ? states : [states];
        const stateOption = { ids: idArr, states: stateArr, value };
        const changes = {
            newValue: [stateOption],
            oldValue: this.getItemPreviousStates(stateOption),
        };
        this.hooks.itemstatechange.emit({
            ids: idArr,
            states: stateArr,
            value,
        });
        this.emit('afteritemstatechange', {
            ids,
            states,
            value,
            action: 'updateState',
            apiName: 'setItemState',
            changes,
        });
    }
    /**
     * Get the state value for an item.
     * @param id the id for the item
     * @param states the state name
     * @returns {boolean} the state value
     * @group Item
     */
    getItemState(id, state) {
        return this.itemController.getItemState(id, state);
    }
    /**
     * Get all the state names with value true for an item.
     * @param id the id for the item
     * @returns {string[]} the state names with value true
     * @group Item
     */
    getItemAllStates(id) {
        return this.itemController.getItemAllStates(id);
    }
    /**
     * Clear all the states for item(s).
     * @param ids the id(s) for the item(s) to be clear
     * @param states the states' names, all the states wil be cleared if states is not assigned
     * @returns
     * @group Item
     */
    clearItemState(ids, states) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        const stateOptions = { ids: idArr, states, value: false };
        const changes = {
            newValue: [stateOptions],
            oldValue: this.getItemPreviousStates(stateOptions),
        };
        this.hooks.itemstatechange.emit({
            ids: idArr,
            states,
            value: false,
        });
        this.emit('afteritemstatechange', {
            ids,
            states,
            value: false,
            action: 'updateState',
            apiName: 'clearItemState',
            changes,
        });
    }
    /**
     * Get the rendering bbox for a node / edge / combo, or the graph (when the id is not assigned).
     * @param id the id for the node / edge / combo, undefined for the whole graph
     * @returns rendering bounding box. returns false if the item is not exist
     * @group Item
     */
    getRenderBBox(id, onlyKeyShape = false, isTransient = false) {
        if (!id)
            return this.canvas.getRoot().getRenderBounds();
        return this.itemController.getItemBBox(id, onlyKeyShape, isTransient);
    }
    /**
     * Get the visibility for a node / edge / combo.
     * @param id the id for the node / edge / combo
     * @returns visibility for the item, false for invisible or unexistence for the item
     * @group Item
     */
    getItemVisible(id) {
        return this.itemController.getItemVisible(id);
    }
    // ===== combo operations =====
    /**
     * Add a new combo to the graph, and update the structure of the existed child in childrenIds to be the children of the new combo.
     * Different from addData with combo type, this API update the succeeds' combo tree strucutres in the same time.
     * @param model combo user data
     * @returns whether success
     * @group Combo
     */
    addCombo(model, childrenIds) {
        const { graphCore } = this.dataController;
        const { specification } = this.themeController;
        graphCore.once('changed', (event) => {
            if (!event.changes.length)
                return;
            const changes = event.changes;
            this.hooks.itemchange.emit({
                type: 'combo',
                changes: graphCore.reduceChanges(event.changes),
                graphCore,
                theme: specification,
            });
            this.emit('afteritemchange', {
                type: 'combo',
                action: 'add',
                models: [model],
                apiName: 'addCombo',
                changes,
            });
        });
        const data = {
            nodes: [],
            edges: [],
            combos: [
                Object.assign(Object.assign({}, model), { data: Object.assign(Object.assign({}, model.data), { _children: childrenIds }) }),
            ],
        };
        this.hooks.datachange.emit({
            data,
            type: 'addCombo',
        });
        return this.dataController.findData('combo', [model.id])[0];
    }
    /**
     * Collapse a combo.
     * @param comboId combo ids
     * @group Combo
     */
    collapseCombo(comboIds) {
        const ids = (0, util_1.isArray)(comboIds) ? comboIds : [comboIds];
        this.executeWithoutStacking(() => {
            this.updateData('combo', ids.map((id) => ({ id, data: { collapsed: true } })));
        });
        this.emit('aftercollapsecombo', {
            type: 'combo',
            action: 'collapseCombo',
            ids: comboIds,
            apiName: 'collapseCombo',
        });
    }
    /**
     * Expand a combo.
     * @param combo combo ids
     * @group Combo
     */
    expandCombo(comboIds) {
        const ids = (0, util_1.isArray)(comboIds) ? comboIds : [comboIds];
        this.executeWithoutStacking(() => {
            this.updateData('combo', ids.map((id) => ({ id, data: { collapsed: false } })));
        });
        this.emit('afterexpandcombo', {
            type: 'combo',
            action: 'expandCombo',
            ids: comboIds,
            apiName: 'expandCombo',
        });
    }
    /**
     * Move one or more combos a distance (dx, dy) relatively,
     * do not update other styles which leads to better performance than updating positions by updateData.
     * In fact, it changes the succeed nodes positions to affect the combo's position, but not modify the combo's position directly.
     * @param models new configurations with x and y for every combo, which has id field to indicate the specific item
     * @group Combo
     */
    moveCombo(ids, dx, dy, upsertAncestors, callback) {
        const idArr = (0, util_1.isArray)(ids) ? ids : [ids];
        const { graphCore } = this.dataController;
        const { specification } = this.themeController;
        graphCore.once('changed', (event) => {
            if (!event.changes.length)
                return;
            const changes = this.extendChanges((0, util_1.clone)(event.changes));
            this.hooks.itemchange.emit({
                type: 'combo',
                changes: event.changes,
                graphCore,
                theme: specification,
                upsertAncestors,
                action: 'updatePosition',
                callback,
            });
            this.emit('afteritemchange', {
                type: 'combo',
                ids: idArr,
                dx,
                dy,
                action: 'updatePosition',
                upsertAncestors,
                apiName: 'moveCombo',
                changes,
            });
        });
        this.hooks.datachange.emit({
            data: {
                nodes: [],
                edges: [],
                combos: idArr.map((id) => ({ id, data: { dx, dy } })),
            },
            type: 'moveCombo',
        });
        return this.dataController.findData('combo', idArr);
    }
    // ===== layout =====
    /**
     * Layout the graph (with current configurations if cfg is not assigned).
     */
    layout(options, disableAnimate = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('beforelayout');
            const { graphCore } = this.dataController;
            const formattedOptions = Object.assign(Object.assign({}, this.getSpecification().layout), options);
            this.updateSpecification({ layout: formattedOptions });
            const layoutUnset = (!options && !this.getSpecification().layout) ||
                !Object.keys(formattedOptions).length;
            if (layoutUnset) {
                const nodes = graphCore.getAllNodes();
                if (nodes.every((node) => (0, util_1.isNil)(node.data.x) && (0, util_1.isNil)(node.data.y))) {
                    // Use `grid` layout as default when x/y of each node is unset.
                    formattedOptions.type = 'grid';
                }
                else {
                    // Use user-defined position(x/y default to 0).
                    formattedOptions.execute = (graph) => __awaiter(this, void 0, void 0, function* () {
                        const nodes = graph.getAllNodes();
                        return {
                            nodes: nodes.map((node) => ({
                                id: node.id,
                                data: {
                                    x: Number(node.data.x) || 0,
                                    y: Number(node.data.y) || 0,
                                },
                            })),
                            edges: [],
                        };
                    });
                }
            }
            yield this.hooks.layout.emitLinearAsync({
                graphCore,
                options: formattedOptions,
                animate: !disableAnimate,
            });
            this.emit('afterlayout');
        });
    }
    /**
     * Some layout algorithms has many iterations which can be stopped at any time.
     */
    stopLayout() {
        this.layoutController.stopLayout();
    }
    /**
     *
     * @returns
     */
    getLayoutCurrentAnimation() {
        return this.layoutController.getCurrentAnimation();
    }
    /**
     * Switch mode.
     * @param mode mode name
     * @returns
     * @group Interaction
     */
    setMode(mode) {
        this.hooks.modechange.emit({ mode });
    }
    /**
     * Get current mode.
     * @returns mode name
     * @group Interaction
     */
    getMode() {
        return this.interactionController.getMode();
    }
    /**
     * Add behavior(s) to mode(s).
     * @param behaviors behavior names or configs
     * @param modes mode names
     * @returns
     * @group Interaction
     */
    addBehaviors(behaviors, modes) {
        const modesArr = (0, util_1.isArray)(modes) ? modes : [modes];
        const behaviorsArr = (0, util_1.isArray)(behaviors) ? behaviors : [behaviors];
        this.hooks.behaviorchange.emit({
            action: 'add',
            modes: modesArr,
            // TODO: update type define.
            // @ts-ignore
            behaviors: behaviorsArr,
        });
        // update the graph specification
        modesArr.forEach((mode) => {
            this.specification.modes[mode] =
                this.specification.modes[mode].concat(behaviorsArr);
        });
    }
    /**
     * Remove behavior(s) from mode(s).
     * @param behaviors behavior configs with unique key
     * @param modes mode names
     * @returns
     * @group Interaction
     */
    removeBehaviors(behaviorKeys, modes) {
        const modesArr = (0, util_1.isArray)(modes) ? modes : [modes];
        this.hooks.behaviorchange.emit({
            action: 'remove',
            modes: modesArr,
            behaviors: behaviorKeys,
        });
        // update the graph specification
        modesArr.forEach((mode) => {
            behaviorKeys.forEach((key) => {
                const oldBehavior = this.specification.modes[mode].find((behavior) => (0, util_1.isObject)(behavior) && behavior.key === key);
                const indexOfOldBehavior = this.specification.modes[mode].indexOf(oldBehavior);
                this.specification.modes[mode].splice(indexOfOldBehavior, 1);
            });
        });
    }
    /**
     * Update a behavior on a mode.
     * @param behavior behavior configs, whose name indicates the behavior to be updated
     * @param mode mode name
     * @returns
     * @group Interaction
     */
    updateBehavior(behavior, mode) {
        this.hooks.behaviorchange.emit({
            action: 'update',
            modes: [mode],
            // TODO: update type define.
            // @ts-ignore
            behaviors: [behavior],
        });
        // no need to update specification since the corresponding part is the same object as the behavior's option
        // this.specification.modes[mode].forEach((b, i) => {
        //   if (isObject(b) && b.key === behavior.key) {
        //     this.specification.modes[mode][i] = behavior;
        //   }
        // });
    }
    /**
     * Add plugin(s) to graph.
     * @param pluginCfgs
     * @returns
     * @group Plugin
     */
    addPlugins(pluginCfgs) {
        const pluginsArr = (0, util_1.isArray)(pluginCfgs) ? pluginCfgs : [pluginCfgs];
        // update the graph specification
        if (!this.specification.plugins)
            this.specification.plugins = [];
        const oldPlugins = this.specification.plugins;
        const validPlugins = [];
        pluginsArr.forEach((config) => {
            const oldPlugin = oldPlugins.find((oldPlugin) => {
                if (typeof oldPlugin === 'string' || typeof config === 'string')
                    return false;
                return oldPlugin.key === config.key;
            });
            if (oldPlugin) {
                console.warn(`Add plugin with key ${config.key} failed, the key is duplicated to the existing plugins on the graph.`);
                return;
            }
            validPlugins.push(config);
        });
        this.specification.plugins = oldPlugins.concat(validPlugins);
        this.hooks.pluginchange.emit({
            action: 'add',
            plugins: validPlugins,
        });
    }
    /**
     * Remove plugin(s) from graph.
     * @param pluginKeys
     * @returns
     * @group Plugin
     */
    removePlugins(pluginKeys) {
        const pluginArr = (0, util_1.isArray)(pluginKeys) ? pluginKeys : [pluginKeys];
        this.hooks.pluginchange.emit({
            action: 'remove',
            plugins: pluginArr,
        });
        // update the graph specification
        const { plugins } = this.specification;
        this.specification.plugins = plugins === null || plugins === void 0 ? void 0 : plugins.filter((plugin) => {
            if ((0, util_1.isObject)(plugin))
                return !(pluginArr.includes(plugin.key) ||
                    pluginArr.includes(plugin));
            return !pluginArr.includes(plugin);
        });
    }
    /**
     * Update a plugin of the graph.
     * @param plugin plugin configs, whose key indicates the behavior to be updated
     * @returns
     * @group Interaction
     */
    updatePlugin(plugin) {
        const { plugins } = this.specification;
        const { key } = plugin;
        if (!key) {
            console.warn('Update plugin failed, the key for the plugin to be updated should be assign.');
            return;
        }
        if (!plugins) {
            console.warn('Update plugin failed, the plugin to be updated does not exist.');
            return;
        }
        const oldPlugin = plugins === null || plugins === void 0 ? void 0 : plugins.find((p) => {
            if (typeof p === 'string')
                return p === key;
            return p.key === key;
        });
        if (!oldPlugin) {
            console.warn('Update plugin failed, the key for the plugin to be updated should be assign.');
            return;
        }
        const idx = plugins.indexOf(oldPlugin);
        plugins[idx] = Object.assign(Object.assign({}, oldPlugin), plugin);
        this.hooks.pluginchange.emit({
            action: 'update',
            plugins: [plugin],
        });
    }
    /**
     * Draw or update a G shape or group to the transient canvas.
     * @param type shape type or item type
     * @param id new shape id or updated shape id for a interation shape, node/edge/combo id for item interaction group drawing
     * @returns upserted shape or group
     * @group Interaction
     */
    drawTransient(type, id, config, canvas) {
        this.hooks.transientupdate.emit({
            type,
            id,
            config,
            canvas: canvas || this.transientCanvas,
            graphCore: this.dataController.graphCore,
        });
        return this.itemController.getTransient(String(id));
    }
    // ===== history operations =====
    /**
     * Determine if history (redo/undo) is enabled.
     */
    isHistoryEnabled() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.isEnable();
    }
    /**
     * Push the operation(s) onto the specified stack
     * @param cmd commands to be pushed
     * @param stackType undo/redo stack
     * @param isNew
     */
    pushStack(cmd, stackType, isNew) {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.push(cmd, stackType, isNew);
    }
    /**
     * Pause stacking operation.
     */
    pauseStacking() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.pauseStacking();
    }
    /**
     * Resume stacking operation.
     */
    resumeStacking() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.resumeStacking();
    }
    /**
     * Retrieve the current redo stack which consists of operations that could be undone
     */
    getUndoStack() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.getUndoStack();
    }
    /**
     * Retrieve the current undo stack which consists of operations that were undone
     */
    getRedoStack() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.getRedoStack();
    }
    /**
     * Retrieve the complete history stack
     * @returns
     */
    getStack() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.getStack();
    }
    /**
     * Restore n operations that were last n reverted on the graph.
     * @param steps The number of operations to undo. Default to 1.
     * @returns
     */
    undo() {
        const history = this.getHistoryPlugin();
        history === null || history === void 0 ? void 0 : history.undo();
    }
    /**
     * Revert recent n operation(s) performed on the graph.
     * @param steps The number of operations to redo. Default to 1.
     * @returns
     */
    redo() {
        const history = this.getHistoryPlugin();
        history === null || history === void 0 ? void 0 : history.redo();
    }
    /**
     * Indicate whether there are any actions available in the undo stack.
     */
    canUndo() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.canUndo();
    }
    /**
     * Indicate whether there are any actions available in the redo stack.
     */
    canRedo() {
        const history = this.getHistoryPlugin();
        return history === null || history === void 0 ? void 0 : history.canRedo();
    }
    /**
     * Begin a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    startBatch() {
        const history = this.getHistoryPlugin();
        history === null || history === void 0 ? void 0 : history.startBatch();
    }
    /**
     * End a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    stopBatch() {
        const history = this.getHistoryPlugin();
        history === null || history === void 0 ? void 0 : history.stopBatch();
    }
    /**
     * Execute a provided function within a batched context
     * All operations performed inside callback will be treated as a composite operation
     * more convenient way without manually invoking `startBatch` and `stopBatch`.
     * @param callback The func containing operations to be batched together.
     */
    batch(callback) {
        const history = this.getHistoryPlugin();
        history === null || history === void 0 ? void 0 : history.batch(callback);
    }
    /**
     * Clear history stack
     * @param {StackType} stackType undo/redo stack
     */
    clearStack(stackType) {
        const history = this.getHistoryPlugin();
        if (!stackType)
            return history === null || history === void 0 ? void 0 : history.clear();
        return stackType === 'undo'
            ? history === null || history === void 0 ? void 0 : history.clearUndoStack()
            : history === null || history === void 0 ? void 0 : history.clearRedoStack();
    }
    /**
     * Collapse sub tree(s).
     * @param ids Root id(s) of the sub trees.
     * @param disableAnimate Whether disable the animations for this operation.
     * @param stack Whether push this operation to stack.
     * @returns
     * @group Tree
     */
    collapse(ids, disableAnimate = false, stack) {
        this.hooks.treecollapseexpand.emit({
            ids: (0, util_1.isArray)(ids) ? ids : [ids],
            action: 'collapse',
            animate: !disableAnimate,
            graphCore: this.dataController.graphCore,
        });
    }
    /**
     * Expand sub tree(s).
     * @param ids Root id(s) of the sub trees.
     * @param disableAnimate Whether disable the animations for this operation.
     * @param stack Whether push this operation to stack.
     * @returns
     * @group Tree
     */
    expand(ids, disableAnimate = false, stack) {
        this.hooks.treecollapseexpand.emit({
            ids: (0, util_1.isArray)(ids) ? ids : [ids],
            action: 'expand',
            animate: !disableAnimate,
            graphCore: this.dataController.graphCore,
        });
    }
    /**
     * Destroy the graph instance and remove the related canvases.
     * @returns
     * @group Graph Instance
     */
    destroy(callback) {
        // TODO: call the destroy functions after items' buildOut animations finished
        // setTimeout(() => {
        this.canvas.destroy();
        this.backgroundCanvas.destroy();
        this.transientCanvas.destroy();
        // clear history stack
        this.clearStack();
        callback === null || callback === void 0 ? void 0 : callback();
        // }, 500);
        this.hooks.destroy.emit({});
        // TODO: destroy controllers and off the listeners
        // this.dataController.destroy();
        // this.interactionController.destroy();
        // this.layoutController.destroy();
        // this.themeController.destroy();
        // this.itemController.destroy();
        // this.extensionController.destroy();
        this.destroyed = true;
    }
}
exports.default = Graph;
//# sourceMappingURL=graph.js.map