"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BrushSelect = void 0;
const stdlib_1 = require("../../stdlib");
const behavior_1 = require("../../types/behavior");
const array_1 = require("../../util/array");
const item_1 = require("../../util/item");
const ALLOWED_TRIGGERS = ['drag', 'shift', 'ctrl', 'alt', 'meta'];
const BRUSH_SHAPE_ID = 'g6-brush-select-brush-shape';
const DEFAULT_OPTIONS = {
    trigger: 'shift',
    selectedState: 'selected',
    itemTypes: ['node', 'edge'],
    eventName: '',
    selectSetMode: 'latest',
    brushStyle: {
        fill: '#EEF6FF',
        fillOpacity: 0.4,
        stroke: '#DDEEFE',
        lineWidth: 1,
    },
    shouldBegin: () => true,
    shouldUpdate: () => true,
    onSelect: () => { },
    onDeselect: () => { },
};
class BrushSelect extends behavior_1.Behavior {
    constructor(options) {
        super(Object.assign({}, DEFAULT_OPTIONS, options));
        this.selectedIds = {
            nodes: [],
            edges: [],
            combos: [],
        };
        this.getEvents = () => {
            return {
                pointerdown: this.onMouseDown,
                pointermove: this.onMouseMove,
                pointerup: this.onMouseUp,
            };
        };
        // Validate options
        if (options.trigger && !ALLOWED_TRIGGERS.includes(options.trigger)) {
            console.warn(`G6: Invalid trigger option "${options.trigger}" for brush-select behavior!`);
            this.options.trigger = DEFAULT_OPTIONS.trigger;
        }
    }
    isKeydown(event) {
        const trigger = this.options.trigger;
        const keyMap = {
            drag: true,
            shift: event.shiftKey,
            ctrl: event.ctrlKey,
            alt: event.altKey,
            meta: event.metaKey,
        };
        return keyMap[trigger];
    }
    onMouseDown(event) {
        if (!this.options.shouldBegin(event))
            return;
        const { itemId, itemType, canvas } = event;
        // should not begin at node
        if (itemId && itemType === 'node')
            return;
        this.beginPoint = {
            x: canvas.x,
            y: canvas.y,
        };
        if (!this.isKeydown(event)) {
            this.clearStates();
            return;
        }
        const { brush } = this;
        if (!brush) {
            this.brush = this.createBrush();
        }
        if (this.brush) {
            this.brush.setAttribute('width', 0);
            this.brush.setAttribute('height', 0);
            this.brush.setAttribute('visibility', 'visible');
        }
        this.mousedown = true; // TODO: no need if the drag events are supported on graph, use dragging instead
        // this.dragging = true;
    }
    onMouseMove(event) {
        if (!this.mousedown)
            return; // TODO: no need if the drag events are supported on graph, use dragging instead
        // if (!this.dragging) return;
        if (!this.isKeydown(event))
            return;
        this.dragging = true;
        this.brush = this.updateBrush(event);
    }
    onMouseUp(event) {
        if (!this.mousedown)
            return;
        if (!this.isKeydown(event)) {
            this.clearStates();
        }
        this.removeBrush();
        this.brush = undefined;
        this.selectedIds = this.selectItems(event);
        this.dragging = false;
        this.mousedown = false;
    }
    clearStates(clearIds = undefined, restIds = undefined) {
        const { graph } = this;
        const { selectedState, eventName, onDeselect } = this.options;
        let nodes = [];
        let edges = [];
        let combos = [];
        if (clearIds) {
            nodes = clearIds.nodes;
            edges = clearIds.edges;
            combos = clearIds.combos;
            this.selectedIds = restIds || {
                nodes: [],
                edges: [],
                combos: [],
            };
        }
        else {
            nodes = graph.findIdByState('node', selectedState);
            edges = graph.findIdByState('edge', selectedState);
            combos = graph.findIdByState('combo', selectedState);
            this.selectedIds = {
                nodes: [],
                edges: [],
                combos: [],
            };
        }
        nodes
            .concat(edges)
            .concat(combos)
            .forEach((id) => graph.setItemState(id, selectedState, false));
        if (nodes.length || edges.length || combos.length) {
            const deselectIds = { nodes, edges, combos };
            onDeselect === null || onDeselect === void 0 ? void 0 : onDeselect(this.selectedIds, deselectIds);
            if (eventName) {
                graph.emit(eventName, {
                    action: 'deselect',
                    selectedIds: this.selectedIds,
                    deselectedIds: deselectIds,
                });
            }
        }
    }
    selectItems(event) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        const { graph, options } = this;
        const { selectedState, itemTypes, eventName, selectSetMode, shouldUpdate, onSelect, } = options;
        const selector = this.getSelector();
        const points = this.getPoints(event).filter(Boolean);
        if (points.length < 2)
            return;
        let selectedIds = selector(graph, points, itemTypes.concat('node'));
        let operateSetFunc = (a, b) => b;
        const currentNotEmpty = ((_a = this.selectedIds) === null || _a === void 0 ? void 0 : _a.nodes.length) ||
            ((_b = this.selectedIds) === null || _b === void 0 ? void 0 : _b.edges.length) ||
            ((_c = this.selectedIds) === null || _c === void 0 ? void 0 : _c.combos.length);
        switch (selectSetMode) {
            case 'diff':
                if (currentNotEmpty) {
                    operateSetFunc = array_1.diffSet;
                }
                break;
            case 'union':
                operateSetFunc = array_1.unionSet;
                break;
            case 'intersect':
                if (currentNotEmpty) {
                    operateSetFunc = array_1.intersectSet;
                }
                break;
            case 'latest':
            default:
                break;
        }
        selectedIds = {
            nodes: operateSetFunc((_d = this.selectedIds) === null || _d === void 0 ? void 0 : _d.nodes, selectedIds.nodes).filter((id) => shouldUpdate('node', id, 'select', this)),
            edges: [],
            combos: operateSetFunc((_e = this.selectedIds) === null || _e === void 0 ? void 0 : _e.combos, selectedIds.combos).filter((id) => shouldUpdate('combo', id, 'select', this)),
        };
        const edgeSelectable = itemTypes.includes('edge');
        if (edgeSelectable) {
            selectedIds.edges = (0, item_1.getEdgesBetween)(graph, selectedIds.nodes.concat(selectedIds.combos)).filter((id) => shouldUpdate('edge', id, 'select', this));
        }
        const nodeSelectable = itemTypes.includes('node');
        if (!nodeSelectable)
            selectedIds.nodes = [];
        const comboSelectable = itemTypes.includes('combo');
        if (!comboSelectable)
            selectedIds.combos = [];
        if (selectedState) {
            const diffToClear = {
                nodes: nodeSelectable
                    ? (0, array_1.diffSet)((_f = this.selectedIds) === null || _f === void 0 ? void 0 : _f.nodes, selectedIds.nodes)
                    : [],
                edges: edgeSelectable
                    ? (0, array_1.diffSet)((_g = this.selectedIds) === null || _g === void 0 ? void 0 : _g.edges, selectedIds.edges)
                    : [],
                combos: comboSelectable
                    ? (0, array_1.diffSet)((_h = this.selectedIds) === null || _h === void 0 ? void 0 : _h.combos, selectedIds.combos)
                    : [],
            };
            const clearNotEmpty = diffToClear.nodes.length ||
                diffToClear.edges.length ||
                diffToClear.combos.length;
            if (clearNotEmpty) {
                this.clearStates(diffToClear, selectedIds);
            }
            const { nodes, edges, combos } = selectedIds;
            nodes
                .concat(edges)
                .concat(combos)
                .forEach((id) => graph.setItemState(id, selectedState, true));
        }
        this.selectedIds = selectedIds;
        const selectNotEmpty = selectedIds.nodes.length ||
            selectedIds.edges.length ||
            selectedIds.combos.length;
        if (selectNotEmpty)
            onSelect === null || onSelect === void 0 ? void 0 : onSelect(this.selectedIds);
        // Emit an event.
        if (eventName) {
            graph.emit(eventName, {
                action: 'select',
                selectedIds: this.selectedIds,
            });
        }
        return this.selectedIds;
    }
    createBrush() {
        const { graph, options } = this;
        const { brushStyle } = options;
        const brush = graph.drawTransient('rect', BRUSH_SHAPE_ID, {
            style: brushStyle,
            capture: false,
        });
        return brush;
    }
    updateBrush(event) {
        const { beginPoint, graph } = this;
        const endPoint = event.canvas;
        const brush = graph.drawTransient('rect', BRUSH_SHAPE_ID, {
            style: {
                width: Math.abs(endPoint.x - beginPoint.x),
                height: Math.abs(endPoint.y - beginPoint.y),
                x: Math.min(endPoint.x, beginPoint.x),
                y: Math.min(endPoint.y, beginPoint.y),
            },
        });
        return brush;
    }
    removeBrush() {
        const { graph } = this;
        graph.drawTransient('rect', BRUSH_SHAPE_ID, { action: 'remove' });
    }
    getSelector() {
        return stdlib_1.utils.rectSelector;
    }
    getPoints(event) {
        const { canvas: point } = event;
        return [this.beginPoint, point];
    }
    destroy() {
        this.removeBrush();
    }
}
exports.BrushSelect = BrushSelect;
//# sourceMappingURL=brush-select.js.map