"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZoomCanvas3D = void 0;
const behavior_1 = require("../../types/behavior");
const VALID_TRIGGERS = ['wheel', 'upDownKeys'];
const DEFAULT_OPTIONS = {
    trigger: 'wheel',
    secondaryKey: '',
    eventName: '',
    sensitivity: 10,
    triggerOnItems: true,
    minZoom: 0.01,
    maxZoom: 10,
    shouldBegin: () => true,
};
/**
 * Zoom the 3d canvas along the ray vertical to the screen.
 */
class ZoomCanvas3D extends behavior_1.Behavior {
    constructor(options) {
        const finalOptions = Object.assign({}, DEFAULT_OPTIONS, options);
        if (!VALID_TRIGGERS.includes(finalOptions.trigger)) {
            console.warn(`The trigger ${finalOptions.trigger} is not valid for zoom-canvas-3d behavior, "wheel" will take effect instead. Only "${VALID_TRIGGERS.join('", "')}" are available.`);
            finalOptions.trigger = 'wheel';
        }
        super(finalOptions);
        this.getEvents = () => {
            const $el = this.graph.canvas.getContextService().getDomElement();
            if ($el && $el.addEventListener) {
                $el.addEventListener('wheel', (e) => {
                    e.preventDefault();
                }, { passive: false });
            }
            if (this.options.trigger === 'wheel') {
                return {
                    wheel: this.onWheel,
                    keydown: this.onKeydown,
                    keyup: this.onKeyup,
                };
            }
            return {
                keydown: this.onKeydown,
                keyup: this.onKeyup,
            };
        };
    }
    onWheel(event) {
        const { graph, options } = this;
        const { secondaryKey, triggerOnItems, eventName, minZoom, maxZoom, sensitivity = 1, shouldBegin, } = options;
        if (event.itemId && !triggerOnItems)
            return;
        if (!shouldBegin(event))
            return;
        if (secondaryKey && !this.keydown)
            return;
        const camera = graph.canvas.getCamera();
        const sign = event.deltaY > 0 ? 1 : -1;
        const currentDistance = camera.getDistance();
        const dolly = ((100 * sign * sensitivity) / currentDistance) *
            Math.sqrt(currentDistance);
        const toDistance = currentDistance + dolly;
        const cameraFrontOfFocalPoint = camera.getDistanceVector()[2] < 0;
        // zoom out constraint
        if (dolly > 0 &&
            cameraFrontOfFocalPoint &&
            toDistance > (1 / minZoom) * 200) {
            return;
        }
        // zoom in constraint
        if (dolly < 0 && !cameraFrontOfFocalPoint && toDistance > maxZoom * 200) {
            return;
        }
        camera.dolly(dolly);
        // Emit event.
        if (eventName) {
            graph.emit(eventName, {
                zoom: camera.getZoom(),
            });
        }
    }
    onKeydown(event) {
        const { key } = event;
        const { trigger, secondaryKey } = this.options;
        if (trigger === 'upDownKeys') {
            if (key === 'ArrowUp')
                this.onWheel({ deltaY: -1 });
            if (key === 'ArrowDown')
                this.onWheel({ deltaY: 1 });
        }
        if (secondaryKey === key.toLowerCase()) {
            this.keydown = true;
        }
    }
    onKeyup(event) {
        const { key } = event;
        const { secondaryKey } = this.options;
        if (secondaryKey === key.toLowerCase())
            this.keydown = false;
    }
}
exports.ZoomCanvas3D = ZoomCanvas3D;
//# sourceMappingURL=zoom-canvas-3d.js.map