"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZoomCanvas = void 0;
const behavior_1 = require("../../types/behavior");
const VALID_TRIGGERS = ['wheel', 'upDownKeys'];
const DEFAULT_OPTIONS = {
    enableOptimize: false,
    triggerOnItems: true,
    sensitivity: 2,
    trigger: 'wheel',
    secondaryKey: '',
    speedUpKey: 'shift',
    eventName: '',
    minZoom: 0.00001,
    maxZoom: 1000,
    shouldBegin: () => true,
};
class ZoomCanvas extends behavior_1.Behavior {
    constructor(options) {
        const finalOptions = Object.assign({}, DEFAULT_OPTIONS, options);
        if (!VALID_TRIGGERS.includes(finalOptions.trigger)) {
            console.warn(`The trigger ${finalOptions.trigger} is not valid, 'wheel' will take effect.`);
            finalOptions.trigger = 'wheel';
        }
        super(finalOptions);
    }
    getEvents() {
        var _a, _b;
        (_b = (_a = this.graph.canvas
            .getContextService()
            .getDomElement()).addEventListener) === null || _b === void 0 ? void 0 : _b.call(_a, 'wheel', (e) => {
            e.preventDefault();
        }, { passive: false });
        if (this.options.trigger === 'upDownKeys') {
            return {
                keydown: this.onKeydown,
                keyup: this.onKeyup,
            };
        }
        return {
            wheel: this.onWheel,
            keydown: this.onKeydown,
            keyup: this.onKeyup,
        };
    }
    hideShapes() {
        const { graph } = this;
        if (this.options.enableOptimize) {
            this.hiddenEdgeIds = graph
                .getAllEdgesData()
                .map((edge) => edge.id)
                .filter((id) => graph.getItemVisible(id) === true);
            // TODO
            graph.hideItem(this.hiddenEdgeIds, false);
            this.hiddenNodeIds = graph
                .getAllNodesData()
                .map((node) => node.id)
                .filter((id) => graph.getItemVisible(id) === true);
            // draw node's keyShapes on transient, and then hidden the real nodes;
            // this.hiddenNodeIds.forEach((id) => {
            //   graph.drawTransient('node', id, {
            //     onlyDrawKeyShape: true,
            //   });
            // });
            // graph.hideItem(this.hiddenNodeIds, false);
        }
    }
    endZoom() {
        const { graph, hiddenEdgeIds, hiddenNodeIds } = this;
        const { enableOptimize } = this.options;
        this.zooming = false;
        if (enableOptimize) {
            // restore hidden items
            if (hiddenEdgeIds) {
                graph.showItem(hiddenEdgeIds, false);
            }
            // TODO
            // if (hiddenNodeIds) {
            //   hiddenNodeIds.forEach((id) => {
            //     graph.drawTransient('node', id, { action: 'remove' });
            //   });
            //   graph.showItem(hiddenNodeIds, false);
            // }
        }
        this.hiddenEdgeIds = [];
        this.hiddenNodeIds = [];
    }
    onWheel(event) {
        const { graph, keydown } = this;
        const { deltaY, client, itemId } = event;
        const { eventName, sensitivity, secondaryKey, triggerOnItems, minZoom, maxZoom, shouldBegin, } = this.options;
        // TODO: CANVAS
        const isOnItem = itemId && itemId !== 'CANVAS';
        if ((secondaryKey && !keydown) ||
            (isOnItem && !triggerOnItems) ||
            !shouldBegin(event)) {
            this.endZoom();
            return;
        }
        // begin zooming
        if (!this.zooming) {
            this.hideShapes();
            this.zooming = true;
        }
        let zoomRatio = 1;
        if (deltaY < 0)
            zoomRatio = (100 + sensitivity) / 100;
        if (deltaY > 0)
            zoomRatio = 100 / (100 + sensitivity);
        const zoomTo = zoomRatio * graph.getZoom();
        if (minZoom && zoomTo < minZoom)
            return;
        if (maxZoom && zoomTo > maxZoom)
            return;
        // TODO: the zoom center is wrong?
        graph.zoom(zoomRatio, { x: client.x, y: client.y });
        clearTimeout(this.zoomTimer);
        this.zoomTimer = setTimeout(() => {
            this.endZoom();
        }, 300);
        // Emit event.
        if (eventName) {
            graph.emit(eventName, { zoom: { ratio: zoomRatio } });
        }
    }
    onKeydown(event) {
        const { key } = event;
        const { secondaryKey, trigger, speedUpKey, eventName, sensitivity, shouldBegin, } = this.options;
        if (secondaryKey && secondaryKey === key.toLowerCase()) {
            this.keydown = true;
        }
        if (speedUpKey && speedUpKey === key.toLowerCase()) {
            this.speedupKeydown = true;
        }
        if (trigger === 'upDownKeys') {
            if (secondaryKey && !this.keydown)
                return;
            if (!shouldBegin(event))
                return;
            const { graph, speedupKeydown } = this;
            const speed = speedupKeydown ? 20 : 1;
            let zoomRatio = 1;
            switch (key) {
                case 'ArrowUp':
                    // Zoom in.
                    zoomRatio = (100 + sensitivity * speed) / 100;
                    break;
                case 'ArrowDown':
                    // Zoom out.
                    zoomRatio = 100 / (100 + sensitivity * speed);
                    break;
            }
            if (zoomRatio !== 1) {
                graph.zoom(zoomRatio);
                if (eventName) {
                    graph.emit(eventName, { zoom: { ratio: zoomRatio } });
                }
            }
        }
    }
    onKeyup(event) {
        const { key } = event;
        const { secondaryKey, speedUpKey } = this.options;
        if (secondaryKey && secondaryKey === key.toLowerCase()) {
            this.keydown = false;
        }
        if (speedUpKey && speedUpKey === key.toLowerCase()) {
            this.speedupKeydown = false;
        }
    }
}
exports.ZoomCanvas = ZoomCanvas;
//# sourceMappingURL=zoom-canvas.js.map