"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CubicEdge = void 0;
const matrix_util_1 = require("@antv/matrix-util");
// eslint-disable-next-line import/namespace
const base_1 = require("./base");
class CubicEdge extends base_1.BaseEdge {
    constructor(props) {
        super(props);
        this.type = 'cubic-edge';
        this.defaultStyles = {
            keyShape: {
                x1: 0,
                y1: 0,
                z1: 0,
                x2: 0,
                y2: 0,
                z2: 0,
                isBillboard: true,
            },
        };
        /**
         * calculate the control points by curvePosition|controlPoints|curveOffset
         * @param startPoint: source point position of edge
         * @param endPoint target point position of edge
         * @param percent the proportion of control points' in the segment, Range 0 to 1
         * @param controlPoints the control point position
         * @param offset the curveOffset
         * @returns control points
         */
        this.getControlPoints = (startPoint, endPoint, percent = 0.5, controlPoints, offset = 20) => {
            if ((controlPoints === null || controlPoints === void 0 ? void 0 : controlPoints.length) > 1)
                return controlPoints;
            const controlPoint1 = this.getControlPoint(startPoint, endPoint, percent, offset);
            const controlPoint2 = this.getControlPoint(startPoint, endPoint, percent, -offset);
            return [controlPoint1, controlPoint2];
        };
        /**
         * control point calculated according to startPoint, endPoint, percent, and offset
         * @param  {Point} startPoint source point position of edge (x, y)
         * @param  {Point} endPoint  target point position of edge (x, y)
         * @param  {Number} percent   the proportion of control points' in the segment, Range 0 to 1
         * @param  {Number} offset    the curveOffset
         * @return {Point} control point (x,y)
         */
        this.getControlPoint = (startPoint, endPoint, percent = 0, offset = 0) => {
            const point = {
                x: (1 - percent) * startPoint.x + percent * endPoint.x,
                y: (1 - percent) * startPoint.y + percent * endPoint.y,
            };
            let tangent = [0, 0];
            matrix_util_1.vec2.normalize(tangent, [
                endPoint.x - startPoint.x,
                endPoint.y - startPoint.y,
            ]);
            if (!tangent || (!tangent[0] && !tangent[1])) {
                tangent = [0, 0];
            }
            const perpendicular = [-tangent[1] * offset, tangent[0] * offset]; // Vertical vector
            point.x += perpendicular[0];
            point.y += perpendicular[1];
            return point;
        };
        // suggest to merge default styles like this to avoid style value missing
        // this.defaultStyles = mergeStyles([this.baseDefaultStyles, this.defaultStyles]);
    }
    draw(model, sourcePoint, targetPoint, shapeMap, diffData, diffState) {
        const { data = {} } = model;
        const shapes = { keyShape: undefined };
        shapes.keyShape = this.drawKeyShape(model, sourcePoint, targetPoint, shapeMap, diffData);
        if (data.haloShape) {
            shapes.haloShape = this.drawHaloShape(model, shapeMap, diffData);
        }
        if (data.labelShape) {
            shapes.labelShape = this.drawLabelShape(model, shapeMap, diffData);
        }
        // labelBackgroundShape
        if (data.labelBackgroundShape) {
            shapes.labelBackgroundShape = this.drawLabelBackgroundShape(model, shapeMap, diffData);
        }
        if (data.iconShape) {
            shapes.iconShape = this.drawIconShape(model, shapeMap, diffData);
        }
        // TODO: other shapes
        return shapes;
    }
    drawKeyShape(model, sourcePoint, targetPoint, shapeMap, diffData, diffState) {
        const { keyShape: keyShapeStyle } = this.mergedStyles;
        const { startArrow, endArrow } = keyShapeStyle, others = __rest(keyShapeStyle, ["startArrow", "endArrow"]);
        const controlPoints = this.getControlPoints(sourcePoint, targetPoint, keyShapeStyle.curvePosition, keyShapeStyle.controlPoints, keyShapeStyle.curveOffset);
        const lineStyle = Object.assign(Object.assign({}, others), { path: [
                ['M', sourcePoint.x, sourcePoint.y],
                [
                    'C',
                    controlPoints[0].x,
                    controlPoints[0].y,
                    controlPoints[1].x,
                    controlPoints[1].y,
                    targetPoint.x,
                    targetPoint.y,
                ],
            ] });
        this.upsertArrow('start', startArrow, others, model, lineStyle);
        this.upsertArrow('end', endArrow, others, model, lineStyle);
        return this.upsertShape('path', 'keyShape', lineStyle, shapeMap, model);
    }
}
exports.CubicEdge = CubicEdge;
//# sourceMappingURL=cubic.js.map