"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseNode = void 0;
const constant_1 = require("../../../constant");
const item_1 = require("../../../types/item");
const shape_1 = require("../../../util/shape");
const text_1 = require("../../../util/text");
const type_1 = require("../../../util/type");
const animate_1 = require("../../../util/animate");
const zoom_1 = require("../../../util/zoom");
class BaseNode {
    constructor(props) {
        //vertex coordinate
        /**
         * Cache the scale transform calculated by balancing size, for restoring.
         */
        this.scaleTransformCache = '';
        // cache the zoom level infomations
        this.zoomCache = {
            hiddenShape: {},
            zoom: 1,
            zoomLevel: 0,
            balanceRatio: 1,
            levelShapes: {},
            wordWrapWidth: 32,
            animateConfig: animate_1.DEFAULT_ANIMATE_CFG.zoom,
            firstRender: true,
        };
        /**
         * The listener for graph zooming.
         * 1. show / hide some shapes while zoom level changed;
         * 2. change the shapes' sizes to make them have same visual size while zooming, e.g. labelShape, labelBackgroundShape.
         * @param shapeMap The shape map that contains all of the elements to show on the node.
         * @param zoom The zoom level of the graph.
         */
        this.onZoom = (shapeMap, zoom) => {
            var _a, _b, _c;
            // zoomLevel changed
            if (this.lodStrategy) {
                const { levels } = this.lodStrategy;
                // last zoom ratio responsed by zoom changing, which might not equal to zoom.previous in props since the function is debounced.
                const { levelShapes, hiddenShape, animateConfig, firstRender = true, zoomLevel: previousLevel, } = this.zoomCache;
                const currentLevel = (0, zoom_1.getZoomLevel)(levels, zoom);
                const levelNums = Object.keys(levelShapes).map(Number);
                const maxLevel = Math.max(...levelNums);
                const minLevel = Math.min(...levelNums);
                if (currentLevel < previousLevel) {
                    if (firstRender) {
                        // zoomLevel changed, from higher to lower, hide something
                        for (let i = currentLevel + 1; i <= maxLevel; i++) {
                            (_a = levelShapes[String(i)]) === null || _a === void 0 ? void 0 : _a.forEach((id) => {
                                if (!shapeMap[id])
                                    return;
                                shapeMap[id].hide();
                                hiddenShape[id] = true;
                            });
                        }
                    }
                    else {
                        // zoomLevel changed, from higher to lower, hide something
                        for (let i = currentLevel + 1; i <= maxLevel; i++) {
                            (_b = levelShapes[String(i)]) === null || _b === void 0 ? void 0 : _b.forEach((id) => (0, animate_1.fadeOut)(id, shapeMap[id], hiddenShape, animateConfig));
                        }
                    }
                }
                else if (currentLevel > previousLevel) {
                    // zoomLevel changed, from lower to higher, show something
                    for (let i = currentLevel; i >= minLevel; i--) {
                        (_c = levelShapes[String(i)]) === null || _c === void 0 ? void 0 : _c.forEach((id) => {
                            (0, animate_1.fadeIn)(id, shapeMap[id], this.mergedStyles[id] ||
                                this.mergedStyles[id.replace('Background', '')], hiddenShape, animateConfig);
                        });
                    }
                }
                this.zoomCache.zoomLevel = currentLevel;
            }
            this.balanceShapeSize(shapeMap, zoom);
            this.zoomCache.zoom = zoom;
        };
        const { themeStyles, lodStrategy, zoom } = props;
        if (themeStyles)
            this.themeStyles = themeStyles;
        this.lodStrategy = lodStrategy;
        this.boundsCache = {};
        this.zoomCache.zoom = zoom;
        this.zoomCache.balanceRatio = 1 / zoom;
        this.zoomCache.animateConfig = Object.assign(Object.assign({}, animate_1.DEFAULT_ANIMATE_CFG.zoom), lodStrategy === null || lodStrategy === void 0 ? void 0 : lodStrategy.animateCfg);
    }
    /**
     * Get merged styles from `getMergedStyles` and assigns the merged styles to the 'mergedStyles' property.
     * @param model - The NodeDisplayModel or ComboDisplayModel to merge the styles from.
     */
    mergeStyles(model) {
        this.mergedStyles = this.getMergedStyles(model);
    }
    /**
     * Merge style
     * @param model - The NodeDisplayModel or ComboDisplayModel to retrieve the merged styles from.
     * @returns The merged styles as a NodeShapeStyles object.
     */
    getMergedStyles(model) {
        var _a;
        const { data } = model;
        const dataStyles = {};
        Object.keys(data).forEach((fieldName) => {
            if (constant_1.RESERVED_SHAPE_IDS.includes(fieldName)) {
                if (fieldName === 'badgeShapes') {
                    dataStyles[fieldName] = {};
                    Object.keys(data[fieldName]).forEach((key) => {
                        if (isNaN(Number(key))) {
                            // key is not a number, it is a common style
                            dataStyles[fieldName][key] = data[fieldName][key];
                            return;
                        }
                        const { position } = data[fieldName][key];
                        dataStyles[`${position}BadgeShape`] = Object.assign(Object.assign({}, data[fieldName][key]), { tag: 'badgeShape' });
                    });
                }
                else if (fieldName === 'anchorShapes') {
                    Object.keys(data[fieldName]).forEach((idx) => {
                        dataStyles[`anchorShape${idx}`] = Object.assign(Object.assign({}, data[fieldName][idx]), { tag: 'anchorShape' });
                    });
                }
                else {
                    dataStyles[fieldName] = data[fieldName];
                }
            }
            else if (fieldName === constant_1.OTHER_SHAPES_FIELD_NAME) {
                Object.keys(data[fieldName]).forEach((otherShapeId) => (dataStyles[otherShapeId] = data[fieldName][otherShapeId]));
            }
        });
        const merged = (0, shape_1.mergeStyles)([
            this.themeStyles,
            this.defaultStyles,
            dataStyles,
        ]);
        const padding = (_a = merged.labelBackgroundShape) === null || _a === void 0 ? void 0 : _a.padding;
        if (padding) {
            merged.labelBackgroundShape.padding = (0, shape_1.formatPadding)(padding);
        }
        return merged;
    }
    /**
     * Call it after calling draw function to update cache about bounds and zoom levels.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     */
    updateCache(shapeMap) {
        ['keyShape', 'labelShape']
            .concat(Object.keys(item_1.BadgePosition))
            .map((pos) => `${pos}BadgeShape`)
            .forEach((id) => {
            const shape = shapeMap[id];
            if (shape === null || shape === void 0 ? void 0 : shape.getAttribute(shape_1.LOCAL_BOUNDS_DIRTY_FLAG_KEY)) {
                if (id === 'labelShape') {
                    this.boundsCache[`${id}Geometry`] = shape.getGeometryBounds();
                }
                else {
                    this.boundsCache[`${id}Local`] = shape.getLocalBounds();
                }
                shape.setAttribute(shape_1.LOCAL_BOUNDS_DIRTY_FLAG_KEY, false);
            }
        });
        const levelShapes = {};
        Object.keys(shapeMap).forEach((shapeId) => {
            const { lod } = shapeMap[shapeId].attributes;
            if (lod !== undefined) {
                levelShapes[lod] = levelShapes[lod] || [];
                levelShapes[lod].push(shapeId);
            }
        });
        this.zoomCache.levelShapes = levelShapes;
        if (shapeMap.labelShape && this.boundsCache.keyShapeLocal) {
            const { maxWidth = '200%' } = this.mergedStyles.labelShape || {};
            this.zoomCache.wordWrapWidth = (0, text_1.getWordWrapWidthByBox)(this.boundsCache.keyShapeLocal, maxWidth, 1);
        }
    }
    /**
     * Perform additional drawing operations or add custom shapes after drawing node.
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param shapesChanged An array of shape IDs that have changed and need to be updated.
     * @returns An object that contains some new shapes to be added to the node.
     */
    afterDraw(model, shapeMap, shapesChanged) {
        return {};
    }
    /**
     * Draw the label shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the label shape of the node.
     */
    drawLabelShape(model, shapeMap, diffData, diffState) {
        const { labelShape: shapeStyle } = this.mergedStyles;
        if (!shapeStyle || !shapeStyle.text || !model.data.labelShape)
            return;
        const { keyShape } = shapeMap;
        this.boundsCache.keyShapeLocal =
            this.boundsCache.keyShapeLocal || keyShape.getLocalBounds();
        const keyShapeBox = (0, shape_1.getShapeLocalBoundsByStyle)(keyShape, this.mergedStyles.keyShape, this.boundsCache.keyShapeLocal);
        const { position, offsetX: propsOffsetX, offsetY: propsOffsetY, offsetZ: propsOffsetZ, maxWidth, 
        // @ts-ignore
        angle } = shapeStyle, otherStyle = __rest(shapeStyle, ["position", "offsetX", "offsetY", "offsetZ", "maxWidth", "angle"]);
        const wordWrapWidth = (0, text_1.getWordWrapWidthByBox)(keyShapeBox, maxWidth, this.zoomCache.zoom);
        const positionPreset = {
            x: keyShapeBox.center[0],
            y: keyShapeBox.max[1],
            z: keyShapeBox.center[2],
            textBaseline: 'top',
            textAlign: 'center',
            offsetX: 0,
            offsetY: 0,
            offsetZ: 0,
            wordWrapWidth,
        };
        switch (position) {
            case 'center':
                positionPreset.y = keyShapeBox.center[1];
                positionPreset.textBaseline = 'middle';
                break;
            case 'top':
                positionPreset.y = keyShapeBox.min[1];
                positionPreset.textBaseline = 'bottom';
                positionPreset.offsetY = -2;
                break;
            case 'left':
                positionPreset.x = keyShapeBox.min[0];
                positionPreset.y = keyShapeBox.center[1];
                positionPreset.textAlign = 'right';
                positionPreset.textBaseline = 'middle';
                positionPreset.offsetX = -4;
                break;
            case 'right':
                positionPreset.x = keyShapeBox.max[0];
                positionPreset.y = keyShapeBox.center[1];
                positionPreset.textAlign = 'left';
                positionPreset.textBaseline = 'middle';
                positionPreset.offsetX = 4;
                break;
            default: // at bottom by default
                positionPreset.offsetY = 2;
                break;
        }
        const offsetX = (propsOffsetX === undefined ? positionPreset.offsetX : propsOffsetX);
        const offsetY = (propsOffsetY === undefined ? positionPreset.offsetY : propsOffsetY);
        const offsetZ = (propsOffsetZ === undefined ? positionPreset.offsetZ : propsOffsetZ);
        positionPreset.x += offsetX;
        positionPreset.y += offsetY;
        positionPreset.z += offsetZ;
        const style = Object.assign(Object.assign(Object.assign(Object.assign({}, this.defaultStyles.labelShape), positionPreset), { isBillboard: true }), otherStyle);
        if (angle) {
            style.transform = `rotate(${angle}rad)`;
        }
        return this.upsertShape('text', 'labelShape', style, shapeMap, model);
    }
    /**
     * Draw the label background shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the label background shape of the node.
     */
    drawLabelBackgroundShape(model, shapeMap, diffData, diffState) {
        const { labelShape } = shapeMap;
        if (!labelShape || !labelShape.style.text || !model.data.labelShape)
            return;
        if (!this.boundsCache.labelShapeGeometry ||
            labelShape.getAttribute(shape_1.LOCAL_BOUNDS_DIRTY_FLAG_KEY)) {
            this.boundsCache.labelShapeGeometry = labelShape.getGeometryBounds();
            labelShape.setAttribute(shape_1.LOCAL_BOUNDS_DIRTY_FLAG_KEY, false);
        }
        // label's local bounds, will take scale into acount
        const { labelShapeGeometry: textBBoxGeo } = this.boundsCache;
        const height = textBBoxGeo.max[1] - textBBoxGeo.min[1];
        const width = Math.min(textBBoxGeo.max[0] - textBBoxGeo.min[0] + 2, labelShape.attributes.wordWrapWidth);
        const _a = this.mergedStyles.labelBackgroundShape, { padding } = _a, backgroundStyle = __rest(_a, ["padding"]);
        const bgStyle = Object.assign(Object.assign({ fill: '#fff' }, backgroundStyle), { x: labelShape.attributes.x + textBBoxGeo.min[0] - padding[3], y: labelShape.attributes.y + textBBoxGeo.min[1] - padding[0], width: width + padding[1] + padding[3], height: height + padding[0] + padding[2] });
        const bgShape = this.upsertShape('rect', 'labelBackgroundShape', bgStyle, shapeMap, model);
        this.balanceShapeSize(shapeMap, this.zoomCache.zoom);
        return bgShape;
    }
    /**
     * Draw the icon shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the icon shape of the node.
     */
    drawIconShape(model, shapeMap, diffData, diffState) {
        const { iconShape: shapeStyle } = this.mergedStyles;
        const { width, height, fontSize, text, offsetX = 0, offsetY = 0, } = shapeStyle;
        const w = (width || fontSize);
        const h = (height || fontSize);
        const iconShapeType = text ? 'text' : 'image';
        if (iconShapeType === 'image') {
            shapeStyle.x = -w / 2 + offsetX;
            shapeStyle.y = -h / 2 + offsetY;
            shapeStyle.width = w;
            shapeStyle.height = h;
        }
        else {
            shapeStyle.textAlign = 'center';
            shapeStyle.textBaseline = 'middle';
            shapeStyle.x = offsetX;
            shapeStyle.y = offsetY;
            shapeStyle.fontSize = w;
        }
        return this.upsertShape(iconShapeType, 'iconShape', shapeStyle, shapeMap, model);
    }
    /**
     * Draw the halo shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the halo shape of the node.
     */
    drawHaloShape(model, shapeMap, diffData, diffState) {
        const { keyShape } = shapeMap;
        const { haloShape: haloShapeStyle, keyShape: keyShapeStyle } = this.mergedStyles;
        if (haloShapeStyle.visible === false)
            return;
        const { nodeName, attributes } = keyShape;
        const { x, y, fill } = attributes;
        return this.upsertShape(nodeName, 'haloShape', Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, attributes), keyShapeStyle), { x,
            y, stroke: fill }), haloShapeStyle), { batchKey: 'halo' }), shapeMap, model);
    }
    /**
     * Draw the anchors shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the anchors shape of the node.
     */
    drawAnchorShapes(model, shapeMap, diffData, diffState) {
        const { anchorShapes: commonStyle, keyShape: keyShapeStyle } = this.mergedStyles;
        const individualConfigs = Object.values(this.mergedStyles).filter((style) => style.tag === 'anchorShape');
        if (!individualConfigs.length)
            return;
        this.boundsCache.keyShapeLocal =
            this.boundsCache.keyShapeLocal || shapeMap.keyShape.getLocalBounds();
        const shapes = {};
        const anchorPositionMap = this.calculateAnchorPosition(keyShapeStyle);
        individualConfigs.forEach((config, i) => {
            const { position, fill = keyShapeStyle.fill } = config, style = __rest(config, ["position", "fill"]);
            const [cx, cy] = this.getAnchorPosition(position, anchorPositionMap);
            const id = `anchorShape${i}`;
            shapes[id] = this.upsertShape('circle', id, Object.assign(Object.assign({ cx,
                cy,
                fill }, commonStyle), style), shapeMap, model);
        });
        return shapes;
    }
    getAnchorPosition(position, anchorPositionMap) {
        const keyShapeBBox = this.boundsCache.keyShapeLocal;
        const keyShapeWidth = keyShapeBBox.max[0] - keyShapeBBox.min[0];
        const keyShapeHeight = keyShapeBBox.max[1] - keyShapeBBox.min[1];
        const defaultPosition = [
            keyShapeBBox.max[0],
            keyShapeBBox.min[1],
        ]; //topRight
        if (position instanceof Array) {
            return [
                keyShapeWidth * (position[0] - 0.5),
                keyShapeHeight * (position[1] - 0.5),
            ];
        }
        else if (typeof position === 'string') {
            position = position.toLowerCase();
            //receive a unknown string, remind the user.
            return (anchorPositionMap[position] ||
                anchorPositionMap['default'] ||
                defaultPosition);
        }
        //receive a position in unknown type (such as a number or undefined).
        return anchorPositionMap['default'] || defaultPosition;
    }
    /**
     * Configures the anchor positions based on the provided keyShapeStyle and returns the configuration.
     * e.g for a CircleNode, it returns: `{"right":keyShapeStyle.x+keyShapeStyle.r, keyShapeStyle.y}`
     * @param keyShapeStyle - The keyShapeStyle object that contains the style information of the key shape.
     * @returns The anchor position configuration as an IAnchorPositionMap object.
     */
    calculateAnchorPosition(keyShapeStyle) {
        const x = (0, type_1.convertToNumber)(keyShapeStyle.x);
        const y = (0, type_1.convertToNumber)(keyShapeStyle.y);
        const r = (0, type_1.convertToNumber)(keyShapeStyle.r);
        const anchorPositionMap = {};
        anchorPositionMap['top'] = [x, y - r];
        anchorPositionMap['left'] = [x - r, y];
        anchorPositionMap['right'] = anchorPositionMap['default'] = [x + r, y];
        anchorPositionMap['bottom'] = [x, y + r];
        return anchorPositionMap;
    }
    /**
     * Draw the badges shape of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the badges shape of the node.
     */
    drawBadgeShapes(model, shapeMap, diffData, diffState) {
        const { badgeShapes: commonStyle, keyShape: keyShapeStyle } = this.mergedStyles;
        const individualConfigs = Object.values(this.mergedStyles).filter((style) => style.tag === 'badgeShape');
        if (!individualConfigs.length)
            return {};
        this.boundsCache.keyShapeLocal =
            this.boundsCache.keyShapeLocal || shapeMap.keyShape.getLocalBounds();
        const keyShapeBBox = (0, shape_1.getShapeLocalBoundsByStyle)(shapeMap.keyShape, keyShapeStyle, this.boundsCache.keyShapeLocal);
        const keyShapeSize = Math.min(keyShapeBBox.max[0] - keyShapeBBox.min[0], keyShapeBBox.max[1] - keyShapeBBox.min[1]);
        const shapes = {};
        individualConfigs.forEach((config, i) => {
            const { position } = config, individualStyle = __rest(config, ["position"]);
            const id = `${position}BadgeShape`;
            const style = Object.assign(Object.assign({}, commonStyle), individualStyle);
            const colorFromPalette = commonStyle.palette
                ? commonStyle.palette[i % commonStyle.palette.length]
                : '#7E92B5';
            const { text = '', type, color = colorFromPalette, size = keyShapeSize / 3, textColor, zIndex = 2, offsetX = 0, offsetY = 0 } = style, otherStyles = __rest(style, ["text", "type", "color", "size", "textColor", "zIndex", "offsetX", "offsetY"]);
            const bgHeight = size;
            const pos = { x: 0, y: 0 };
            switch (position) {
                case 'rightTop':
                case 'topRight':
                    pos.x = keyShapeBBox.max[0] / 2 + offsetX;
                    pos.y = keyShapeBBox.min[1] * 0.44 - bgHeight + offsetY;
                    break;
                case 'right':
                    pos.x = keyShapeBBox.max[0] / 2 + offsetX;
                    pos.y = -bgHeight / 2 + offsetY;
                    break;
                case 'rightBottom':
                case 'bottomRight':
                    pos.x = keyShapeBBox.max[0] / 2 + offsetX;
                    pos.y = keyShapeBBox.max[1] * 0.44 + offsetY;
                    break;
                case 'leftTop':
                case 'topLeft':
                    pos.x = keyShapeBBox.min[0] / 2 + offsetX;
                    pos.y = keyShapeBBox.min[1] * 0.44 - bgHeight + offsetY;
                    break;
                case 'left':
                    pos.x = keyShapeBBox.min[0] / 2 + offsetX;
                    pos.y = -bgHeight / 2 + offsetY;
                    break;
                case 'leftBottom':
                case 'bottomLeft':
                    pos.x = keyShapeBBox.min[0] / 2 + offsetX;
                    pos.y = keyShapeBBox.max[1] * 0.44 + offsetY;
                    break;
                case 'top':
                    pos.x = offsetX;
                    pos.y = keyShapeBBox.min[1] * 0.44 - bgHeight + offsetY;
                    break;
                default:
                    // bottom
                    pos.x = offsetX;
                    pos.y = keyShapeBBox.max[1] * 0.44 + offsetY;
                    break;
            }
            const positionLowerCase = position.toLowerCase();
            const isLeft = positionLowerCase.includes('left');
            const isCenter = !isLeft && !positionLowerCase.includes('right');
            // a radius rect (as container) + a text / icon
            const fontSize = bgHeight - 3;
            let textAlign = isLeft ? 'right' : 'left';
            let textX = isLeft ? pos.x - 3 : pos.x + 2;
            if (isCenter) {
                textAlign = 'center';
                textX = pos.x;
            }
            shapes[id] = this.upsertShape('text', id, Object.assign(Object.assign({ text, fill: textColor, fontSize, x: textX, y: pos.y + bgHeight / 2 }, otherStyles), { textAlign, textBaseline: 'middle', zIndex: zIndex + 1 }), shapeMap, model);
            this.boundsCache[`${id}Local`] = shapes[id].getLocalBounds();
            const bbox = this.boundsCache[`${id}Local`];
            const bgShapeId = `${position}BadgeBackgroundShape`;
            const bgWidth = text.length <= 1
                ? bgHeight
                : Math.max(bgHeight, bbox.max[0] - bbox.min[0]) + 8;
            let bgX = isLeft ? pos.x - bgWidth : pos.x;
            if (isCenter)
                bgX = pos.x - bgWidth / 2;
            shapes[bgShapeId] = this.upsertShape('rect', bgShapeId, Object.assign({ text, fill: color, height: bgHeight, width: bgWidth, radius: bgHeight / 2, zIndex, x: bgX, y: pos.y }, otherStyles), shapeMap, model);
        });
        return shapes;
    }
    /**
     * Draw other shapes(such as preRect,stateIcon) of the node
     * @param model The displayed model of this node, only for drawing and not received by users.
     * @param shapeMap The shape map that contains all of the elements to show on the node.
     * @param diffData An object that contains previous and current data.
     * @param diffState An object that contains previous and current node's state.
     * @returns The display object representing the other shapes of the node.
     */
    drawOtherShapes(model, shapeMap, diffData, diffState) {
        return {};
    }
    /**
     * Update the shapes' sizes e.g. labelShape, labelBackgroundShape, to keep the visual size while zooming.
     * @param shapeMap
     * @param zoom
     * @returns
     */
    balanceShapeSize(shapeMap, zoom) {
        // balance the size for label, badges
        const { keyShape, labelShape, labelBackgroundShape } = shapeMap;
        const balanceRatio = 1 / zoom || 1;
        this.zoomCache.balanceRatio = balanceRatio;
        if (!labelShape || !labelShape.isVisible())
            return;
        const { labelShape: labelStyle } = this.mergedStyles;
        const { position = 'bottom' } = labelStyle;
        const keyShapeLocal = keyShape.getLocalBounds();
        if (zoom < 1) {
            // if it is zoom-out, do not scale the gap between keyShape and labelShape, differentiate from zoom-in by adjusting transformOrigin
            if (position === 'bottom')
                labelShape.style.transformOrigin = '0';
            else
                labelShape.style.transformOrigin = '';
        }
        else {
            switch (position) {
                case 'bottom':
                    labelShape.style.transformOrigin = `0 ${keyShapeLocal.max[1] - labelShape.attributes.y}`;
                    break;
                case 'right':
                    labelShape.style.transformOrigin = `${keyShapeLocal.max[0] - labelShape.attributes.x} 0`;
                    break;
                case 'top':
                    labelShape.style.transformOrigin = `0 ${keyShapeLocal.min[1] - labelShape.attributes.y}`;
                    break;
                case 'left':
                    labelShape.style.transformOrigin = `${keyShapeLocal.min[0] - labelShape.attributes.x} 0`;
                    break;
                default:
                    // center
                    labelShape.style.transformOrigin = ``;
                    break;
            }
        }
        const oriTransform = (labelShape.style.transform || '').replace(this.scaleTransformCache, '');
        const scaleTransform = `scale(${balanceRatio}, ${balanceRatio})`;
        labelShape.style.transform = `${oriTransform} ${scaleTransform}`;
        this.scaleTransformCache = scaleTransform;
        const wordWrapWidth = this.zoomCache.wordWrapWidth * zoom;
        labelShape.style.wordWrapWidth = wordWrapWidth;
        if (!labelBackgroundShape || !labelBackgroundShape.isVisible())
            return;
        const { padding } = this.mergedStyles.labelBackgroundShape;
        const { width, height, x, y } = labelBackgroundShape.attributes;
        const [paddingTop, paddingRight, paddingBottom, paddingLeft] = padding;
        switch (position) {
            case 'top':
                // if it is zoom-out, do not scale the gap between keyShape and labelShape, differentiate from zoom-in by adjusting transformOrigin
                labelBackgroundShape.style.transformOrigin = `${paddingLeft + (width - paddingLeft - paddingRight) / 2} ${zoom < 1 ? height - paddingBottom : keyShapeLocal.min[1] - y}`;
                break;
            case 'left':
                labelBackgroundShape.style.transformOrigin = `${zoom < 1 ? width - paddingRight : keyShapeLocal.min[0] - x} ${paddingTop + (height - paddingTop - paddingBottom) / 2}`;
                break;
            case 'right':
                labelBackgroundShape.style.transformOrigin = `${zoom < 1 ? paddingLeft : keyShapeLocal.max[0] - x} ${paddingTop + (height - paddingTop - paddingBottom) / 2}`;
                break;
            case 'bottom':
                labelBackgroundShape.style.transformOrigin = `${paddingLeft + (width - paddingLeft - paddingRight) / 2} ${zoom < 1
                    ? paddingTop + (height - paddingTop - paddingBottom) / 2
                    : keyShapeLocal.max[1] - y}`;
                break;
            default:
                // center
                labelBackgroundShape.style.transformOrigin = `${paddingLeft + (width - paddingLeft - paddingRight) / 2} ${paddingTop + (height - paddingTop - paddingBottom) / 2}`;
        }
        const { labelShapeGeometry: labelBBox } = this.boundsCache;
        const labelWidth = labelBBox.max[0] - labelBBox.min[0];
        const xAxistRatio = (labelWidth * balanceRatio + paddingLeft + paddingRight) / width;
        labelBackgroundShape.style.transform = `scale(${xAxistRatio}, ${balanceRatio})`;
    }
    /**
     * Create (if does not exit in shapeMap) or update the shape according to the configurations.
     * @param type shape's type
     * @param id unique string to indicates the shape
     * @param style style to be updated
     * @param shapeMap the shape map of a node / combo
     * @param model data model of the node / combo
     * @returns The display object representing the shape.
     */
    upsertShape(type, id, style, shapeMap, model) {
        return (0, shape_1.upsertShape)(type, id, style, shapeMap, model);
    }
}
exports.BaseNode = BaseNode;
//# sourceMappingURL=base.js.map