import { DisplayObject } from '@antv/g';
import { IGraph } from '../../../types';
import { IG6GraphEvent } from '../../../types/event';
import { ShapeStyle } from '../../../types/item';
import { Plugin as Base, IPluginBaseConfig } from '../../../types/plugin';
/**
 * This is an interface named `FisheyeConfig`, which extends the `IPluginBaseConfig` interface. It contains the following properties:
- `trigger`: The trigger method, which can be `'mousemove'`, `'click'`, or `'drag'`.
- `d`: A number representing the magnification factor of the fisheye.
- `r`: A number representing the radius of the fisheye.
- `delegateStyle`: The shape style.
- `showLabel`: A boolean indicating whether to show the label.
- `scaleRBy`: Can be `'wheel'`, `'drag'`, `'unset'`, or `undefined`, representing the scaling method for the fisheye radius.
- `scaleDBy`: Can be `'wheel'`, `'drag'`, `'unset'`, or `undefined`, representing the scaling method for the fisheye magnification factor.
- `maxR`: A number representing the maximum value of the fisheye radius.
- `minR`: A number representing the minimum value of the fisheye radius.
- `maxD`: A number representing the maximum value of the fisheye magnification factor.
- `minD`: A number representing the minimum value of the fisheye magnification factor.
- `throttle`: A number representing the throttle time (in milliseconds).
- `showDPercent`: A boolean indicating whether to show the percentage of the fisheye magnification factor.
 */
export interface FisheyeConfig extends IPluginBaseConfig {
    /** The trigger method, which can be `'mousemove'`, `'click'`, or `'drag'`. */
    trigger?: 'mousemove' | 'click' | 'drag';
    /**  A number representing the magnification factor of the fisheye. */
    d?: number;
    /**  A number representing the radius of the fisheye. */
    r?: number;
    /** The shape style */
    delegateStyle?: ShapeStyle;
    /** A boolean indicating whether to show the label. */
    showLabel?: boolean;
    /**  Can be `'wheel'`, `'drag'`, `'unset'`, or `undefined`, representing the scaling method for the fisheye radius. */
    scaleRBy?: 'wheel' | 'drag' | 'unset' | undefined;
    /** Can be `'wheel'`, `'drag'`, `'unset'`, or `undefined`, representing the scaling method for the fisheye magnification factor. */
    scaleDBy?: 'wheel' | 'drag' | 'unset' | undefined;
    /**  A number representing the maximum value of the fisheye radius. */
    maxR?: number;
    /** A number representing the minimum value of the fisheye radius. */
    minR?: number;
    /** A number representing the maximum value of the fisheye magnification factor. */
    maxD?: number;
    /** A number representing the minimum value of the fisheye magnification factor. */
    minD?: number;
    /**  A number representing the throttle time (in milliseconds). */
    throttle?: number;
    /**  A boolean indicating whether to show the percentage of the fisheye magnification factor. */
    showDPercent?: boolean;
}
export declare class Fisheye extends Base {
    private pointerDown;
    private cacheCenter;
    private dragPrePos;
    private dragging;
    private molecularParam;
    private dPercentText;
    private delegate;
    private r2;
    private cachedOriginPositions;
    private cachedTransientTexts;
    private delegateCenterDiff;
    constructor(options?: FisheyeConfig);
    getDefaultCfgs(): FisheyeConfig;
    getEvents(): {
        click: (e: IG6GraphEvent, mousePos?: any) => void;
    } | {
        pointermove: (e: IG6GraphEvent, mousePos?: any) => void;
    };
    init(graph: IGraph): void;
    private initOriginPositions;
    protected isInLensDelegate(lensDelegate: any, pointer: any): boolean;
    protected onPointerUp(e: IG6GraphEvent): void;
    protected onPointerMove(e: IG6GraphEvent): void;
    protected onPointerDown(e: IG6GraphEvent): void;
    protected moveDelegate(lensDelegate: any, e: any): void;
    protected onWheel(e: IG6GraphEvent): void;
    /**
     * Scale the range by wheel
     * @param e mouse wheel event
     */
    protected scaleRByWheel(e: IG6GraphEvent): void;
    /**
     * Scale the range by dragging
     * @param e mouse event
     */
    protected scaleRByDrag: Function;
    protected scaleRByDragMethod(e: IG6GraphEvent): void;
    /**
     * Scale the magnifying factor by wheel
     * @param e mouse wheel event
     */
    protected scaleDByWheel(evt: IG6GraphEvent): void;
    /**
     * Scale the magnifying factor by dragging
     * @param e mouse event
     */
    protected scaleDByDrag: Function;
    protected scaleDByDragMethod(e: IG6GraphEvent): void;
    /**
     * Response function for mousemove, click, or drag to magnify
     * @param e mouse event
     */
    protected magnify(e: IG6GraphEvent, mousePos?: any): void;
    protected restoreCache(): {
        [x: string]: {
            x: number;
            y: number;
            texts: DisplayObject<any, any>[];
        };
    };
    /**
     * Adjust part of the parameters, including trigger, d, r, maxR, minR, maxD, minD, scaleRBy, and scaleDBy
     * @param {FisheyeConfig} cfg
     */
    updateParams(cfg: FisheyeConfig): void;
    /**
     * Update the delegate shape of the lens
     * @param {Point} mCenter the center of the shape
     * @param {number} r the radius of the shape
     */
    private updateDelegate;
    /**
     * Clear the fisheye lens
     */
    clear(): void;
    /**
     * Destroy the component
     */
    destroy(): void;
}
