"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.History = exports.getCategoryByApiName = void 0;
const util_1 = require("@antv/util");
const history_1 = require("../../../types/history");
const plugin_1 = require("../../../types/plugin");
const command_1 = __importDefault(require("./command"));
const apiMap = {
    add: ['addData', 'addCombo'],
    remove: ['removeData'],
    update: [
        'updateData',
        'updateNodePosition',
        'updateComboPosition',
        'updatePosition',
        'moveCombo',
        'collapseCombo',
        'expandCombo',
    ],
    stateChange: ['setItemState', 'clearItemState'],
    layerChange: ['frontItem', 'backItem'],
    displayChange: ['showItem', 'hideItem'],
};
/**
 * Retrieve the category of a given API based on its name.
 * @param apiName name of API
 */
const getCategoryByApiName = (apiName) => {
    for (const [categoryKey, apiList] of Object.entries(apiMap)) {
        if (apiList.includes(apiName)) {
            return categoryKey;
        }
    }
    throw new Error(`Unknown apiName: ${apiName}. Unable to determine category.`);
};
exports.getCategoryByApiName = getCategoryByApiName;
class HistoryStack {
    constructor(maxStep = 0) {
        this.items = [];
        this.maxStep = maxStep;
    }
    push(item) {
        if (this.maxStep !== 0 && this.items.length === this.maxStep) {
            this.items.shift();
        }
        this.items.push(item);
    }
    pop() {
        if (this.items.length !== 0) {
            return this.items.pop();
        }
        throw new Error('Stack is empty');
    }
    size() {
        return this.items.length;
    }
    clear() {
        this.items = [];
    }
    toArray() {
        return this.items;
    }
}
class History extends plugin_1.Plugin {
    constructor(options) {
        super();
        this.stackSize = 0;
        this.stackActive = true;
        this.withoutStackingCounter = 0;
        const { enableStack, stackCfg } = (0, util_1.deepMix)(this.getDefaultCfgs(), options);
        this.enableStack = enableStack;
        this.cfg = stackCfg;
        this.isBatching = false;
        this.undoStack = new HistoryStack(this.cfg.stackSize);
        this.redoStack = new HistoryStack(this.cfg.stackSize);
        this.initBatchCommands();
    }
    getDefaultCfgs() {
        return {
            enableStack: true,
            stackCfg: {
                stackSize: 0,
                stackActive: true,
                includes: [],
                excludes: [],
                ignoreAdd: false,
                ignoreRemove: false,
                ignoreUpdate: false,
                ignoreStateChange: false,
                ignoreLayerChange: false,
                ignoreDisplayChange: false,
            },
        };
    }
    init(graph) {
        super.init(graph);
        this.clear();
    }
    getUndoStack() {
        return this.undoStack.toArray();
    }
    getRedoStack() {
        return this.redoStack.toArray();
    }
    getStack() {
        return {
            redoStack: this.getRedoStack(),
            undoStack: this.getUndoStack(),
        };
    }
    clear() {
        this.clearUndoStack();
        this.clearRedoStack();
    }
    clearUndoStack() {
        this.undoStack.clear();
    }
    clearRedoStack() {
        this.redoStack.clear();
    }
    isEnable() {
        return this.enableStack;
    }
    push(cmd, stackType = history_1.STACK_TYPE.undo, isNew = true) {
        if (this.stackActive) {
            if (stackType === history_1.STACK_TYPE.undo) {
                if (isNew) {
                    // Clear the redo stack when a new action is performed to maintain state consistency
                    this.clearRedoStack();
                }
                if ((0, util_1.isEmpty)(cmd))
                    return;
                this.undoStack.push(cmd);
                this.initBatchCommands();
            }
            else {
                this.redoStack.push(cmd);
            }
            this.graph.emit('history:change', cmd, stackType, isNew);
        }
        else {
            console.error('Stacking operations are currently paused. Unable to push to the stack.');
        }
    }
    /**
     * Pause stacking operations.
     */
    pauseStacking() {
        this.withoutStackingCounter++;
        if (this.withoutStackingCounter === 1) {
            // Only disable on the first call
            this.stackActive = false;
        }
    }
    /**
     * Resume stacking operations.
     */
    resumeStacking() {
        if (this.withoutStackingCounter > 0) {
            this.withoutStackingCounter--;
        }
        if (this.withoutStackingCounter === 0) {
            // Only enable when all pause requests are cleared
            this.stackActive = true;
        }
    }
    undo() {
        if (this.isEnable()) {
            const cmds = this.undoStack.pop();
            if (cmds) {
                this.push(cmds, history_1.STACK_TYPE.redo, false);
                for (let i = cmds.length - 1; i >= 0; i--) {
                    cmds[i].undo(this.graph);
                }
            }
        }
        return this;
    }
    redo() {
        if (this.isEnable()) {
            const cmds = this.redoStack.pop();
            if (cmds) {
                this.push(cmds, history_1.STACK_TYPE.undo, false);
                (0, util_1.each)(cmds, (cmd) => cmd.redo(this.graph));
            }
        }
        return this;
    }
    canUndo() {
        return this.isEnable() && this.undoStack.size() > 0;
    }
    canRedo() {
        return this.isEnable() && this.redoStack.size() > 0;
    }
    initBatchCommands() {
        this.batchCommands = [];
    }
    /**
     * Begin a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    startBatch() {
        if (this.isBatching) {
            throw new Error('Ensure that batch processing is stopped before starting.');
        }
        this.initBatchCommands();
        this.isBatching = true;
    }
    /**
     * End a batch operation.
     * Any operations performed between `startBatch` and `stopBatch` are grouped together.
     * treated as a single operation when undoing or redoing.
     */
    stopBatch() {
        if (!this.isBatching) {
            throw new Error('Ensure that batch processing is started before stopping.');
        }
        this.push(this.batchCommands);
        this.isBatching = false;
    }
    /**
     * Execute the provided function in a batch mode
     * @param callback
     */
    batch(callback) {
        this.startBatch();
        callback();
        this.stopBatch();
    }
    getEvents() {
        return {
            afteritemchange: this.handleUpdateHistory,
            afteritemzindexchange: this.handleUpdateHistory,
            afteritemstatechange: this.handleUpdateHistory,
            afteritemvisibilitychange: this.handleUpdateHistory,
            aftercollapsecombo: this.handleUpdateHistory,
            afterexpandcombo: this.handleUpdateHistory,
        };
    }
    handleUpdateHistory(props) {
        const { apiName, changes } = props;
        if (changes && changes.length === 0)
            return;
        if (this.shouldPushToStack(apiName)) {
            this.batchCommands = [
                ...this.batchCommands,
                ...command_1.default.create(props),
            ];
            if (this.isBatching) {
                return;
            }
            this.push(this.batchCommands);
        }
    }
    /**
     * Determine if a given operation should be pushed onto the history stack based on various configurations.
     * @param {string} apiName name of the API operation.
     * @param {boolean} stack Optional. whether push this operation into graph's stack.
     */
    shouldPushToStack(apiName) {
        const { includes = [], excludes = [] } = this.cfg;
        if (!this.enableStack || !this.stackActive)
            return false;
        if (includes.includes(apiName))
            return true;
        if (excludes.includes(apiName))
            return false;
        let categoryKey = (0, exports.getCategoryByApiName)(apiName);
        if (!categoryKey)
            return true;
        categoryKey = 'ignore' + (0, util_1.upperFirst)(categoryKey);
        return !this.cfg[categoryKey];
    }
    notify(graph, eventName, ...data) {
        graph.emit(eventName, data);
    }
}
exports.History = History;
//# sourceMappingURL=index.js.map