"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_util_1 = require("@antv/path-util");
const util_1 = require("@antv/util");
const shape_1 = require("../../../util/shape");
const path_1 = require("../../../util/path");
const util_2 = require("./util");
const convexHull_1 = require("./convexHull");
const bubbleset_1 = require("./bubbleset");
/**
 * The outline of the member used to wrap the interior.
 * convex hull(convex hull): http://geomalgorithms.com/a10-_hull-1.html#Monotone%20Chain
 * bubble: bubbleset algorithm, refer: http://vialab.science.uoit.ca/wp-content/papercite-data/pdf/col2009c.pdf
 * 通By configuring padding, you can adjust the tightness of the wrapping contour to the node.
 */
class Hull {
    constructor(graph, options) {
        const { members: memberIds = [], nonMembers: nonMemberIds = [] } = options;
        this.options = options;
        this.graph = graph;
        this.members = memberIds.map((id) => graph.getNodeData(id) || graph.getComboData(id));
        this.nonMembers = nonMemberIds.map((id) => graph.getNodeData(id) || graph.getComboData(id));
        this.setPadding();
        this.setType();
        this.path = this.getPath(this.members, this.nonMembers);
        this.render();
    }
    getDefaultCfg() {
        return {
            id: 'g6-hull',
            type: 'round-convex',
            members: [],
            nonMembers: [],
        };
    }
    setPadding() {
        let nodeSize = 30;
        if (this.members.length) {
            const firstBBox = this.graph.getRenderBBox(this.members[0].id);
            if (firstBBox)
                nodeSize = firstBBox.halfExtents[0];
        }
        this.padding =
            this.options.padding > 0
                ? this.options.padding + nodeSize
                : 10 + nodeSize;
        this.bubbleCfg = Object.assign(Object.assign({}, this.options.bubbleCfg), { nodeR0: this.padding - nodeSize, nodeR1: this.padding - nodeSize, morphBuffer: this.padding - nodeSize });
    }
    setType() {
        this.type = this.options.type;
        if (this.members.length < 3) {
            this.type = 'round-convex';
        }
        if (this.type !== 'round-convex' &&
            this.type !== 'smooth-convex' &&
            this.type !== 'bubble') {
            console.warn('The hull type should be either round-convex, smooth-convex or bubble, round-convex is used by default.');
            this.type = 'round-convex';
        }
    }
    getPath(members, nonMembers) {
        let contour, path, hull;
        switch (this.type) {
            case 'round-convex':
                contour = (0, convexHull_1.genConvexHull)(members);
                hull = (0, util_2.roundedHull)(contour.map((p) => [p.x, p.y]), this.padding);
                path = (0, path_util_1.parsePathString)(hull);
                break;
            case 'smooth-convex':
                contour = (0, convexHull_1.genConvexHull)(members);
                if (contour.length === 2) {
                    hull = (0, util_2.roundedHull)(contour.map((p) => [p.x, p.y]), this.padding);
                    path = (0, path_util_1.parsePathString)(hull);
                }
                else if (contour.length > 2) {
                    hull = (0, util_2.paddedHull)(contour.map((p) => [p.x, p.y]), this.padding);
                    path = (0, util_2.getClosedSpline)(hull);
                }
                break;
            case 'bubble':
                contour = (0, bubbleset_1.genBubbleSet)(this.graph, members, nonMembers, this.bubbleCfg);
                path = contour.length >= 2 && (0, util_2.getClosedSpline)(contour);
                break;
            default:
        }
        return path;
    }
    render() {
        if (!this.path)
            return;
        const shape = this.graph.drawTransient('path', this.options.id, {
            style: Object.assign({ path: this.path }, this.options.style),
            capture: false,
        });
        const { labelShape } = this.options;
        if (labelShape === null || labelShape === void 0 ? void 0 : labelShape.text) {
            const shapeBounds = shape.getRenderBounds();
            const { position, offsetX, offsetY } = labelShape, labelShapeStyle = __rest(labelShape, ["position", "offsetX", "offsetY"]);
            const formattedLabelStyle = this.getLabelStyle(labelShape, shapeBounds);
            this.graph.drawTransient('text', `${this.options.id}-label`, {
                style: Object.assign(Object.assign({}, labelShapeStyle), formattedLabelStyle),
                capture: false,
            });
        }
        shape.toBack();
        this.hullShape = shape;
    }
    getLabelStyle(propsStyle, keyShapeBounds) {
        const { position, offsetX = 0, offsetY = 0, autoRotate = true, } = propsStyle;
        const { min, max, center } = keyShapeBounds;
        const style = {
            x: 0,
            y: 0,
            textAlign: 'center',
            textBaseline: 'middle',
        };
        const point = { x: 0, y: 0 };
        switch (position) {
            case 'bottom':
            case 'outside-bottom':
                point.x = (min[0] + max[0]) / 2;
                point.y = max[1];
                style.textBaseline = 'bottom';
                break;
            case 'left':
            case 'outside-left':
                point.x = min[0];
                point.y = (min[1] + max[1]) / 2;
                style.textAlign = 'left';
                break;
            case 'right':
            case 'outside-right':
                point.x = max[0];
                point.y = (min[1] + max[1]) / 2;
                style.textAlign = 'right';
                break;
            case 'top':
            case 'outside-top':
            default:
                point.x = (min[0] + max[0]) / 2;
                point.y = min[1];
                style.textBaseline = 'top';
                break;
        }
        let startPoint;
        const points = (0, path_1.pathToPoints)(this.path);
        if (this.type === 'bubble') {
            switch (position) {
                case 'bottom':
                    startPoint = { x: center[0], y: max[0] + 10 };
                    break;
                case 'left':
                    startPoint = { x: max[0] + 10, y: center[1] };
                    break;
                case 'right':
                    startPoint = { x: min[0] - 10, y: center[1] };
                    break;
                case 'top':
                default:
                    startPoint = { x: center[0], y: max[1] - 10 };
                    break;
            }
        }
        else {
            startPoint = { x: center[0], y: center[1] };
        }
        const extendedPoint = (0, util_2.getExtendPoint)(point, startPoint, 100);
        const intersect = (0, util_2.getPolygonIntersectByLine)(points, [
            startPoint,
            extendedPoint,
        ]);
        if (intersect) {
            const { point: intersectPoint, line } = intersect;
            style.x = intersectPoint.x + offsetX;
            style.y = intersectPoint.y + offsetY;
            if (autoRotate) {
                const angle = Math.atan((line[0].y - line[1].y) / (line[0].x - line[1].x));
                style.transform = `rotate(${(angle / Math.PI) * 180}deg)`;
                style.textAlign = 'center';
                if (position === 'right' || position === 'left') {
                    if (angle > 0) {
                        style.textBaseline = position === 'right' ? 'top' : 'bottom';
                    }
                    else {
                        style.textBaseline = position === 'right' ? 'bottom' : 'top';
                    }
                }
            }
        }
        else {
            style.x = point.x + offsetX;
            style.y = point.y + offsetY;
        }
        if (position.split('-')[0] === 'outside') {
            // @ts-ignore
            style.textBaseline = (0, util_2.getInvertBaseline)(style.textBaseline);
        }
        return style;
    }
    /**
     * Add a member of hull, and if the member was originally in nonMembers, remove it from nonMembers.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Whether successfully added.
     */
    addMember(itemIds) {
        if (!itemIds)
            return;
        const idArr = (0, util_1.isArray)(itemIds) ? itemIds : [itemIds];
        let shouldUpdate = false;
        idArr.forEach((id) => {
            const model = this.graph.getNodeData(id) || this.graph.getComboData(id);
            const memberIdx = this.members.indexOf(model);
            if (memberIdx === -1) {
                this.members.push(model);
                shouldUpdate = true;
            }
            const nonMemberIdx = this.nonMembers.indexOf(model);
            if (nonMemberIdx > -1) {
                this.nonMembers.splice(nonMemberIdx, 1);
                shouldUpdate = true;
            }
        });
        if (shouldUpdate) {
            this.updateMembers(this.members.map((model) => model.id), this.nonMembers.map((model) => model.id));
        }
        return false;
    }
    /**
     * Add the nodes that the hull needs to exclude, and if the member was originally in the members, remove it from the members.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Whether successfully added.
     */
    addNonMember(itemIds) {
        if (!itemIds)
            return;
        const idArr = (0, util_1.isArray)(itemIds) ? itemIds : [itemIds];
        let shouldUpdate = false;
        idArr.forEach((id) => {
            const model = this.graph.getNodeData(id) || this.graph.getComboData(id);
            const nonMemberIdx = this.nonMembers.indexOf(model);
            if (nonMemberIdx === -1) {
                this.nonMembers.push(model);
                shouldUpdate = true;
            }
            const memberIdx = this.members.indexOf(model);
            if (memberIdx > -1) {
                this.members.splice(memberIdx, 1);
                shouldUpdate = true;
            }
        });
        if (shouldUpdate) {
            this.updateMembers(this.members.map((model) => model.id), this.nonMembers.map((model) => model.id));
            return true;
        }
        return false;
    }
    /**
     * Remove members from the hull.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Returns true if the removal is successful, otherwise returns false.
     */
    removeMember(itemIds) {
        if (!itemIds)
            return;
        const idArr = (0, util_1.isArray)(itemIds) ? itemIds : [itemIds];
        let removed = false;
        idArr.forEach((id) => {
            const index = this.members.findIndex((member) => member.id === id);
            if (index > -1) {
                this.members.splice(index, 1);
                removed = true;
            }
        });
        if (removed) {
            this.updateMembers(this.members.map((model) => model.id), this.nonMembers.map((model) => model.id));
            return true;
        }
        return false;
    }
    /**
     * Remove non members from the hull.
     * @param itemIds id(s) of member nodes/combos
     * @return boolean Returns true if the removal is successful, otherwise returns false.
     */
    removeNonMember(itemIds) {
        if (!itemIds)
            return;
        const idArr = (0, util_1.isArray)(itemIds) ? itemIds : [itemIds];
        let removed = false;
        idArr.forEach((id) => {
            const model = this.graph.getNodeData(id) || this.graph.getComboData(id);
            const index = this.nonMembers.indexOf(model);
            if (index > -1) {
                this.nonMembers.splice(index, 1);
                removed = true;
            }
        });
        if (removed) {
            this.updateMembers(this.members.map((model) => model.id), this.nonMembers.map((model) => model.id));
            return true;
        }
        return false;
    }
    updateMembers(memberIds, nonMemberIds) {
        if (memberIds) {
            this.options.members = memberIds;
            this.members = memberIds.map((id) => this.graph.getNodeData(id) || this.graph.getComboData(id));
        }
        if (nonMemberIds) {
            this.options.nonMembers = nonMemberIds;
            this.nonMembers = nonMemberIds.map((id) => this.graph.getNodeData(id) || this.graph.getComboData(id));
        }
        this.path = this.getPath(this.members, this.nonMembers);
        this.render();
    }
    updateStyle(style) {
        this.graph.drawTransient('path', this.options.id, {
            style,
            capture: false,
        });
    }
    /**
     * Update the options
     * @param options hull options
     */
    updateOptions(options) {
        const { members: memberIds, nonMembers: nonMemberIds } = options;
        this.options = Object.assign(Object.assign(Object.assign({}, this.options), options), { style: Object.assign(Object.assign({}, this.options.style), options.style), labelShape: Object.assign(Object.assign({}, this.options.labelShape), options.labelShape) });
        if (memberIds) {
            this.members = memberIds.map((id) => this.graph.getNodeData(id) || this.graph.getComboData(id));
        }
        if (nonMemberIds) {
            this.nonMembers = nonMemberIds.map((id) => this.graph.getNodeData(id) || this.graph.getComboData(id));
        }
        this.setPadding();
        this.setType();
        this.path = this.getPath(this.members, this.nonMembers);
        this.render();
    }
    /**
     * Whether a node/edge is inside the range of the hull.
     * @param itemId
     */
    contain(itemId) {
        let shapePoints;
        // TODO
        // const shape = nodeItem.getKeyShape();
        // if (nodeItem.get('type') === 'path') {
        //   shapePoints = pathToPoints(shape.attr('path'));
        // } else {
        const shapeBBox = this.graph.getRenderBBox(itemId);
        if (!shapeBBox) {
            console.warn(`The item with ${itemId} is not existed in the graph.`);
            return;
        }
        shapePoints = [
            [shapeBBox.min[0], shapeBBox.min[1]],
            [shapeBBox.max[0], shapeBBox.min[1]],
            [shapeBBox.max[0], shapeBBox.max[1]],
            [shapeBBox.min[0], shapeBBox.max[1]],
        ];
        // }
        shapePoints = shapePoints.map((canvasPoint) => {
            const point = this.graph.getPointByCanvas(canvasPoint[0], canvasPoint[1]);
            return [point.x, point.y];
        });
        return (0, shape_1.isPolygonsIntersect)(shapePoints, (0, path_1.pathToPoints)(this.path));
    }
    destroy() {
        this.graph.drawTransient('path', this.options.id, { action: 'remove' });
    }
}
exports.default = Hull;
//# sourceMappingURL=hullComponent.js.map