import { Canvas } from '@antv/g';
import { IGraph } from '../../../types';
import { ShapeStyle } from '../../../types/item';
import { Plugin as Base, IPluginBaseConfig } from '../../../types/plugin';
export interface MiniMapConfig extends IPluginBaseConfig {
    /** Class name of viewport */
    viewportClassName?: string;
    /** Class name of minimap */
    className?: string;
    /** Mode of minimap */
    mode?: 'default' | 'keyShape' | 'delegate';
    /** Size of minimap */
    size?: number[];
    /** Style of delegate shape */
    delegateStyle?: ShapeStyle;
    /** Whether to refresh minimap */
    refresh?: boolean;
    /** Padding of minimap */
    padding?: number;
    /** Whether to hide edges on minimap to enhance performance */
    hideEdge?: boolean;
    /** Container for minimap */
    container?: HTMLDivElement | null;
}
export declare class Minimap extends Base {
    private canvas;
    /** The viewport DOM on the minimap. */
    private viewport;
    /** Cache the mapping of graphics of nodes/edges/combos on main graph and minimap graph. */
    private itemMap;
    private container;
    /** Ratio of (minimap graph size / main graph size). */
    private ratio;
    /** Distance from top of minimap graph to the top of minimap container. */
    private dx;
    /** Distance from left of minimap graph to the left of minimap container. */
    private dy;
    constructor(options?: MiniMapConfig);
    getDefaultCfgs(): MiniMapConfig;
    getEvents(): {
        afterupdateitem: () => void;
        afteritemstatechange: () => void;
        afterlayout: () => void;
        viewportchange: () => void;
        afteritemchange: () => void;
    };
    /**
     * If it is animating, disable refresh.
     */
    protected disableRefresh(): void;
    protected enableRefresh(): void;
    private initViewport;
    /**
     * Update the viewport DOM.
     */
    private updateViewport;
    /**
     * Clone all the graphic from main graph to the minimap graph.
     */
    private updateGraphShapes;
    /**
     * Only draw keyShapes on the minimap.
     */
    private updateKeyShapes;
    /**
     * Add or update keyShape of one node.
     * @param nodeModel node data model
     * @param group container graphics group on minimap
     */
    private updateOneNodeKeyShape;
    /**
     * Draw the delegate rects for nodes and line edges on minimap.
     */
    private updateDelegateShapes;
    private clearDestroyedShapes;
    /**
     * Add or update keyShape of one edge.
     * @param edgeModel edge data model
     * @param group container graphics group on minimap
     */
    private updateOneEdgeKeyShape;
    /**
     * Add or update delegate rect of one node.
     * @param nodeModel node data model
     * @param group container graphics group on minimap
     */
    private updateOneNodeDelegateShape;
    /**
     * Listener for main graph updating, update the viewport DOM.
     */
    private handleUpdateCanvas;
    init(graph: IGraph): void;
    /**
     * Init the DOM container for minimap.
     */
    initContainer(): Promise<any>;
    updateCanvas(): void;
    /**
     * Get the canvas of the minimap.
     * @return {Canvas} G Canvas
     */
    getCanvas(): Canvas;
    /**
     * Get the viewport DOM of the minimap.
     * @return {HTMLElement} viewport DOM
     */
    getViewport(): HTMLElement;
    /**
     * Get the container DOM of the minimap.
     * @return {HTMLElement} container DOM
     */
    getContainer(): HTMLElement;
    destroy(): void;
}
