"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.traverse = exports.graphData2TreeData = exports.treeData2GraphData = exports.validateComboStrucutre = exports.isSucceed = exports.traverseAncestors = exports.traverseGraphAncestors = exports.traverseAncestorsAndSucceeds = exports.graphComboTreeDfs = exports.graphCoreTreeDfs = exports.deconstructData = void 0;
const util_1 = require("@antv/util");
const algorithm_1 = require("@antv/algorithm");
/**
 * Deconstruct data and distinguish nodes and combos from graphcore data.
 * @param data data from graphcore
 * @returns
 */
const deconstructData = (data) => {
    const { nodes: nodesAndCombos = [], edges = [] } = data;
    const nodes = [];
    const combos = [];
    nodesAndCombos.forEach((item) => {
        if (item.data._isCombo)
            combos.push(item);
        else
            nodes.push(item);
    });
    return {
        nodes,
        edges,
        combos,
    };
};
exports.deconstructData = deconstructData;
/**
 * Depth first search begin from nodes in graphCore data.
 * @param graphCore graphlib data structure
 * @param nodes begin nodes
 * @param fn will be called while visiting each node
 * @param mode 'TB' - visit from top to bottom; 'BT' - visit from bottom to top;
 * @returns
 */
const graphCoreTreeDfs = (graphCore, nodes, fn, mode = 'TB', treeKey = 'combo', stopFns = {}) => {
    if (!(nodes === null || nodes === void 0 ? void 0 : nodes.length))
        return;
    const { stopBranchFn, stopAllFn } = stopFns;
    for (let i = 0; i < nodes.length; i++) {
        const node = nodes[i];
        if (!graphCore.hasNode(node.id))
            continue;
        if (stopBranchFn === null || stopBranchFn === void 0 ? void 0 : stopBranchFn(node))
            continue; // Stop this branch
        if (stopAllFn === null || stopAllFn === void 0 ? void 0 : stopAllFn(node))
            return; // Stop all
        if (mode === 'TB')
            fn(node); // Traverse from top to bottom
        (0, exports.graphCoreTreeDfs)(graphCore, graphCore.getChildren(node.id, treeKey), fn, mode, treeKey, stopFns);
        if (mode !== 'TB')
            fn(node); // Traverse from bottom to top
    }
};
exports.graphCoreTreeDfs = graphCoreTreeDfs;
/**
 * Depth first search begin from nodes in g6 graph data.
 * @param graph G6 graph instance
 * @param nodes begin nodes
 * @param fn will be called while visiting each node
 * @param mode 'TB' - visit from top to bottom; 'BT' - visit from bottom to top;
 * @returns
 */
const graphComboTreeDfs = (graph, nodes, fn, mode = 'TB') => {
    if (!(nodes === null || nodes === void 0 ? void 0 : nodes.length))
        return;
    nodes.forEach((node) => {
        if (mode === 'TB')
            fn(node); // Traverse from top to bottom
        (0, exports.graphComboTreeDfs)(graph, graph.getComboChildrenData(node.id), fn, mode);
        if (mode !== 'TB')
            fn(node); // Traverse from bottom to top
    });
};
exports.graphComboTreeDfs = graphComboTreeDfs;
/**
 * Depth first search begin from nodes in g6 graph data, and then search the ancestirs from begin nodes.
 * @param graph G6 graph instance
 * @param graphCore graphlib data structure
 * @param nodes begin ndoes
 * @param fn will be called while visiting each node
 * @param mode 'TB' - visit from top to bottom; 'BT' - visit from bottom to top;
 * @returns
 */
const traverseAncestorsAndSucceeds = (graph, graphCore, nodes, fn, mode = 'TB') => {
    if (!(nodes === null || nodes === void 0 ? void 0 : nodes.length))
        return;
    (0, exports.graphComboTreeDfs)(graph, nodes, fn, mode);
    (0, exports.traverseAncestors)(graphCore, nodes, fn);
    return;
};
exports.traverseAncestorsAndSucceeds = traverseAncestorsAndSucceeds;
const traverseGraphAncestors = (graph, nodes, fn) => {
    if (!(nodes === null || nodes === void 0 ? void 0 : nodes.length))
        return;
    nodes.forEach((node) => {
        if (!node.data.parentId)
            return;
        let ancestor = graph.getComboData(node.data.parentId);
        while (ancestor) {
            if (fn(ancestor))
                return;
            if (!ancestor.data.parentId)
                return;
            ancestor = graph.getComboData(ancestor.data.parentId);
        }
    });
};
exports.traverseGraphAncestors = traverseGraphAncestors;
/**
 * Traverse the ancestors from the begin nodes.
 * @param graphCore graphlib data structure
 * @param nodes begin ndoes
 * @param fn will be called while visiting each node
 */
const traverseAncestors = (graphCore, nodes, fn) => {
    nodes.forEach((node) => {
        let ancestor = graphCore.getParent(node.id, 'combo');
        while (ancestor) {
            if (fn(ancestor))
                return;
            ancestor = graphCore.getParent(ancestor.id, 'combo');
        }
    });
};
exports.traverseAncestors = traverseAncestors;
/**
 * Whether the test succeed is a parent of the test parent.
 * @param graph G6 graph instance
 * @param testParent id of the node to be the parent
 * @param parentId id of the node to be the succeed
 * @returns
 */
const isSucceed = (graph, testParent, testSucceed) => {
    const succeedIds = [];
    if (graph.getNodeData(testParent))
        return false;
    (0, exports.graphComboTreeDfs)(graph, [graph.getComboData(testParent)], (node) => {
        succeedIds.push(node.id);
    });
    if (succeedIds.includes(testSucceed))
        return true;
    return false;
};
exports.isSucceed = isSucceed;
/**
 * Check the validation if move the child into the subtree of parent. e.g. If the new parent (toBeAncestorId) is a succeed of new child (toBeSucceedId), it is invalid.
 * @param graph G6 graph instance
 * @param toBeSucceedId id of the new child to check if it is valid to be a child of the node with id toBeAncestorId
 * @param toBeAncestorId id of the new parent to check if it is valid to be a parent of the node with id toBeSucceedId
 * @returns
 */
const validateComboStrucutre = (graph, toBeSucceedId, toBeAncestorId) => {
    if (toBeAncestorId && !graph.getComboData(toBeAncestorId)) {
        console.warn(`Setting parent combo failed. The parent combo with id ${toBeAncestorId} does not exist`);
        return false;
    }
    if (toBeSucceedId === toBeAncestorId) {
        console.warn(`Setting parent combo failed. Cannot set combo/node with id ${toBeSucceedId} to be the child of itself.`);
        return false;
    }
    if (toBeAncestorId && (0, exports.isSucceed)(graph, toBeSucceedId, toBeAncestorId)) {
        console.warn(`Setting parent combo failed, since the parent combo with id ${toBeAncestorId} is a succeed of the combo with id ${toBeSucceedId}.`);
        return false;
    }
    return true;
};
exports.validateComboStrucutre = validateComboStrucutre;
/**
 * Transform tree graph data into graph data. Edges from parent-child structure.
 * @param treeData Tree structured data or an array of it.
 * @returns Graph formatted data object with nodes, edges and combos.
 */
const treeData2GraphData = (treeData) => {
    const graphData = {
        nodes: [],
        edges: [],
        combos: [],
    };
    const trees = (0, util_1.isArray)(treeData) ? treeData : [treeData];
    trees.forEach((tree) => {
        (0, exports.traverse)(tree, (child) => {
            var _a, _b;
            const { id, children, data } = child, others = __rest(child, ["id", "children", "data"]);
            graphData.nodes.push({
                id: child.id,
                data: Object.assign(Object.assign({ childrenIds: ((_a = child.children) === null || _a === void 0 ? void 0 : _a.map((c) => c.id)) || [] }, others), data),
            });
            (_b = child.children) === null || _b === void 0 ? void 0 : _b.forEach((subChild) => {
                graphData.edges.push({
                    id: `tree-edge-${child.id}-${subChild.id}`,
                    source: child.id,
                    target: subChild.id,
                    data: {},
                });
            });
        });
    });
    return graphData;
};
exports.treeData2GraphData = treeData2GraphData;
/**
 * Transform graph data into tree graph data.
 * @param nodeMap This function will cache the nodes into nodeMap
 * @param graphData Graph data.
 * @param propRootIds Ids of root nodes. There should be at least one node for each connected component, or the first node in a connected component will be added to the roots array.
 * @returns
 */
const graphData2TreeData = (nodeMap, graphData, propRootIds = []) => {
    var _a;
    const trees = [];
    const graphDataWithoutCombos = {
        nodes: (_a = graphData.nodes) === null || _a === void 0 ? void 0 : _a.filter((node) => !node.data._isCombo),
        edges: graphData.edges,
    };
    const connectedComponents = (0, algorithm_1.connectedComponent)(graphDataWithoutCombos, false);
    const rootIds = [];
    const componentsNodeIds = [];
    connectedComponents.forEach((com, i) => {
        componentsNodeIds[i] = com.map((node) => node.id);
        if (propRootIds.length) {
            const root = componentsNodeIds[0].find((id) => propRootIds.includes(id));
            rootIds.push(root !== undefined ? root : com[0].id);
        }
        else {
            rootIds.push(com[0].id);
        }
    });
    rootIds.forEach((id, i) => {
        nodeMap[id] = { id, children: [] };
        trees.push(nodeMap[id]);
        (0, algorithm_1.depthFirstSearch)(graphData, id, {
            enter: ({ previous, current }) => {
                if (!previous ||
                    current === id ||
                    !componentsNodeIds[i].includes(current))
                    return;
                nodeMap[previous] = nodeMap[previous] || {
                    id: previous,
                    children: [],
                };
                nodeMap[current] = { id: current, children: [] };
                nodeMap[previous].children.push(nodeMap[current]);
            },
        }, false);
    });
    return trees;
};
exports.graphData2TreeData = graphData2TreeData;
/**
 * Travere a tree data from top to bottom.
 * @param treeData
 * @param callback
 */
const traverse = (treeData, callback) => {
    callback(treeData);
    if (treeData.children) {
        treeData.children.forEach((child) => {
            if (child)
                (0, exports.traverse)(child, callback);
        });
    }
};
exports.traverse = traverse;
//# sourceMappingURL=data.js.map