"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getGroupedChanges = exports.getContextMenuEventProps = exports.getItemInfoFromElement = void 0;
/**
 * Given an element, which might be a child shape of a node/edge,
 * get its belonging item.
 */
const getItemInfoFromElement = (element) => {
    var _a;
    if (element.nodeName === 'document') {
        return {
            itemType: 'canvas',
            itemId: undefined,
            groupElement: element,
        };
    }
    let parent = element;
    while (parent && !((_a = parent.getAttribute) === null || _a === void 0 ? void 0 : _a.call(parent, 'data-item-type'))) {
        parent = parent.parentElement;
    }
    if (!parent)
        return null;
    return {
        itemType: parent.getAttribute('data-item-type'),
        itemId: parent.getAttribute('data-item-id'),
        groupElement: parent,
    };
};
exports.getItemInfoFromElement = getItemInfoFromElement;
/**
 * Returns the event props for graph's contextmenu event.
 * @param event Original g6 event props.
 * @param graph Graph instance.
 * @returns Contextmenu event props.
 */
const getContextMenuEventProps = (event, graph) => {
    return Object.assign(Object.assign({}, event), { type: 'contextmenu', preventDefault: () => {
            graph.canvas
                .getContextService()
                .getDomElement()
                .addEventListener('contextmenu', (e) => {
                e.preventDefault();
            }, { once: true });
        } });
};
exports.getContextMenuEventProps = getContextMenuEventProps;
/**
 * Group the changes with change types from graphCore's changes.
 * @param graphCore
 * @param changes
 * @returns
 */
const getGroupedChanges = (graphCore, changes) => {
    const groupedChanges = {
        NodeRemoved: [],
        EdgeRemoved: [],
        NodeAdded: [],
        EdgeAdded: [],
        NodeDataUpdated: [],
        EdgeUpdated: [],
        EdgeDataUpdated: [],
        TreeStructureChanged: [],
        ComboStructureChanged: [],
    };
    changes.forEach((change) => {
        const { type: changeType } = change;
        if (['NodeDataUpdated', 'EdgeUpdated', 'EdgeDataUpdated'].includes(changeType)) {
            const { id: oid } = change;
            if (!graphCore.hasNode(oid) && !graphCore.hasEdge(oid)) {
                const nid = Number(oid);
                if ((!isNaN(nid) && graphCore.hasNode(nid)) || graphCore.hasEdge(nid)) {
                    groupedChanges[changeType].push(Object.assign(Object.assign({}, change), { id: nid }));
                }
                return;
            }
        }
        else if (changeType === 'TreeStructureChanged') {
            if (change.treeKey === 'combo')
                groupedChanges.ComboStructureChanged.push(change);
            else if (change.treeKey === 'tree')
                groupedChanges.TreeStructureChanged.push(change);
            return;
        }
        else if (['NodeRemoved', 'EdgeRemoved'].includes(changeType)) {
            groupedChanges[changeType].push(change);
        }
        else {
            const { id: oid } = change.value;
            if (!graphCore.hasNode(oid) && !graphCore.hasEdge(oid)) {
                const nid = Number(oid);
                if ((!isNaN(nid) && graphCore.hasNode(nid)) || graphCore.hasEdge(nid)) {
                    groupedChanges[changeType].push(Object.assign(Object.assign({}, change), { value: Object.assign(Object.assign({}, change.value), { id: nid }) }));
                }
                return;
            }
        }
        groupedChanges[changeType].push(change);
    });
    return groupedChanges;
};
exports.getGroupedChanges = getGroupedChanges;
//# sourceMappingURL=event.js.map