"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isBending = exports.getLineIntersect = exports.getRectIntersectByPoint = exports.getEllipseIntersectByPoint = exports.getCircleIntersectByPoint = exports.isSamePoint = exports.distance = exports.distanceVec = exports.getNearestPoint = void 0;
const math_1 = require("./math");
/**
 * Find the nearest on in points to the curPoint.
 * @param points
 * @param curPoint
 * @returns
 */
const getNearestPoint = (points, curPoint) => {
    let index = 0;
    let nearestPoint = points[0];
    let minDistance = (0, exports.distance)(points[0], curPoint);
    for (let i = 0; i < points.length; i++) {
        const point = points[i];
        const dis = (0, exports.distance)(point, curPoint);
        if (dis < minDistance) {
            nearestPoint = point;
            minDistance = dis;
            index = i;
        }
    }
    return {
        index,
        nearestPoint,
    };
};
exports.getNearestPoint = getNearestPoint;
/**
 * Get distance by two points.
 * @param {vec2} p1 first point
 * @param {vec2} p2 second point
 */
const distanceVec = (p1, p2) => {
    const vx = p1[0] - p2[0];
    const vy = p1[1] - p2[1];
    return Math.sqrt(vx * vx + vy * vy);
};
exports.distanceVec = distanceVec;
/**
 * Get distance by two points.
 * @param {Point} p1 first point
 * @param {Point} p2 second point
 */
const distance = (p1, p2) => {
    const vx = p1.x - p2.x;
    const vy = p1.y - p2.y;
    return Math.sqrt(vx * vx + vy * vy);
};
exports.distance = distance;
/**
 * If the p1 and p2 are the same.
 * @param p1
 * @param p2
 * @returns
 */
const isSamePoint = (p1, p2) => p1.x === p2.x && p1.y === p2.y && p1.z === p2.z;
exports.isSamePoint = isSamePoint;
/**
 * Get point and circle intersect point.
 * @param {ICircle} circle Circle's center x,y and radius r
 * @param {Point} point Point x,y
 * @return {Point} calculated intersect point
 */
const getCircleIntersectByPoint = (circleProps, point) => {
    const { x: cx, y: cy, r } = circleProps;
    const { x, y } = point;
    const dx = x - cx;
    const dy = y - cy;
    if (dx * dx + dy * dy < r * r) {
        return null;
    }
    const angle = Math.atan(dy / dx);
    return {
        x: cx + Math.abs(r * Math.cos(angle)) * Math.sign(dx),
        y: cy + Math.abs(r * Math.sin(angle)) * Math.sign(dy),
    };
};
exports.getCircleIntersectByPoint = getCircleIntersectByPoint;
/**
 * Get point and ellipse inIntersect.
 * @param {Object} ellipse ellipse center x,y and radius rx,ry
 * @param {Object} point Point x,y
 * @return {object} calculated intersect point
 */
const getEllipseIntersectByPoint = (ellipseProps, point) => {
    const { rx: a, ry: b, x: cx, y: cy } = ellipseProps;
    const dx = point.x - cx;
    const dy = point.y - cy;
    // The angle will be in range [-PI, PI]
    let angle = Math.atan2(dy / b, dx / a);
    if (angle < 0) {
        // transfer to [0, 2*PI]
        angle += 2 * Math.PI;
    }
    return {
        x: cx + a * Math.cos(angle),
        y: cy + b * Math.sin(angle),
    };
};
exports.getEllipseIntersectByPoint = getEllipseIntersectByPoint;
/**
 * Point and rectangular intersection point.
 * @param  {IRect} rect  rect
 * @param  {Point} point point
 * @return {PointPoint} rst;
 */
const getRectIntersectByPoint = (rectProps, point) => {
    const { x, y, width, height } = rectProps;
    const cx = x + width / 2;
    const cy = y + height / 2;
    const points = [];
    const center = {
        x: cx,
        y: cy,
    };
    points.push({
        x,
        y,
    });
    points.push({
        x: x + width,
        y,
    });
    points.push({
        x: x + width,
        y: y + height,
    });
    points.push({
        x,
        y: y + height,
    });
    points.push({
        x,
        y,
    });
    let rst = null;
    for (let i = 1; i < points.length; i++) {
        rst = (0, exports.getLineIntersect)(points[i - 1], points[i], center, point);
        if (rst) {
            break;
        }
    }
    return rst;
};
exports.getRectIntersectByPoint = getRectIntersectByPoint;
/**
 * Get the intersect point of two lines.
 * @param  {Point}  p0 The start point of the first line.
 * @param  {Point}  p1 The end point of the first line.
 * @param  {Point}  p2 The start point of the second line.
 * @param  {Point}  p3 The end point of the second line.
 * @return {Point}  Calculated intersect point.
 */
const getLineIntersect = (p0, p1, p2, p3) => {
    const tolerance = 0.0001;
    const E = {
        x: p2.x - p0.x,
        y: p2.y - p0.y,
    };
    const D0 = {
        x: p1.x - p0.x,
        y: p1.y - p0.y,
    };
    const D1 = {
        x: p3.x - p2.x,
        y: p3.y - p2.y,
    };
    const kross = D0.x * D1.y - D0.y * D1.x;
    const sqrKross = kross * kross;
    const invertKross = 1 / kross;
    const sqrLen0 = D0.x * D0.x + D0.y * D0.y;
    const sqrLen1 = D1.x * D1.x + D1.y * D1.y;
    if (sqrKross > tolerance * sqrLen0 * sqrLen1) {
        const s = (E.x * D1.y - E.y * D1.x) * invertKross;
        const t = (E.x * D0.y - E.y * D0.x) * invertKross;
        if (!(0, math_1.isBetween)(s, 0, 1) || !(0, math_1.isBetween)(t, 0, 1))
            return null;
        return {
            x: p0.x + s * D0.x,
            y: p0.y + s * D0.y,
        };
    }
    return null;
};
exports.getLineIntersect = getLineIntersect;
/**
 * Determine if three points are bending (not lie on a straight line)
 * @param p0 the first 2d point
 * @param p1 the second 2d point
 * @param p2 the third 2d point
 * @returns
 */
const isBending = (p0, p1, p2) => !((p0.x === p1.x && p1.x === p2.x) || (p0.y === p1.y && p1.y === p2.y));
exports.isBending = isBending;
//# sourceMappingURL=point.js.map