"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EdgeCollisionChecker = exports.QuadTree = exports.SortedArray = exports.removeRedundantPoint = exports.getPolylinePoints = exports.getPolylinePath = exports.pointsToPolyline = exports.getBorderRadiusPoints = exports.pathFinderUsingWaypoints = exports.getNeighborPoints = exports.removeFrom = exports.reconstructPath = exports.heuristicCostEstimate = exports._costByPoints = exports.simplePathFinder = exports.simplifyPolylinePoints = void 0;
const g_1 = require("@antv/g");
const util_1 = require("@antv/util");
const bbox_1 = require("./bbox");
const math_1 = require("./math");
const point_1 = require("./point");
// class PolylineHistory {
//   private pathMap: Map<ID, Point[]>;
//   constructor() {
//     this.pathMap = new Map<ID, Point[]>();
//   }
//   public addPath(edgeId: ID, points: Point[]): void {
//     this.pathMap.set(edgeId, points);
//   }
//   public getPointsById(edgeId: ID): Point[] | undefined {
//     return this.pathMap.get(edgeId);
//   }
// }
// const polylineHistory = new PolylineHistory();
/**
 * Simplify points of polyline by removing duplicated points
 * @param points list of 2d points
 * @returns unique set of points
 */
const simplifyPolylinePoints = (points) => {
    const result = [];
    const map = {};
    (0, util_1.each)(points, (p) => {
        p.id = `${p.x}|||${p.y}`;
        if (!map[p.id]) {
            map[p.id] = p;
            result.push(p);
        }
    });
    return result;
};
exports.simplifyPolylinePoints = simplifyPolylinePoints;
/**
 * Use the A* path-finding algorithm to obtain the shortest path. Considering that there is no need to implement automatic obstacle avoidance, use waypoints to plan the path
 */
const simplePathFinder = (start, end, startNode, endNode, cfg) => {
    return (0, exports.simplifyPolylinePoints)((0, exports.getPolylinePoints)(start, end, cfg.offset, startNode, endNode));
};
exports.simplePathFinder = simplePathFinder;
/**
 * 如果 points 中的一个节点 x 与 p 相等，则消耗 -2。y 同
 * 即优先选择和 points 在同一水平线 / 垂直线上的点
 */
// eslint-disable-next-line @typescript-eslint/naming-convention
const _costByPoints = (p, points) => {
    const offset = -2;
    let result = 0;
    points.forEach((point) => {
        if (point) {
            if (p.x === point.x) {
                result += offset;
            }
            if (p.y === point.y) {
                result += offset;
            }
        }
    });
    return result;
};
exports._costByPoints = _costByPoints;
/**
 * ps 经过 p 到 pt 的距离，减去其他路过节点造成的消耗
 */
const heuristicCostEstimate = (p, ps, pt, source, target) => (0, math_1.manhattanDist)(p, ps) +
    (0, math_1.manhattanDist)(p, pt) +
    (0, exports._costByPoints)(p, [ps, pt, source, target]);
exports.heuristicCostEstimate = heuristicCostEstimate;
const reconstructPath = (pathPoints, pointById, cameFrom, currentId, iterator = 0) => {
    pathPoints.unshift(pointById[currentId]);
    if (cameFrom[currentId] &&
        cameFrom[currentId] !== currentId &&
        iterator <= 100) {
        (0, exports.reconstructPath)(pathPoints, pointById, cameFrom, cameFrom[currentId], iterator + 1);
    }
};
exports.reconstructPath = reconstructPath;
/**
 * 从 arr 中删去 item
 */
const removeFrom = (arr, item) => {
    const index = arr.indexOf(item);
    if (index > -1) {
        arr.splice(index, 1);
    }
};
exports.removeFrom = removeFrom;
/**
 * 在 points 中找到满足 x 或 y 和 point 的 x 或 y 相等，且与 point 连线不经过 bbox1 与 bbox2 的点
 */
const getNeighborPoints = (points, point, bbox1, bbox2) => {
    const neighbors = [];
    points.forEach((p) => {
        if (p === point)
            return;
        if (p.x === point.x || p.y === point.y) {
            if ((0, bbox_1.isSegmentCrossingBBox)(p, point, bbox1) ||
                (0, bbox_1.isSegmentCrossingBBox)(p, point, bbox2)) {
                return;
            }
            neighbors.push(p);
        }
    });
    return (0, exports.simplifyPolylinePoints)(neighbors);
};
exports.getNeighborPoints = getNeighborPoints;
/** A-Star Algorithm using waypoints */
const pathFinderUsingWaypoints = (points, start, goal, sBBox, tBBox, os, ot) => {
    // A-Star Algorithm
    const closedSet = [];
    const openSet = {
        [start.id]: start,
    };
    const cameFrom = {};
    const gScore = {}; // all default values are Infinity
    const fScore = {}; // all default values are Infinity
    gScore[start.id] = 0;
    fScore[start.id] = (0, exports.heuristicCostEstimate)(start, goal, start);
    const sortedOpenSet = new SortedArray();
    sortedOpenSet.add({
        id: start.id,
        value: fScore[start.id],
    });
    const pointById = {};
    points.forEach((p) => {
        pointById[p.id] = p;
    });
    let current;
    while (Object.keys(openSet).length) {
        const minId = sortedOpenSet.minId(false);
        if (minId) {
            current = openSet[minId];
        }
        else {
            break;
        }
        // point with the smallest fScore in openSet is the end
        if (current.id === goal.id) {
            // ending condition
            const pathPoints = [];
            (0, exports.reconstructPath)(pathPoints, pointById, cameFrom, goal.id);
            return pathPoints;
        }
        delete openSet[current.id];
        sortedOpenSet.remove(current.id);
        closedSet.push(current);
        const neighborPoints = (0, exports.getNeighborPoints)(points, current, sBBox, tBBox);
        const iterateNeighbors = (items) => {
            items.forEach((neighbor) => {
                if (closedSet.indexOf(neighbor) !== -1) {
                    return;
                }
                const neighborId = neighbor.id;
                if (!openSet[neighborId]) {
                    openSet[neighborId] = neighbor;
                }
                const tentativeGScore = fScore[current.id] + (0, math_1.manhattanDist)(current, neighbor); // + manhattanDist(neighbor, goal);
                if (gScore[neighborId] && tentativeGScore >= gScore[neighborId]) {
                    sortedOpenSet.add({
                        id: neighborId,
                        value: fScore[neighborId],
                    });
                    return;
                }
                cameFrom[neighborId] = current.id;
                gScore[neighborId] = tentativeGScore;
                fScore[neighborId] =
                    gScore[neighborId] +
                        (0, exports.heuristicCostEstimate)(neighbor, goal, start, os, ot);
                sortedOpenSet.add({
                    id: neighborId,
                    value: fScore[neighborId],
                });
            });
        };
        iterateNeighbors(neighborPoints);
    }
    return [start, goal];
};
exports.pathFinderUsingWaypoints = pathFinderUsingWaypoints;
/**
 * Calculate the two points necessary to draw a rounded corner between three points, given a radius.
 */
const getBorderRadiusPoints = (p0, p1, p2, r) => {
    const d0 = (0, math_1.manhattanDist)(p0, p1);
    const d1 = (0, math_1.manhattanDist)(p2, p1);
    const maxR = Math.min(d0, d1) / 2; // calculate the minimum possible radius
    r = Math.min(r, maxR); // use the smaller value between the given radius and the minimum radius
    const ps = {
        x: p1.x - (r / d0) * (p1.x - p0.x),
        y: p1.y - (r / d0) * (p1.y - p0.y),
    };
    const pt = {
        x: p1.x - (r / d1) * (p1.x - p2.x),
        y: p1.y - (r / d1) * (p1.y - p2.y),
    };
    return [ps, pt];
};
exports.getBorderRadiusPoints = getBorderRadiusPoints;
/** Draw a polyline path */
const pointsToPolyline = (points, borderRadius, z) => {
    const pathSegments = [];
    const startPoint = points[0];
    pathSegments.push(`M ${startPoint.x} ${startPoint.y}`);
    (0, util_1.each)(points, (p, i) => {
        const p1 = points[i + 1];
        const p2 = points[i + 2];
        if (p1 && p2) {
            if ((0, point_1.isBending)(p, p1, p2) && borderRadius) {
                const [ps, pt] = (0, exports.getBorderRadiusPoints)(p, p1, p2, borderRadius);
                pathSegments.push(`L ${ps.x} ${ps.y}`);
                pathSegments.push(`Q ${p1.x} ${p1.y} ${pt.x} ${pt.y}`);
                pathSegments.push(`L ${pt.x} ${pt.y}`);
            }
            else {
                pathSegments.push(`L ${p1.x} ${p1.y}`);
            }
        }
        else if (p1) {
            pathSegments.push(`L ${p1.x} ${p1.y}`);
        }
    });
    if (z) {
        pathSegments.push('Z');
    }
    return pathSegments.join(' ');
};
exports.pointsToPolyline = pointsToPolyline;
/**
 * Get polyline path
 * @param id current edge id
 * @param points list of vertices
 * @param radius radius of the corner
 * @param z  whether is closed from the last point to the first point
 * @returns
 */
const getPolylinePath = (id, points, radius, z) => {
    if (!points || points.length < 2) {
        return 'M 0 0 L 0 0'; // Cannot draw a single point
    }
    const newPoints = (0, exports.removeRedundantPoint)(points);
    // Record polyline points
    // polylineHistory.addPath(id, newPoints);
    return (0, exports.pointsToPolyline)(newPoints, radius, z);
};
exports.getPolylinePath = getPolylinePath;
/**
 * Get key waypoints between two points
 * @param start 2d start point
 * @param end 2d end point
 * @param offset offset
 * @param sNode source node
 * @param tNode target node
 * @returns key waypoints between two points
 */
const getPolylinePoints = (start, end, offset, sNode, tNode) => {
    let sBBox, tBBox;
    // todo: combo
    if (!sNode || !sNode.getType()) {
        sBBox = (0, bbox_1.getBBoxFromPoint)(start);
    }
    else {
        sBBox = sNode === null || sNode === void 0 ? void 0 : sNode.getBBox();
    }
    if (!tNode || !tNode.getType()) {
        tBBox = (0, bbox_1.getBBoxFromPoint)(end);
    }
    else {
        tBBox = tNode === null || tNode === void 0 ? void 0 : tNode.getBBox();
    }
    const sxBBox = (0, bbox_1.getExpandedBBox)(sBBox, offset);
    const txBBox = (0, bbox_1.getExpandedBBox)(tBBox, offset);
    const sPoint = (0, bbox_1.getExpandedBBoxPoint)(sxBBox, start, end);
    const tPoint = (0, bbox_1.getExpandedBBoxPoint)(txBBox, end, start);
    const lineBBox = (0, bbox_1.getBBoxFromPoints)([sPoint, tPoint]);
    const sMixBBox = (0, bbox_1.mergeBBox)(sxBBox, lineBBox);
    const tMixBBox = (0, bbox_1.mergeBBox)(txBBox, lineBBox);
    let connectPoints = [];
    connectPoints = connectPoints
        .concat((0, bbox_1.getBBoxPoints)(sMixBBox))
        .concat((0, bbox_1.getBBoxPoints)(tMixBBox));
    const centerPoint = {
        x: (start.x + end.x) / 2,
        y: (start.y + end.y) / 2,
    };
    [lineBBox, sMixBBox, tMixBBox].forEach((bbox) => {
        connectPoints = connectPoints.concat((0, bbox_1.getBBoxCrossPointsByPoint)(bbox, centerPoint).filter((p) => (0, bbox_1.isPointOutsideBBox)(p, sxBBox) && (0, bbox_1.isPointOutsideBBox)(p, txBBox)));
    });
    [
        {
            x: sPoint.x,
            y: tPoint.y,
        },
        {
            x: tPoint.x,
            y: sPoint.y,
        },
    ].forEach((p) => {
        // impossible!!
        if ((0, bbox_1.isPointOutsideBBox)(p, sxBBox) &&
            (0, bbox_1.isPointOutsideBBox)(p, txBBox) // &&
        // isPointInsideBBox(p, sMixBBox) && isPointInsideBBox(p, tMixBBox)
        ) {
            connectPoints.push(p);
        }
    });
    connectPoints.unshift(sPoint);
    connectPoints.push(tPoint);
    connectPoints = (0, exports.simplifyPolylinePoints)(connectPoints);
    const pathPoints = (0, exports.pathFinderUsingWaypoints)(connectPoints, sPoint, tPoint, sBBox, tBBox, start, end);
    pathPoints.unshift(start);
    pathPoints.push(end);
    return (0, exports.simplifyPolylinePoints)(pathPoints);
};
exports.getPolylinePoints = getPolylinePoints;
/**
 * Remove the intermediate points that are consecutively the same as x and different from y; remove the intermediate points that are consecutively the same as y but different from x
 * @param points list of 2d points
 * @returns
 */
const removeRedundantPoint = (points) => {
    if (!(points === null || points === void 0 ? void 0 : points.length))
        return points;
    const beginPoint = points[points.length - 1];
    const current = {
        x: beginPoint.x,
        y: beginPoint.y,
    };
    let continueSameX = [beginPoint];
    let continueSameY = [beginPoint];
    for (let i = points.length - 2; i >= 0; i--) {
        const point = points[i];
        if (point.x === current.x) {
            continueSameX.push(point);
        }
        else {
            continueSameX = [point];
            current.x = point.x;
        }
        if (point.y === current.y) {
            continueSameY.push(point);
        }
        else {
            continueSameY = [point];
            current.y = point.y;
        }
        if (continueSameX.length > 2) {
            const removeIdx = points.indexOf(continueSameX[1]);
            if (removeIdx > -1)
                points.splice(removeIdx, 1);
            continue;
        }
        if (continueSameY.length > 2) {
            const removeIdx = points.indexOf(continueSameY[1]);
            if (removeIdx > -1)
                points.splice(removeIdx, 1);
        }
    }
    return points;
};
exports.removeRedundantPoint = removeRedundantPoint;
/**
 * sorted array ascendly
 * add new item to proper index when calling add
 */
class SortedArray {
    constructor() {
        this.arr = [];
        this.map = {};
        this.arr = [];
        this.map = {};
    }
    _innerAdd(item, length) {
        const idxRange = [0, length - 1];
        while (idxRange[1] - idxRange[0] > 1) {
            const midIdx = Math.floor((idxRange[0] + idxRange[1]) / 2);
            if (this.arr[midIdx].value > item.value) {
                idxRange[1] = midIdx;
            }
            else if (this.arr[midIdx].value < item.value) {
                idxRange[0] = midIdx;
            }
            else {
                this.arr.splice(midIdx, 0, item);
                this.map[item.id] = true;
                return;
            }
        }
        this.arr.splice(idxRange[1], 0, item);
        this.map[item.id] = true;
    }
    add(item) {
        // 已经存在，先移除
        delete this.map[item.id];
        const length = this.arr.length;
        if (!length) {
            this.arr.push(item);
            this.map[item.id] = true;
            return;
        }
        // 比最后一个大，加入尾部
        if (this.arr[length - 1].value < item.value) {
            this.arr.push(item);
            this.map[item.id] = true;
            return;
        }
        this._innerAdd(item, length);
    }
    // only remove from the map to avoid cost
    // clear the invalid (not in the map) item when calling minId(true)
    remove(id) {
        if (!this.map[id])
            return;
        delete this.map[id];
    }
    _clearAndGetMinId() {
        let res;
        for (let i = this.arr.length - 1; i >= 0; i--) {
            if (this.map[this.arr[i].id])
                res = this.arr[i].id;
            else
                this.arr.splice(i, 1);
        }
        return res;
    }
    _findFirstId() {
        while (this.arr.length) {
            const first = this.arr.shift();
            if (this.map[first.id])
                return first.id;
        }
    }
    minId(clear) {
        if (clear) {
            return this._clearAndGetMinId();
        }
        else {
            return this._findFirstId();
        }
    }
}
exports.SortedArray = SortedArray;
/** Define a quadtree */
class QuadTree {
    constructor(boundary, capacity) {
        this.boundary = boundary;
        this.edges = [];
        this.divided = false;
        this.capacity = capacity;
    }
    // Insert a new edge
    insert(edge) {
        var _a, _b, _c, _d;
        if (!this.boundary.intersects(edge.bbox)) {
            return false;
        }
        if (this.edges.length < this.capacity) {
            this.edges.push(edge);
            return true;
        }
        if (!this.divided) {
            this.subdivide();
        }
        return (((_a = this.northeast) === null || _a === void 0 ? void 0 : _a.insert(edge)) ||
            ((_b = this.northwest) === null || _b === void 0 ? void 0 : _b.insert(edge)) ||
            ((_c = this.southeast) === null || _c === void 0 ? void 0 : _c.insert(edge)) ||
            ((_d = this.southwest) === null || _d === void 0 ? void 0 : _d.insert(edge)) ||
            false);
    }
    // Divide the quadtree into four subtrees
    subdivide() {
        const x = this.boundary.center[0];
        const y = this.boundary.center[1];
        const hw = this.boundary.halfExtents[0] / 2;
        const hh = this.boundary.halfExtents[1] / 2;
        const nwBoundary = new g_1.AABB();
        nwBoundary.setMinMax([x - 2 * hw, y - 2 * hh, 0], [x, y, 0]);
        this.northwest = new QuadTree(nwBoundary, this.capacity);
        const neBoundary = new g_1.AABB();
        neBoundary.setMinMax([x, y - 2 * hh, 0], [x + 2 * hw, y, 0]);
        this.northeast = new QuadTree(neBoundary, this.capacity);
        const swBoundary = new g_1.AABB();
        swBoundary.setMinMax([x - 2 * hw, y, 0], [x, y + 2 * hh, 0]);
        this.southwest = new QuadTree(swBoundary, this.capacity);
        const seBoundary = new g_1.AABB();
        seBoundary.setMinMax([x, y, 0], [x + 2 * hw, y + 2 * hh, 0]);
        this.southeast = new QuadTree(seBoundary, this.capacity);
        this.divided = true;
    }
    // Query all edges within the specified range
    queryRange(range) {
        var _a, _b, _c, _d;
        if (!range.intersects(this.boundary)) {
            return [];
        }
        let found = this.edges.filter((edge) => edge.bbox.intersects(range));
        if (this.divided) {
            found = found.concat(((_a = this.northeast) === null || _a === void 0 ? void 0 : _a.queryRange(range)) || [], ((_b = this.northwest) === null || _b === void 0 ? void 0 : _b.queryRange(range)) || [], ((_c = this.southeast) === null || _c === void 0 ? void 0 : _c.queryRange(range)) || [], ((_d = this.southwest) === null || _d === void 0 ? void 0 : _d.queryRange(range)) || []);
        }
        return found;
    }
}
exports.QuadTree = QuadTree;
/** Collision detector */
class EdgeCollisionChecker {
    constructor(quadTree) {
        this.quadTree = quadTree;
    }
    /** Check if node and edge intersect */
    getCollidingEdges(nodeBBox) {
        // Get edge's bounding box intersects moving node's bounding box
        const potentialCollisions = this.quadTree.queryRange(nodeBBox);
        // todo: edge intersects node bbox
        return potentialCollisions;
    }
}
exports.EdgeCollisionChecker = EdgeCollisionChecker;
//# sourceMappingURL=polyline.js.map